/*
Module : DTime+CDate.h
Purpose: Defines the interface for the DTime+ CDate class
Created: PJN / 01-05-2006
History: PJN / 01-01-2018 1. Fixed a bug in CDate::GetWindowsFormatString where it was calling a non existent
                          String::GetLength method

Copyright (c) 2006 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#if _MSC_VER > 1000
#pragma once
#endif //#if _MSC_VER > 1000

#ifndef __DTIMEPLUSCDATE_H__
#define __DTIMEPLUSCDATE_H__


//////////////////// Includes /////////////////////////////////////////////////

#include <cstdint>
#include <ctime>
#include <sstream>
#include <iomanip>
#include <climits>
#include <cstring>
#include <array>
#include "DTime+Defines.h"


//////////////////// Classes //////////////////////////////////////////////////

namespace DTimePlus
{

//A Date class with a granularity of 1 Day in the Gregorian Calendar
template<class TCALENDAR = CGregorianCalendar>
class CDate : public TCALENDAR
{
public:
//Constructors
#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CDate() noexcept
  {
    Set();
  }

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CDate(int32_t Year, uint16_t month, uint16_t Day, bool bDoAssert = true) noexcept
  {
    Set(Year, month, Day, bDoAssert);
  }

#ifdef _WIN32
#pragma warning(suppress: 26495)
  CDate(const SYSTEMTIME& SystemTime, bool bDoAssert = true) noexcept
  {
    Set(SystemTime, bDoAssert);
  }
#endif //#ifdef _WIN32

#ifdef _WIN32
#pragma warning(suppress: 26495)
  CDate(const FILETIME& FileTime, bool bDoAssert = true) noexcept
  {
    Set(FileTime, bDoAssert);
  }
#endif //#ifdef _WIN32

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CDate(int32_t Year, uint16_t month, uint16_t WeekOfMonth, DayOfWeek dow, bool bDoAssert = true) noexcept
  {
    Set(Year, month, WeekOfMonth, dow, bDoAssert);
  }

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CDate(int32_t Days, DateEpoch e, bool bDoAssert = true) noexcept
  {
    Set(Days, e, bDoAssert);
  }

  CDate(const CDate<TCALENDAR>&) = default;
  CDate(CDate<TCALENDAR>&&) = default;

#ifdef _AFX
#pragma warning(suppress: 26495)
  CDate(const CTime& ctime, bool bDoAssert = true) noexcept
  {
    Set(ctime, bDoAssert);
  }
#endif //#ifdef _AFX

#ifdef _AFX
#pragma warning(suppress: 26495)
  CDate(const COleDateTime& oleTime, bool bDoAssert = true) noexcept
  {
    Set(oleTime, bDoAssert);
  }
#endif //#ifdef _AFX

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CDate(const String& sDate, const String& sFormat, bool bDoAssert = true)
  {
    Set(sDate, sFormat, bDoAssert);
  }

  ~CDate() = default;

  //Set operators
  CDate<TCALENDAR>& Set() noexcept
  {
    //Create an invalid value
    m_lDays = INT32_MIN;
    return *this;
  }

  CDate<TCALENDAR>& Set(int32_t Year, uint16_t month, uint16_t Day, bool bDoAssert = true) noexcept
  {
    if (!TCALENDAR::DMYToNum(Year, month, Day, m_lDays, bDoAssert))
      Set();
    return *this;
  }

#ifdef _DEBUG
  CDate<TCALENDAR>& Set(int32_t Year, uint16_t month, uint16_t WeekOfMonth, DayOfWeek dow, bool bDoAssert = true) noexcept
#else
  CDate<TCALENDAR>& Set(int32_t Year, uint16_t month, uint16_t WeekOfMonth, DayOfWeek dow, bool /*bDoAssert*/ = true) noexcept
#endif //#ifdef _DEBUG
  {
    //Validate our parameters
    if ((WeekOfMonth < 1) || (WeekOfMonth > 5) || (dow < DayOfWeek::FirstDayOfWeek) || (dow > DayOfWeek::LastDayOfWeek))
    {
    #ifdef _DEBUG
      if (bDoAssert)
        DTIME_ASSERT(false);
    #endif //#ifdef _DEBUG
      Set();
      return *this;
    }

    if (WeekOfMonth < 5)
    {
      const CDate<TCALENDAR> FirstMonth{Year, month, 1};
      if (!FirstMonth.IsValid())
      {
      #ifdef _DEBUG
        if (bDoAssert)
          DTIME_ASSERT(false);
      #endif //#ifdef _DEBUG
        Set();
        return *this;
      }
      const auto dow1{static_cast<uint16_t>(FirstMonth.GetDayOfWeek())};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      const auto Day{static_cast<uint16_t>(((static_cast<int>(dow) - dow1 + 7) % 7) + (7 * (WeekOfMonth - 1)) + 1)};
      Set(Year, month, Day);
    }
    else //5 means the last week of the month
    {
      DTIME_ASSERT(WeekOfMonth == 5);

#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      auto NewMonth{static_cast<uint16_t>(month + 1)};
      int32_t NewYear{Year};
      if (NewMonth > 12)
      {
        NewMonth = 1;
        ++NewYear;
      }

      Set(NewYear, NewMonth, 1, dow);
      AddWeek(-1);
    }
    return *this;
  }

#ifdef _DEBUG
  CDate<TCALENDAR>& Set(int32_t Days, DateEpoch e, bool bDoAssert = true) noexcept
#else
  CDate<TCALENDAR>& Set(int32_t Days, DateEpoch e, bool /*bDoAssert*/ = true) noexcept
#endif //#ifdef _DEBUG
  {
    switch (e)
    {
      case DateEpoch::EPOCH_GREG:
      {
        m_lDays = Days;
        break;
      }
      case DateEpoch::EPOCH_JD:
      {
        m_lDays = Days - 2299161; //-2299161 corresponds to the GDN of JD Epoch namely 1 January -4712 
        break;
      }
      case DateEpoch::EPOCH_MJD:
      {
        m_lDays = Days + 100839; //100839 corresponds to the GDN of MJD Epoch namely 17 November 1858 
        break;
      }
      case DateEpoch::EPOCH_1900:
      {
        m_lDays = Days + 115860; //115860 corresponds to the GDN of 1 January 1900
        break;
      }
      case DateEpoch::EPOCH_1950:
      {
        m_lDays = Days + 134122; //134122 corresponds to the GDN of 1 January 1950
        break;
      }
      case DateEpoch::EPOCH_CTIME:
      {
        m_lDays = Days + 141427; //141427 corresponds to the GDN of 1 January 1970
        break;
      }
      case DateEpoch::EPOCH_2000:
      {
        m_lDays = Days + 152384; //152384 corresponds to the GDN of 1 January 2000
        break;
      }
      default:
      {
        DTIME_ASSERT(false); //should not occur
        break;
      }
    }
    if (m_lDays > 2146905911) //Largest valid GDN
    {
    #ifdef _DEBUG
      if (bDoAssert)
        DTIME_ASSERT(false);
    #endif //#ifdef _DEBUG
      Set();
    }
    return *this;
  }

#ifdef _WIN32
  CDate<TCALENDAR>& Set(const SYSTEMTIME& SystemTime, bool bDoAssert = true) noexcept
  {
    Set(SystemTime.wYear, SystemTime.wMonth, SystemTime.wDay, bDoAssert);
    return *this;
  }
#endif //#ifdef _WIN32

#ifdef _WIN32
  CDate<TCALENDAR>& Set(const FILETIME& FileTime, bool bDoAssert = true) noexcept
  {
  #ifndef _DEBUG
    UNREFERENCED_PARAMETER(bDoAssert);
  #endif //#ifndef _DEBUG

    SYSTEMTIME st{};
    if (::FileTimeToSystemTime(&FileTime, &st))
      Set(st);
    else
    {
    #ifdef _DEBUG
      if (bDoAssert)
        DTIME_ASSERT(false);
    #endif //#ifdef _DEBUG
      Set();
    }
    return *this;
  }
#endif //#ifdef _WIN32

#ifdef _AFX
  CDate<TCALENDAR>& Set(const CTime& ctime, bool bDoAssert = true) noexcept
  {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Set(static_cast<int32_t>(ctime.GetTime() / 86400), DateEpoch::EPOCH_CTIME, bDoAssert);
    return *this;
  }
#endif //#ifdef _AFX

#ifdef _AFX
  CDate<TCALENDAR>& Set(const COleDateTime& oleTime, bool bDoAssert = true) noexcept
  {
    if (oleTime.GetStatus() == COleDateTime::valid)
    {
      SYSTEMTIME st{};
      if (oleTime.GetAsSystemTime(st))
        Set(st, bDoAssert);
      else
        Set();
    }
    else
      Set();
    return *this;
  }
#endif //#ifdef _AFX

  CDate<TCALENDAR>& SetFromCollate(int32_t Date, bool bDoAssert = true)
  {
    dt_ostringstream ss;
    ss << Date;
    String sDate(ss.str());

    const auto nLen{sDate.length()};
    String sDay{sDate.substr(nLen - 2, 2)};
    String sMonth{sDate.substr(nLen - 4, 2)};
    String sYear{sDate.substr(0, nLen - 4)};

#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    const auto Day{static_cast<uint16_t>(_ttoi(sDay.c_str()))};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    const auto month{static_cast<uint16_t>(_ttoi(sMonth.c_str()))};
    const int32_t Year{_ttol(sYear.c_str())};
#ifdef _MSC_VER
#pragma warning(suppress: 26487 26489)
#endif //#ifdef _MSC_VER
    return Set(Year, month, Day, bDoAssert);
  }

  /*
  sDate is a legal date string as defined by sFormat.

  sFormat is a 5 character string where the first 3 indicate the
  order of the month/day/year, the fourth character indicates the length
  of the year, and the fifth character indicates the separator.

  For example:

  sFormat   sDate          Result

  "dmy2/"   "20/03/96"     20th March 1996
  "mdy4-"   "03-28-1996"   28th March 1996

  This feature allows for the creation of CDate's from a string
  as entered in a masked edit field or from a parsed report
  */
  CDate<TCALENDAR>& Set(const String& sDate, const String& sFormat, bool bDoAssert = true)
  {
    if (sFormat.length() != 5)
    {
    #ifdef _DEBUG
      if (bDoAssert)
        DTIME_ASSERT(false); //sFormat string provided was invalid!
    #endif //#ifdef _DEBUG
      Set();
    }
    else
    {
      uint16_t Day{0};
      uint16_t month{1};
      int32_t Year{0};

#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
      std::array<TCHAR, 2> tokens{ sFormat[4], _T('\0') };
      String sD{sDate};
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
      TCHAR* pszDate{sD.data()};
    #ifdef _MSC_VER
      TCHAR* pszNextToken{nullptr};
      TCHAR* pszToken{_tcstok_s(pszDate, tokens.data(), &pszNextToken)};
    #else
      TCHAR* pszToken{_tcstok(pszDate, tokens.data())};
    #endif //#ifdef _MSC_VER

      for (int i{0}; i<3; i++)
      {
        DTIME_ASSUME(pszToken != nullptr);
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
        switch (sFormat[i])
        {
          case _T('d'):
          {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
            Day = static_cast<uint16_t>(_ttoi(pszToken));
            break;
          }
          case _T('m'):
          {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
            month = static_cast<uint16_t>(_ttoi(pszToken));
            break;
          }
          case _T('y'):
          {
            Year  = _ttol(pszToken);
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
            const TCHAR c{sFormat[3]};
            DTIME_ASSERT((c == _T('2')) || (c == _T('4')));
            if (c == _T('2'))
            {
              if (Year >= 50)
                Year += 1900;
              else if (Year < 50)
                Year += 2000;
            }
            break;
          }
          default:
          {
            DTIME_ASSERT(false);
            break;
          }
        }

      #ifdef _MSC_VER
        pszToken = _tcstok_s(nullptr, tokens.data(), &pszNextToken);
      #else
        pszToken = _tcstok(nullptr, tokens.data());
      #endif //#ifdef _MSC_VER
      }
      Set(Year, month, Day, bDoAssert);
    }
    return *this;
  }

//Static constructors
  static CDate<TCALENDAR> CurrentSystemDate(bool bDoAssert = true) noexcept
  {
    //Get the current time from the CRT
    time_t now{0};
    if (time(&now) != -1)
    {
      //PC clock will always return time in Gregorian calendar
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      const CDate<CGregorianCalendar> date{static_cast<int32_t>(now / 86400), DateEpoch::EPOCH_CTIME, bDoAssert};
      if (date.IsValid())
        return CDate<TCALENDAR>{date.GetValue(), DateEpoch::EPOCH_GREG};
      else
        return CDate<TCALENDAR>{};
    }
    else
      return CDate<TCALENDAR>{};
  }

  static CDate<TCALENDAR> CurrentLocalDate(bool bDoAssert = true) noexcept
  {
    //Get the current time from the CRT
    time_t now{0};
    if (time(&now) != -1)
    {
      //Convert to local time
    #ifdef _MSC_VER
      tm lt{};
      const errno_t e{localtime_s(&lt, &now)};
      if (e != 0)
        return CDate<TCALENDAR>{};
      const tm* pLocalTime{&lt};
    #else
      const tm* pLocalTime{localtime(&now)};
    #endif //#ifdef _MSC_VER
      if (pLocalTime == nullptr)
        return CDate<TCALENDAR>{};
      else
      {
        //PC clock will always return time in Gregorian calendar
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
        const CDate<CGregorianCalendar> date{static_cast<int32_t>(pLocalTime->tm_year + 1900), static_cast<uint16_t>(pLocalTime->tm_mon + 1), static_cast<uint16_t>(pLocalTime->tm_mday), bDoAssert};
        if (date.IsValid())
          return CDate<TCALENDAR>{date.GetValue(), DateEpoch::EPOCH_GREG};
        else
          return CDate<TCALENDAR>{};
      }
    }
    else
      return CDate<TCALENDAR>{};
  }

  static CDate<TCALENDAR> JDEpoch() noexcept
  {
    return CDate<TCALENDAR>{0, DateEpoch::EPOCH_JD};
  }

  static CDate<TCALENDAR> MJDEpoch() noexcept
  {
    return CDate<TCALENDAR>{0, DateEpoch::EPOCH_MJD};
  }

  static CDate<TCALENDAR> Epoch1900() noexcept
  {
    return CDate<TCALENDAR>{0, DateEpoch::EPOCH_1900};
  }

  static CDate<TCALENDAR> Epoch1950() noexcept
  {
    return CDate<TCALENDAR>{0, DateEpoch::EPOCH_1950};
  }

  static CDate<TCALENDAR> EpochCTime() noexcept
  {
    return CDate<TCALENDAR>{0, DateEpoch::EPOCH_CTIME};
  }

  static CDate<TCALENDAR> Epoch2000() noexcept
  {
    return CDate<TCALENDAR>{0, DateEpoch::EPOCH_2000};
  }

  static CDate<TCALENDAR> GregorianEpoch() noexcept
  {
    return CDate<TCALENDAR>{0, DateEpoch::EPOCH_GREG};
  }

//Static operations
#ifdef _WIN32
#pragma warning(suppress: 26477)
  static String GetWindowsFormatString(LPCWSTR lpLocaleName = LOCALE_NAME_USER_DEFAULT)
  {
    //Is there to be a leading 0 on the day?
    bool bDayLZ{false};
    String sBuf{SafeGetLocaleInfo(lpLocaleName, LOCALE_IDAYLZERO)};
    if (sBuf == _T("0")) //No leading zeros for days
      bDayLZ = false;
    else if (sBuf == _T("1")) //Leading zeros for days
      bDayLZ = true;
    else
      DTIME_ASSERT(false);

    //Is there to be a leading 0 on the month?
    bool bMoLZ{true};
    sBuf = SafeGetLocaleInfo(lpLocaleName, LOCALE_IMONLZERO);
    if (sBuf == _T("0")) //No leading zeros for months
      bMoLZ = false;
    else if (sBuf == _T("1")) //Leading zeros for months
      bMoLZ = true;
    else
      DTIME_ASSERT(false);

    //Find out if only 2 digits are to be displayed for the year
    bool b2DigitYear{false};
    sBuf = SafeGetLocaleInfo(lpLocaleName, LOCALE_ICENTURY);
    if (sBuf == _T("0")) //Abbreviated 2-digit century
      b2DigitYear = true;
    else if (sBuf == _T("1")) //Full century
      b2DigitYear = false;
    else
      DTIME_ASSERT(false);

    //Get the date separator character
    String sSep{_T("/")};
    sBuf = SafeGetLocaleInfo(lpLocaleName, LOCALE_SDATE);
    if (sBuf.length())
      sSep = sBuf;

    //Retrieve the short date formatting string
    sBuf = SafeGetLocaleInfo(lpLocaleName, LOCALE_IDATE);

    //What will be the return value from this function
    String sFormatString;

    //Format according to the short date formatting string
    if (sBuf == _T("0")) //MDY format
    {
      if (bMoLZ)
        sFormatString += _T("%m");
      else
        sFormatString += _T("%#m");
      sFormatString += sSep;
      if (bDayLZ)
        sFormatString += _T("%d");
      else
        sFormatString += _T("%#d");
      sFormatString += sSep;
      if (b2DigitYear)
        sFormatString += _T("%y");
      else
        sFormatString += _T("%Y");
    }
    else if (sBuf == _T("1")) //DMY format
    {
      if (bDayLZ)
        sFormatString += _T("%d");
      else
        sFormatString += _T("%#d");
      sFormatString += sSep;
      if (bMoLZ)
        sFormatString += _T("%m");
      else
        sFormatString += _T("%#m");
      sFormatString += sSep;
      if (b2DigitYear)
        sFormatString += _T("%y");
      else
        sFormatString += _T("%Y");
    }
    else
    {
      DTIME_ASSERT(sBuf != _T("2")); //No Recognized format available

      //YMD Format
      if (b2DigitYear)
        sFormatString += _T("%y");
      else
        sFormatString += _T("%Y");
      sFormatString += sSep;
      if (bMoLZ)
        sFormatString += _T("%m");
      else
        sFormatString += _T("%#m");
      sFormatString += sSep;
      if (bDayLZ)
        sFormatString += _T("%d");
      else
        sFormatString += _T("%#d");
    }

    return sFormatString;
  }
#endif //#ifdef _WIN32

//Operations
  DTIME_NODISCARD uint16_t DaysSinceFirstOfYear() const noexcept
  {
    DTIME_ASSERT(IsValid());

#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    return static_cast<uint16_t>(*this - CDate<TCALENDAR>{GetYear(), 1, 1});
  }

  DTIME_NODISCARD uint16_t DaysSinceEndPreviousYear() const noexcept
  {
    DTIME_ASSERT(IsValid());

#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    return static_cast<uint16_t>(DaysSinceFirstOfYear() + 1);
  }

  DTIME_NODISCARD uint16_t GetDay() const noexcept
  {
    DTIME_ASSERT(IsValid());

    return GetDate().Day;
  }

  DTIME_NODISCARD uint16_t GetMonth() const noexcept
  {
    DTIME_ASSERT(IsValid());

    return GetDate().Month;
  }

  DTIME_NODISCARD int32_t GetYear() const noexcept
  {
    DTIME_ASSERT(IsValid());

    return GetDate().Year;
  }

  int32_t GetCEBCEYear(bool& IsCE) const noexcept
  {
    DTIME_ASSERT(IsValid());

    const int32_t Year{GetDate().Year};
    if (Year > 0)
    {
      IsCE = true;
      return Year;
    }
    else
    {
      IsCE = false;
      return labs(GetYear() - 1);
    }
  }

  static uint16_t Get2DigitYear(int32_t Year) noexcept
  {
    Year = labs(Year);
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    return static_cast<uint16_t>(Year - ((Year/100) * 100));
  }

  DTIME_NODISCARD uint16_t Get2DigitYear() const
  {
    DTIME_ASSERT(IsValid());

    return Get2DigitYear(GetYear());
  }

  DTIME_NODISCARD int32_t JD() const noexcept
  {
    DTIME_ASSERT(IsValid());

    return *this - JDEpoch();
  }

  DTIME_NODISCARD int32_t GetValue() const noexcept
  {
    DTIME_ASSERT(IsValid());

    return m_lDays;
  }

  DTIME_NODISCARD DateS GetDate() const noexcept
  {
    DTIME_ASSERT(IsValid());

    DateS ds{};
    TCALENDAR::NumToDMY(m_lDays, ds.Year, ds.Month, ds.Day);
    return ds;
  }

  DTIME_NODISCARD bool IsLeap() const noexcept
  {
    DTIME_ASSERT(IsValid());

    return TCALENDAR::IsLeap(GetYear());
  }

  DTIME_NODISCARD uint16_t DaysInYear() const noexcept
  {
    DTIME_ASSERT(IsValid());

    return TCALENDAR::DaysInYear(GetYear());
  }

  DTIME_NODISCARD uint16_t DaysInMonth() const noexcept
  {
    DTIME_ASSERT(IsValid());

    const DateS ds{GetDate()};
    return TCALENDAR::DaysInMonth(ds.Year, ds.Month);
  }

  void AddYear(int Years) noexcept
  {
    DTIME_ASSERT(IsValid());

    const int32_t Months{Years * 12};
    if ((Months >= INT_MIN) && (Months <= INT_MAX))
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      AddMonth(static_cast<int>(Months));
    }
    else
    {
      DTIME_ASSERT(false);
      Set();
    }
  }

  void AddMonth(int Months) noexcept
  {
    DTIME_ASSERT(IsValid());

    int NewMonth{GetMonth()};
    int32_t NewYear{GetYear()};
    NewMonth += Months;
    if ((NewMonth > 12) || (NewMonth < 1))
    {
      NewYear += NewMonth / 12;
      NewMonth = NewMonth % 12;
    }
    if (NewMonth < 1)
    {
      --NewYear;
      NewMonth += 12;
    }

    uint16_t Day{GetDay()};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    const uint16_t MaxDays{TCALENDAR::DaysInMonth(NewYear, static_cast<uint16_t>(NewMonth))};
    if (Day > MaxDays)
      Day = MaxDays;
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Set(NewYear, static_cast<uint16_t>(NewMonth), Day);
  }

  void AddWeek(int Weeks) noexcept
  {
    DTIME_ASSERT(IsValid());

#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    *this += 7L * (static_cast<int32_t>(Weeks));
  }

  DTIME_NODISCARD DayOfWeek GetDayOfWeek() const noexcept
  {
    DTIME_ASSERT(IsValid());

    return static_cast<DayOfWeek>(lmod(m_lDays+12, 7) + 1);
  }

  DTIME_NODISCARD String GetFullStringDayOfWeek() const
  {
    DTIME_ASSERT(IsValid());

    return TCALENDAR::GetFullStringDayOfWeek(GetDayOfWeek());
  }

  DTIME_NODISCARD String GetAbrStringDayOfWeek() const
  {
    DTIME_ASSERT(IsValid());

    return TCALENDAR::GetAbrStringDayOfWeek(GetDayOfWeek());
  }

  DTIME_NODISCARD String GetFullStringMonth() const
  {
    DTIME_ASSERT(IsValid());

    return TCALENDAR::GetFullStringMonth(GetMonth());
  }

  DTIME_NODISCARD String GetAbrStringMonth() const
  {
    DTIME_ASSERT(IsValid());

    return TCALENDAR::GetAbrStringMonth(GetMonth());
  }

#ifdef _WIN32
  bool GetAsSystemTime(SYSTEMTIME& SystemTime) const noexcept
  {
    if (!IsValid())
      return false;

    const DateS ds{GetDate()};

    //If our date cannot fit into the range allowed by SYSTEMTIME
    if ((ds.Year < 0) || (ds.Year > USHRT_MAX))
      return false;

#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    SystemTime.wYear = static_cast<uint16_t>(ds.Year);
    SystemTime.wMonth = ds.Month;
    SystemTime.wDayOfWeek = static_cast<WORD>(GetDayOfWeek()) - 1; //SYSTEMTIME uses 0 based indices
    SystemTime.wDay = ds.Day;
    SystemTime.wHour = 0;
    SystemTime.wMinute = 0;
    SystemTime.wSecond = 0;
    SystemTime.wMilliseconds = 0;
    return true;
  }
#endif //#ifdef _WIN32

#ifdef _WIN32
  bool GetAsFileTime(FILETIME& FileTime) const noexcept
  {
    SYSTEMTIME SystemTime{};
    bool bSuccess{GetAsSystemTime(SystemTime)};
    bSuccess = bSuccess && ::SystemTimeToFileTime(&SystemTime, &FileTime);
    return bSuccess;
  }
#endif //#ifdef _WIN32

  bool GetAsTM(tm& time) const noexcept
  {
    //What will be the return value from this function (assume the best)
    bool rVal{true};

    if (!IsValid())
      return false;

    constexpr const int32_t MinYear{1900 + INT32_MIN};
    constexpr const int32_t MaxYear{INT32_MAX};

    //handle range errors
    const DateS ds{GetDate()};
    if ((ds.Year < MinYear) || (ds.Year > MaxYear))
      rVal = false;
    else
    {
      time.tm_sec = 0;
      time.tm_min = 0;
      time.tm_hour = 0;
      time.tm_isdst = 0;
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      time.tm_year = static_cast<int>(ds.Year - 1900);
      time.tm_mon = ds.Month - 1; //tm struct uses 0 based indices
      time.tm_wday = static_cast<int>(GetDayOfWeek()) - 1; //tm struct uses 0 based indices
      time.tm_mday = ds.Day;
      time.tm_yday = DaysSinceFirstOfYear(); //Returns days since Jan 1
    }
    return rVal;
  }

#ifdef _MSC_VER
#pragma warning(suppress: 26440)
#endif //#ifdef _MSC_VER
  DTIME_NODISCARD uint16_t GetWeekOfYear() const
  {
    DTIME_ASSERT(IsValid());

    const CDate<TCALENDAR> FirstOfYear{GetYear(), 1, 1, TCALENDAR::GetBeginningDayOfWeek()};
    const int32_t diff{*this - FirstOfYear};
    if (diff < 0)
      return 0;
    else
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      return static_cast<uint16_t>((diff / 7) + 1);
    }
  }

  DTIME_NODISCARD uint16_t GetWeekOfMonth() const
  {
    DTIME_ASSERT(IsValid());

    const CDate<TCALENDAR> FirstOfMonth{GetYear(), GetMonth(), 1, TCALENDAR::GetBeginningDayOfWeek()};
    const int32_t diff{*this - FirstOfMonth};
    if (diff < 0)
      return 0;
    else
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      return static_cast<uint16_t>((diff / 7) + 1);
    }
  }

  DTIME_NODISCARD bool IsValid() const noexcept
  {
    return (m_lDays != INT32_MIN);
  }

  DTIME_NODISCARD int32_t Collate() const noexcept
  {
    const DateS s{GetDate()};
    return (s.Year * 10000) + (s.Month * 100) + s.Day;
  }

  DTIME_NODISCARD CDate<CJulianCalendar> AsJulianCDate() const noexcept
  {
    CDate<CJulianCalendar> rVal;
    rVal.Set(m_lDays, DateEpoch::EPOCH_GREG);
    return rVal;
  }

  DTIME_NODISCARD CDate<CGregorianCalendar> AsGregorianCDate() const noexcept
  {
    CDate<CGregorianCalendar> rVal;
    rVal.Set(m_lDays, DateEpoch::EPOCH_GREG);
    return rVal;
  }

  DTIME_NODISCARD CDate<CHebrewCalendar> AsHebrewCDate() const noexcept
  {
    CDate<CHebrewCalendar> rVal;
    rVal.Set(m_lDays, DateEpoch::EPOCH_GREG);
    return rVal;
  }

//Creation of other CDate's from this instance
  DTIME_NODISCARD CDate<TCALENDAR> FirstThisMonth() const noexcept
  {
    const DateS s{GetDate()};
    return CDate<TCALENDAR>{s.Year, s.Month, 1};
  }

  DTIME_NODISCARD CDate<TCALENDAR> LastThisMonth() const noexcept
  {
    const DateS s{GetDate()};
    return CDate<TCALENDAR>{s.Year, s.Month, TCALENDAR::DaysInMonth(s.Year, s.Month)};
  }

  DTIME_NODISCARD CDate<TCALENDAR> FirstThisYear() const noexcept
  {
    const DateS s{GetDate()};
    return CDate<TCALENDAR>{s.Year, 1, 1};
  }

  DTIME_NODISCARD CDate<TCALENDAR> LastThisYear() const noexcept
  {
    const DateS s{GetDate()};
    return CDate<TCALENDAR>{s.Year, 12, TCALENDAR::DaysInMonth(s.Year, 12)};
  }

  CDate<TCALENDAR>& operator=(const CDate<TCALENDAR>&) = default;
  CDate<TCALENDAR>& operator=(CDate<TCALENDAR>&&) = default;

//Overloaded arithmetic operators
  CDate<TCALENDAR> operator+(int32_t Days) const noexcept
  {
    if (IsValid())
      return CDate<TCALENDAR>{GetValue() + Days, DateEpoch::EPOCH_GREG};
    else
      return CDate<TCALENDAR>{};
  }

  int32_t operator-(const CDate<TCALENDAR>& d) const noexcept
  {
    if (IsValid() && d.IsValid())
      return m_lDays - d.m_lDays;
    else
      return 0;
  }

  CDate operator-(int32_t Days) const noexcept
  {
    if (IsValid())
      return CDate<TCALENDAR>{GetValue() - Days, DateEpoch::EPOCH_GREG};
    else
      return CDate<TCALENDAR>{};
  }

  CDate<TCALENDAR>& operator+=(int32_t Days) noexcept
  {
    *this = *this + Days;
    return *this;
  }

  CDate<TCALENDAR>& operator-=(int32_t Days) noexcept
  {
    *this = *this - Days;
    return *this;
  }

  CDate<TCALENDAR>& operator++() noexcept
  {
    *this = *this + 1;
    return *this;
  }

  CDate<TCALENDAR>& operator--() noexcept
  {
    *this = *this - 1;
    return *this;
  }

//Overloaded equality operators
  bool operator==(const CDate<TCALENDAR>& d) const noexcept
  {
    return (m_lDays == d.m_lDays);
  }

  bool operator>(const CDate<TCALENDAR>& d) const noexcept
  {
    if (IsValid() && d.IsValid())
      return m_lDays > d.m_lDays;
    else
      return false; //Only makes sense to compare 2 valid values
  }

  bool operator>=(const CDate<TCALENDAR>& d) const noexcept
  {
    return operator==(d) || operator>(d);
  }

  bool operator<(const CDate<TCALENDAR>& d) const noexcept
  {
    if (IsValid() && d.IsValid())
      return m_lDays < d.m_lDays;
    else
      return false; //Only makes sense to compare 2 valid values
  }

  bool operator<=(const CDate<TCALENDAR>& d) const noexcept
  {
    return operator==(d) || operator<(d);
  }

  bool operator!=(const CDate<TCALENDAR>& d) const noexcept
  {
    return !operator==(d);
  }

//Display
  /* The Following Format parameters are supported
  %a Abbreviated weekday name
  %A Full weekday name
  %b Abbreviated month name
  %B Full month name
  %d Day of month as decimal number (01 - 31)
  %j Day of year as decimal number (001 - 366)
  %m Month as decimal number (01 - 12)
  %U Week of year as decimal number
  %w Weekday as decimal number (1 - 7; Sunday is 1)
  %x Windows Short date representation using GetDateFormat() Win32 API
  %y Year without century, as decimal number (00 - 99)
  %Y Year with century, as decimal number
  %#x Windows Long date representation using GetDateFormat() Win32 API
  %#d, %#j, %#m, %#U, %#y Remove leading zeros (if any)
  */
  DTIME_NODISCARD String Format(const DTimePlus::String& sFormat) const
  {
    //Cache the date values for performance reasons
    const DateS ds{GetDate()};

    //What will be the return value from this function
    String rVal;

    if (IsValid())
    {
      const auto nFmtLength{sFormat.length()};
      for (String::size_type i{0}; i<nFmtLength; i++)
      {
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
        TCHAR c{sFormat[i]};
        if (c == _T('%'))
        {
          ++i;
          if (i < nFmtLength)
          {
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
            c = sFormat[i];
            switch (c)
            {
            #ifdef _WIN32
              case _T('a'):
              {
                rVal += GetAbrStringDayOfWeek();
                break;
              }
              case _T('A'):
              {
                rVal += GetFullStringDayOfWeek();
                break;
              }
              case _T('b'):
              {
                rVal += GetAbrStringMonth();
                break;
              }
              case _T('B'):
              {
                rVal += GetFullStringMonth();
                break;
              }
            #endif //#ifdef _WIN32
              case _T('d'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << ds.Day;
                rVal += ss.str();
                break;
              }
              case _T('j'):
              {
                dt_ostringstream ss;
                ss << std::setw(3) << std::setfill(_T('0')) << DaysSinceEndPreviousYear();
                rVal += ss.str();
                break;
              }
              case _T('m'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << ds.Month;
                rVal += ss.str();
                break;
              }
              case _T('U'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << GetWeekOfYear();
                rVal += ss.str();
                break;
              }
              case _T('w'):
              {
                dt_ostringstream ss;
                ss << static_cast<int>(GetDayOfWeek());
                rVal += ss.str();
                break;
              }
          #ifdef _WIN32
              case _T('x'):
              {
                SYSTEMTIME SystemTime{};
                if (GetAsSystemTime(SystemTime))
#pragma warning(suppress: 26477)
                  rVal += SafeGetDateFormat(LOCALE_NAME_USER_DEFAULT, DATE_SHORTDATE, &SystemTime);
                break;
              }
            #endif //#ifdef _WIN32
              case _T('y'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << Get2DigitYear(ds.Year);
                rVal += ss.str();
                break;
              }
              case _T('Y'):
              {
                dt_ostringstream ss;
                ss << ds.Year;
                rVal += ss.str();
                break;
              }
              case _T('#'):
              {
                if (i < nFmtLength)
                {
                  ++i;
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
                  c = sFormat[i];
                  switch (c)
                  {
                    case _T('d'):
                    {
                      dt_ostringstream ss;
                      ss << ds.Day;
                      rVal += ss.str();
                      break;
                    }
                    case _T('j'):
                    {
                      dt_ostringstream ss;
                      ss << DaysSinceEndPreviousYear();
                      rVal += ss.str();
                      break;
                    }
                    case _T('m'):
                    {
                      dt_ostringstream ss;
                      ss << ds.Month;
                      rVal += ss.str();
                      break;
                    }
                    case _T('U'):
                    {
                      dt_ostringstream ss;
                      ss << GetWeekOfYear();
                      rVal += ss.str();
                      break;
                    } 
                  #ifdef _WIN32
                    case _T('x'):
                    {
                      SYSTEMTIME SystemTime{};
                      if (GetAsSystemTime(SystemTime))
#pragma warning(suppress: 26477)
                        rVal += SafeGetDateFormat(LOCALE_NAME_USER_DEFAULT, DATE_LONGDATE, &SystemTime);
                      break;
                    }
                  #endif //#ifdef _WIN32
                    case _T('y'):
                    {
                      dt_ostringstream ss;
                      ss << Get2DigitYear(ds.Year);
                      rVal += ss.str();
                      break;
                    }
                    default:
                    {
                      rVal += c;
                      break;
                    }
                  }
                }
                break;
              }
              default:
              {
                rVal += c;
                break;
              }
            }
          }
        }
        else
        {
          rVal += c;
        }
      }
    }

    return rVal;
  }

//Serialization
#ifdef _AFX
  void Serialize(CArchive& ar)
  {
    if (ar.IsStoring())
    {
      constexpr const WORD wVersion{0x101};
      ar << wVersion;
      ar << m_lDays;
    }
    else
    {
      WORD wVersion{0};
      ar >> wVersion;
      ASSERT(wVersion == 0x100 || wVersion == 0x101);
      ar >> m_lDays;
      if (wVersion == 0x100)
      {
        bool bOldInGregCalendar{false};
        ar.Read(&bOldInGregCalendar, sizeof(bOldInGregCalendar));
      }
    }
  }
#endif //#ifdef _AFX

#ifdef _AFX
  friend CArchive& operator<<(CArchive& ar, CDate<TCALENDAR>& Date)
  {
    ASSERT(ar.IsStoring());
    Date.Serialize(ar);
    return ar;
  }
#endif //#ifdef _AFX

#ifdef _AFX
  friend CArchive& operator>>(CArchive& ar, CDate<TCALENDAR>& Date)
  {
    ASSERT(!ar.IsStoring());
    Date.Serialize(ar);
    return ar;
  }
#endif //#ifdef _AFX

protected:
//Member variables
  int32_t m_lDays; //Count of days since 15 October 1582 (Gregorian Calendrical System or n.s)
};

} //namespace DTimePlus

#endif //#ifndef __DTIMEPLUSCDATE_H__
