/*
Module : DTime+CLDate.h
Purpose: Defines the interface for the DTime+ CLDate class
Created: PJN / 01-05-2006
History: PJN / 01-01-2018 1. Made CLDate::GetWindowsFormatString method publicly accessible

Copyright (c) 2006 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#if _MSC_VER > 1000
#pragma once
#endif //#if _MSC_VER > 1000

#ifndef __DTIMEPLUSCLDATE_H__
#define __DTIMEPLUSCLDATE_H__


//////////////////// Includes /////////////////////////////////////////////////

#include <cmath>
#include <climits>
#include "DTime+CLTimeOfDay.h"
#include "DTime+CLTimeSpan.h"


//////////////////// Classes //////////////////////////////////////////////////

namespace DTimePlus
{

//A Date class with a granularity of 1 second
template<class TCALENDAR = CGregorianCalendar>
class CLDate
{
public:
//Constructors
#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLDate() noexcept
  {
    Set();
  }

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLDate(int32_t Year, uint16_t month, uint16_t Day, uint16_t Hour, uint16_t Minute, uint16_t Second, bool bDoAssert = true) noexcept
  {
    Set(Year, month, Day, Hour, Minute, Second, bDoAssert);
  }

#ifdef _WIN32
#pragma warning(suppress: 26495)
  CLDate(const SYSTEMTIME& SystemTime, bool bDoAssert = true) noexcept
  {
    Set(SystemTime, bDoAssert);
  }
#endif //#ifdef _WIN32

#ifdef _WIN32
#pragma warning(suppress: 26495)
  CLDate(const FILETIME& FileTime, bool bDoAssert = true) noexcept
  {
    Set(FileTime, bDoAssert);
  }
#endif //#ifdef _WIN32

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLDate(int32_t Year, uint16_t month, uint16_t WeekOfMonth, DayOfWeek dow, uint16_t Hour, uint16_t Minute, uint16_t Second, bool bDoAssert = true) noexcept
  {
    Set(Year, month, WeekOfMonth, dow, Hour, Minute, Second, bDoAssert);
  }

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLDate(int32_t Days, DateEpoch e, uint16_t Hour, uint16_t Minute, uint16_t Second, bool bDoAssert = true) noexcept
  {
    Set(Days, e, Hour, Minute, Second, bDoAssert);
  }

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLDate(double Days, DateEpoch e, bool bDoAssert = true) noexcept
  {
    Set(Days, e, bDoAssert);
  }

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLDate(const CDate<TCALENDAR>& Date, const CLTimeOfDay<>& Tod) noexcept
  {
    Set(Date, Tod);
  }

  CLDate(const CLDate<TCALENDAR>&) = default;
  CLDate(CLDate<TCALENDAR>&&) = default;

  CLDate(const int64_t& nSeconds) noexcept : m_nSeconds{nSeconds}
  {
  }

#ifdef _AFX
#pragma warning(suppress: 26495)
  CLDate(const CTime& ctime, bool bDoAssert = true) noexcept
  {
    Set(ctime, bDoAssert);
  }
#endif //#ifdef _AFX

#ifdef _AFX
#pragma warning(suppress: 26495)
  CLDate(const COleDateTime& oleTime, bool bDoAssert = true) noexcept
  {
    Set(oleTime, bDoAssert);
  }
#endif //#ifdef _AFX

  ~CLDate() = default;

//Set operators
  CLDate<TCALENDAR>& Set() noexcept
  {
    //Create an invalid value
    m_nSeconds = INT64_MIN;
    return *this;
  }

  CLDate<TCALENDAR>& Set(int32_t Year, uint16_t month, uint16_t Day, uint16_t Hour, uint16_t Minute, uint16_t Second, bool bDoAssert = true) noexcept
  {
    const CDate<TCALENDAR> Date{Year, month, Day, bDoAssert};
    const CLTimeOfDay<> Tod{Hour, Minute, Second, bDoAssert};
    if (Date.IsValid() && Tod.IsValid())
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      m_nSeconds = (static_cast<int64_t>(Date.GetValue()) * 86400) + Tod.GetTotalSeconds();
    }
    else
      Set();
    return *this;
  }

#ifdef _WIN32
  CLDate<TCALENDAR>& Set(const SYSTEMTIME& st, bool bDoAssert = true) noexcept
  {
    Set(st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond, bDoAssert);
    return *this;
  }
#endif //#ifdef _WIN32

#ifdef _WIN32
  CLDate<TCALENDAR>& Set(const FILETIME& FileTime, bool bDoAssert = true) noexcept
  {
    SYSTEMTIME SystemTime{};
    if (::FileTimeToSystemTime(&FileTime, &SystemTime))
      Set(SystemTime, bDoAssert);
    else
    {
    #ifdef _DEBUG
      if (bDoAssert)
        DTIME_ASSERT(false);
    #endif //#ifdef _DEBUG
      Set();
    }
    return *this;
  }
#endif //#ifdef _WIN32

  CLDate<TCALENDAR>& Set(int32_t Year, uint16_t month, uint16_t WeekOfMonth, DayOfWeek dow, uint16_t Hour, uint16_t Minute, uint16_t Second, bool bDoAssert = true) noexcept
  {
    const CDate<TCALENDAR> Date{Year, month, WeekOfMonth, dow, bDoAssert};
    const CLTimeOfDay<> Tod{Hour, Minute, Second, bDoAssert};

    if (Date.IsValid() && Tod.IsValid())
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      m_nSeconds = (static_cast<int64_t>(Date.GetValue()) * 86400) + Tod.GetTotalSeconds();
    }
    else
      Set();
    return *this;
  }

  CLDate<TCALENDAR>& Set(int32_t Days, DateEpoch e, uint16_t Hour, uint16_t Minute, uint16_t Second, bool bDoAssert = true) noexcept
  {
    const CDate<TCALENDAR> Date{Days, e, bDoAssert};
    const CLTimeOfDay<> Tod{Hour, Minute, Second, bDoAssert};

    if (Date.IsValid() && Tod.IsValid())
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      m_nSeconds = (static_cast<int64_t>(Date.GetValue()) * 86400) + Tod.GetTotalSeconds();
    }
    else
      Set();
    return *this;
  }

  CLDate<TCALENDAR>& Set(double Days, DateEpoch e, bool bDoAssert = true) noexcept
  {
    CDate<TCALENDAR> Date;
    if (Days < 0)
      Date.Set(static_cast<int32_t>(Days-1), e, bDoAssert);
    else
      Date.Set(static_cast<int32_t>(Days), e, bDoAssert);
    double dInt{0};
    const double dFrac{modf(Days, &dInt)};
    CLTimeOfDay<> Tod;
    if (dFrac < 0)
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26467)
#endif //#ifdef _MSC_VER
      Tod.Set(static_cast<uint32_t>(86400 + (dFrac * 86400)), bDoAssert);
    }
    else
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26467)
#endif //#ifdef _MSC_VER
      Tod.Set(static_cast<uint32_t>(dFrac * 86400), bDoAssert);
    }

    if (Date.IsValid() && Tod.IsValid())
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      m_nSeconds = (static_cast<int64_t>(Date.GetValue()) * 86400) + Tod.GetTotalSeconds();
    }
    else
      Set();
    return *this;
  }

  CLDate<TCALENDAR>& Set(const int64_t& nSeconds) noexcept
  {
    m_nSeconds = nSeconds;
    return *this;
  }

  CLDate<TCALENDAR>& Set(const CDate<TCALENDAR>& Date, const CLTimeOfDay<>& Tod) noexcept
  {
    if (!Date.IsValid() || !Tod.IsValid())
      Set();
    else
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      m_nSeconds = (static_cast<int64_t>(Date.GetValue()) * 86400) + Tod.GetTotalSeconds();
    }
    return *this;
  }

#ifdef _AFX
  CLDate<TCALENDAR>& Set(const CTime& ctime, bool bDoAssert = true) noexcept
  {
    const time_t t{ctime.GetTime()};
#pragma warning(suppress: 26472)
    const auto nDays{static_cast<int32_t>(t / 86400)};
#pragma warning(suppress: 26472)
    const time_t nSecondsLeftOver{t - (static_cast<time_t>(nDays) * 86400)};
#pragma warning(suppress: 26472)
    const auto nHours{static_cast<uint16_t>(nSecondsLeftOver / 3600)};
#pragma warning(suppress: 26472)
    const auto nMinutes{static_cast<uint16_t>((nSecondsLeftOver - (static_cast<time_t>(nHours) * 3600)) / 60)};
#pragma warning(suppress: 26472)
    const auto nSeconds{static_cast<uint16_t>(nSecondsLeftOver - (static_cast<time_t>(nHours) * 3600) - (static_cast<time_t>(nMinutes) * 60))};
    Set(nDays, DateEpoch::EPOCH_CTIME, nHours, nMinutes, nSeconds, bDoAssert);
    return *this;
  }
#endif //#ifdef _AFX

#ifdef _AFX
  CLDate<TCALENDAR>& Set(const COleDateTime& oleTime, bool bDoAssert = true) noexcept
  {
    if (oleTime.GetStatus() == COleDateTime::valid)
    {
      SYSTEMTIME st{};
      if (oleTime.GetAsSystemTime(st))
        Set(st, bDoAssert);
      else
        Set();
    }
    else
      Set();
    return *this;
  }
#endif //#ifdef _AFX

//Static constructors
  static CLDate<TCALENDAR> CurrentLocalTime(bool bDoAssert = true) noexcept
  {
    //Get the current time from the CRT
    time_t now{0};
    if (time(&now) != -1)
    {
      //Convert to local time
#ifdef _MSC_VER
      tm lt{};
      const errno_t e{localtime_s(&lt, &now)};
      if (e != 0)
        return CLDate<TCALENDAR>{};
      const tm* pLocalTime{&lt};
#else
      const tm* pLocalTime{localtime(&now)};
#endif //#ifdef _MSC_VER
      if (pLocalTime == nullptr)
        return CLDate<TCALENDAR>{};
      else
      {
        //PC clock will always return time in Gregorian calendar
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
        const CLDate<CGregorianCalendar> date{static_cast<int32_t>(pLocalTime->tm_year + 1900), static_cast<uint16_t>(pLocalTime->tm_mon + 1), static_cast<uint16_t>(pLocalTime->tm_mday), static_cast<uint16_t>(pLocalTime->tm_hour), static_cast<uint16_t>(pLocalTime->tm_min), static_cast<uint16_t>(pLocalTime->tm_sec), bDoAssert};
        if (date.IsValid())
          return CLDate<TCALENDAR>{date.GetValue()};
        else
          return CLDate<TCALENDAR>{};
      }
    }
    else
      return CLDate<TCALENDAR>();
  }

  static CLDate<TCALENDAR> CurrentSystemTime(bool bDoAssert = true) noexcept
  {
    //Get the current time from the CRT
    time_t now{0};
    if (time(&now) != -1)
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      const auto nDays{static_cast<int32_t>(now / 86400)};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      const time_t nSecondsLeftOver{now - (static_cast<time_t>(nDays) * 86400)};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      const auto nHours{static_cast<uint16_t>(nSecondsLeftOver / 3600)};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      const auto nMinutes{static_cast<uint16_t>((nSecondsLeftOver - (static_cast<time_t>(nHours) * 3600)) / 60)};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      const auto nSeconds{static_cast<uint16_t>(nSecondsLeftOver - (static_cast<time_t>(nHours) * 3600) - (static_cast<time_t>(nMinutes) * 60))};
      //PC clock will always return time in Gregorian calendar
      const CLDate<CGregorianCalendar> date{nDays, DateEpoch::EPOCH_CTIME, nHours, nMinutes, nSeconds, bDoAssert};
      if (date.IsValid())
        return CLDate<TCALENDAR>{date.GetValue()};
      else
        return CLDate<TCALENDAR>{};
    }
    else
      return CLDate<TCALENDAR>{};
  }

//Static operations
#ifdef _WIN32
#pragma warning(suppress: 26477)
  static String GetWindowsFormatString(LPCWSTR lpLocaleName = LOCALE_NAME_USER_DEFAULT)
  {
    return CDate<TCALENDAR>::GetWindowsFormatString(lpLocaleName) + _T(" ") + CLTimeOfDay<>::GetWindowsFormatString(lpLocaleName);
  }
#endif //#ifdef _WIN32

//Operations
  DTIME_NODISCARD CDate<TCALENDAR> GetCDate() const noexcept
  {
    DTIME_ASSERT(IsValid());

    CDate<TCALENDAR> rVal;
    if (m_nSeconds >= 0)
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      rVal.Set(static_cast<int32_t>(m_nSeconds / 86400), DateEpoch::EPOCH_GREG);
    }
    else if (m_nSeconds >= -86400)
      rVal.Set(-1, DateEpoch::EPOCH_GREG);
    else
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      rVal.Set(static_cast<int32_t>((m_nSeconds - 86399) / 86400), DateEpoch::EPOCH_GREG);
    }

    return rVal;
  }

  DTIME_NODISCARD CLTimeOfDay<> GetCLTimeOfDay() const noexcept
  {
    DTIME_ASSERT(IsValid());

    CLTimeOfDay<> rVal;
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    const int64_t nSeconds{m_nSeconds - static_cast<int64_t>(GetCDate().GetValue()) * 86400};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    rVal.Set(static_cast<uint32_t>(nSeconds));
    return rVal;
  }

  DTIME_NODISCARD int64_t GetValue() const noexcept
  {
    DTIME_ASSERT(IsValid());

    return m_nSeconds;
  }

  DTIME_NODISCARD DateLS GetDate() const noexcept
  {
    DTIME_ASSERT(IsValid());

    //What will be the return value from this function
    DateLS rVal;

    const DateS ds{GetCDate().GetDate()};
    rVal.Year = ds.Year;
    rVal.Month = ds.Month;
    rVal.Day = ds.Day;
    const CLTimeOfDay<> tod{GetCLTimeOfDay()};
    rVal.Hour = tod.GetHours();
    rVal.Minute = tod.GetMinutes();
    rVal.Second = tod.GetSeconds();
    return rVal;
  };

#ifdef _WIN32
  bool GetAsSystemTime(SYSTEMTIME& SystemTime) const noexcept
  {
    if (!IsValid())
      return false;

    if (!GetCDate().GetAsSystemTime(SystemTime))
      return false;

    const CLTimeOfDay<> tod{GetCLTimeOfDay()};
    SystemTime.wHour = tod.GetHours();
    SystemTime.wMinute = tod.GetMinutes();
    SystemTime.wSecond = tod.GetSeconds();
    SystemTime.wMilliseconds = 0;
    return true;
  }
#endif //#ifdef _WIN32

#ifdef _WIN32
  bool GetAsFileTime(FILETIME& FileTime) const noexcept
  {
    SYSTEMTIME SystemTime{};
    bool bSuccess = GetAsSystemTime(SystemTime);
    bSuccess = bSuccess && ::SystemTimeToFileTime(&SystemTime, &FileTime);
    return bSuccess;
  }
#endif //#ifdef _WIN32

  bool GetAsTM(tm& time) const noexcept
  {
    if (!IsValid())
     return false;

    //What will be the return value from this function
    bool bSuccess{true};

    const CDate<TCALENDAR> date{GetCDate()};
    const DateLS ds{GetDate()};
    constexpr const int32_t MinYear{1900 + INT32_MIN};
    constexpr const int32_t MaxYear{INT32_MAX};

    //handle range errors
    if ((ds.Year < MinYear) || (ds.Year > MaxYear))
      bSuccess = false;
    else
    {
      time.tm_hour = ds.Hour;
      time.tm_min = ds.Minute;
      time.tm_sec = ds.Second;
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      time.tm_year = static_cast<int>(ds.Year - 1900);
      time.tm_mon = ds.Month - 1; //tm struct uses 0 based indices
      time.tm_wday = static_cast<int>(date.GetDayOfWeek()) - 1; //tm struct uses 0 based indices
      time.tm_mday = ds.Day;
      time.tm_yday = date.DaysSinceFirstOfYear(); //Returns days since Jan 1
      time.tm_isdst = 0; //As you can see, GetTM always return 0 for this value. Client code should not rely on this value!
    }
    return bSuccess;
  }

#ifdef _AFX
  bool GetAsCOleDateTime(COleDateTime& OleDateTime) const noexcept
  {
    SYSTEMTIME st{};
    const bool rVal = GetAsSystemTime(st);
    if (rVal)
      OleDateTime = COleDateTime(st);
    else
      OleDateTime.SetStatus(COleDateTime::invalid);
    return rVal;
  }
#endif //#ifdef _AFX

  DTIME_NODISCARD CLDate<CJulianCalendar> AsJulianCLDate() const noexcept
  {
    CLDate<CJulianCalendar> rVal;
    rVal.Set(m_nSeconds);
    return rVal;
  }

  DTIME_NODISCARD CLDate<CGregorianCalendar> AsGregorianCLDate() const noexcept
  {
    CLDate<CGregorianCalendar> rVal;
    rVal.Set(m_nSeconds);
    return rVal;
  }

  void AddYear(int Years = 1) noexcept
  {
    DTIME_ASSERT(IsValid());

    CDate<TCALENDAR> Date{GetCDate()};
    const CLTimeOfDay<> Tod{GetCLTimeOfDay()};
    Date.AddYear(Years);
    Set(Date, Tod);
  }

  void AddMonth(int Months = 1) noexcept
  {
    DTIME_ASSERT(IsValid());

    CDate<TCALENDAR> Date{GetCDate()};
    const CLTimeOfDay<> Tod{GetCLTimeOfDay()};
    Date.AddMonth(Months);
    Set(Date, Tod);
  }

  void AddWeek(int Weeks = 1) noexcept
  {
    DTIME_ASSERT(IsValid());

    CDate<TCALENDAR> Date{GetCDate()};
    const CLTimeOfDay<> Tod{GetCLTimeOfDay()};
    Date.AddWeek(Weeks);
    Set(Date, Tod);
  }

  DTIME_NODISCARD bool IsValid() const noexcept
  {
    return (m_nSeconds != INT64_MIN);
  }

  CLDate<TCALENDAR>& operator=(const CLDate<TCALENDAR>&) = default;
  CLDate<TCALENDAR>& operator=(CLDate<TCALENDAR>&&) = default;

//Overloaded arithmetic operators
  CLDate<TCALENDAR> operator+(const CLTimeSpan& TimeSpan) const noexcept
  {
    if (IsValid() && TimeSpan.IsValid())
    {
      CLDate<TCALENDAR> rVal;
      rVal.m_nSeconds = m_nSeconds + TimeSpan.GetValue();
      return rVal;
    }
    else 
      return CLDate<TCALENDAR>{};
  }

  CLTimeSpan operator-(const CLDate<TCALENDAR>& ld) const noexcept
  {
    if (IsValid() && ld.IsValid())
      return CLTimeSpan{m_nSeconds - ld.m_nSeconds};
    else
      return {};
  }

  CLDate<TCALENDAR> operator-(const CLTimeSpan& TimeSpan) const noexcept
  {
    if (IsValid() && TimeSpan.IsValid())
    {
      const int64_t s{m_nSeconds - TimeSpan.GetValue()};
      CLDate<TCALENDAR> rVal;
      rVal.m_nSeconds = s;
      return rVal;
    }
    else
      return CLDate<TCALENDAR>{};
  }

  CLDate<TCALENDAR>& operator+=(const CLTimeSpan& TimeSpan) noexcept
  {
    *this = *this + TimeSpan;
    return *this;
  }

  CLDate<TCALENDAR>& operator-=(const CLTimeSpan& TimeSpan) noexcept
  {
    *this = *this - TimeSpan;
    return *this;
  }

  CLDate<TCALENDAR>& operator++() noexcept
  {
    *this = *this + CLTimeSpan{1, 0, 0, 0};
    return *this;
  }

  CLDate<TCALENDAR>& operator--() noexcept
  {
    *this = *this - CLTimeSpan{1, 0, 0, 0};
    return *this;
  }

//Overloaded equality operators
  bool operator==(const CLDate<TCALENDAR>& ld) noexcept
  {
    return (m_nSeconds == ld.m_nSeconds);
  }

  bool operator>(const CLDate<TCALENDAR>& ld) noexcept
  {
    if (IsValid() && ld.IsValid())
      return (m_nSeconds > ld.m_nSeconds);
    else
      return false; //Only makes sense to compare 2 valid values
  }

  bool operator>=(const CLDate<TCALENDAR>& ld) noexcept
  {
    return operator==(ld) || operator>(ld);
  }

  bool operator<(const CLDate<TCALENDAR>& ld) noexcept
  {
    if (IsValid() && ld.IsValid())
      return (m_nSeconds < ld.m_nSeconds);
    else
      return false; //Only makes sense to compare 2 valid values
  }

  bool operator<=(const CLDate<TCALENDAR>& ld) noexcept
  {
    return operator==(ld) || operator<(ld);
  }

  bool operator!=(const CLDate<TCALENDAR>& ld) noexcept
  {
    return !operator==(ld);
  }

//Display
  /* The Following Format parameters are supported
  %a  Abbreviated weekday name
  %A  Full weekday name
  %b  Abbreviated month name
  %B  Full month name
  %d  Day of month as decimal number (01 - 31)
  %j  Day of year as decimal number (001 - 366)
  %m  Month as decimal number (01 - 12)
  %P  AM / PM indicator
  %U  Week of year as decimal number
  %w  Weekday as decimal number (1 - 7; Sunday is 1)
  %x  Windows short date representation, namely Short Date representation + " " + Time Representation using GetDateFormatEx() and GetTimeFormatEx() Win32 API
  %y  Year without century, as decimal number (00 - 99)
  %Y  Year with century, as decimal number
  %H  Hours in the day
  %h  12 Hour format Hours in (00 - 12)
  %M  Minutes in the hour
  %S  Seconds in the minute
  %%  Percent sign
  %#x Long date representation, namely Long Date representation + " " + Time Representation using GetDateFormatEx() and GetTimeFormatEx() Win32 API
  %#d, %#j, %#m, %#U, %#w, %#y, %#H, %#h, %#M, %#S  Remove leading zeros (if any)
  */
  DTIME_NODISCARD String Format(const String& sFormat) const
  {
    //What will be the return value from this function
    String rVal;

    //Cache some values for speed optimization
    const CDate<TCALENDAR> date{GetCDate()};
    const CLTimeOfDay<> tod{GetCLTimeOfDay()};
    const DateLS ds{GetDate()};

    if (IsValid())
    {
      const auto nFmtLength{sFormat.length()};
      for (String::size_type i{0}; i<nFmtLength; i++)
      {
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
        TCHAR c{sFormat[i]};
        if (c == _T('%'))
        {
          ++i;
          if (i < nFmtLength)
          {
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
            c = sFormat[i];
            switch (c)
            {
              case _T('H'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << ds.Hour;
                rVal += ss.str();
                break;
              }
              case _T('h'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << tod.GetAMPMHours();
                rVal += ss.str();
                break;
              }
              case _T('M'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << ds.Minute;
                rVal += ss.str();
                break;
              }
            #ifdef _WIN32
              case _T('P'):
              {
                rVal += tod.GetAMPMString();
                break;
              }
            #endif //#ifdef _WIN32
              case _T('S'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << ds.Second;
                rVal += ss.str();
                break;
              }
            #ifdef _WIN32
              case _T('a'):
              {
                rVal += TCALENDAR::GetAbrStringDayOfWeek(date.GetDayOfWeek());
                break;
              }
              case _T('A'):
              {
                rVal += TCALENDAR::GetFullStringDayOfWeek(date.GetDayOfWeek());
                break;
              }
              case _T('b'):
              {
                rVal += TCALENDAR::GetAbrStringMonth(ds.Month);
                break;
              }
              case _T('B'):
              {
                rVal += TCALENDAR::GetFullStringMonth(ds.Month);
                break;
              }
            #endif //#ifdef _WIN32
              case _T('d'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << ds.Day;
                rVal += ss.str();
                break;
              }
              case _T('j'):
              {
                dt_ostringstream ss;
                ss << std::setw(3) << std::setfill(_T('0')) << date.DaysSinceEndPreviousYear();
                rVal += ss.str();
                break;
              }
              case _T('m'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << ds.Month;
                rVal += ss.str();
                break;
              }
              case _T('U'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << date.GetWeekOfYear();
                rVal += ss.str();
                break;
              }
              case _T('w'):
              {
                dt_ostringstream ss;
                ss << static_cast<int>(date.GetDayOfWeek());
                rVal += ss.str();
                break;
              }
            #ifdef _WIN32
              case _T('x'):
              {
                rVal += date.Format(_T("%x")) + _T(" ") + tod.Format(_T("%x"));
                break;
              }
            #endif //#ifdef _WIN32
              case _T('y'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << CDate<TCALENDAR>::Get2DigitYear(ds.Year);
                rVal += ss.str();
                break;
              }
              case _T('Y'):
              {
                dt_ostringstream ss;
                ss << ds.Year;
                rVal += ss.str();
                break;
              }
              case _T('#'):
              {
                if (i < nFmtLength)
                {
                  ++i;
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
                  c = sFormat[i];
                  switch (c)
                  {
                    case _T('d'):
                    {
                      dt_ostringstream ss;
                      ss << ds.Day;
                      rVal += ss.str();
                      break;
                    }
                    case _T('j'):
                    {
                      dt_ostringstream ss;
                      ss << date.DaysSinceEndPreviousYear();
                      rVal += ss.str();
                      break;
                    }
                    case _T('m'):
                    {
                      dt_ostringstream ss;
                      ss << ds.Month;
                      rVal += ss.str();
                      break;
                    }
                    case _T('U'):
                    {
                      dt_ostringstream ss;
                      ss << date.GetWeekOfYear();
                      rVal += ss.str();
                      break;
                    }
                  #ifdef _WIN32
                    case _T('x'):
                    {
                      rVal += date.Format(_T("%#x")) + _T(" ") + tod.Format(_T("%x"));
                      break;
                    }
                  #endif //#ifdef _WIN32
                    case _T('y'):
                    {
                      dt_ostringstream ss;
                      ss << CDate<TCALENDAR>::Get2DigitYear(ds.Year);
                      rVal += ss.str();
                      break;
                    }
                    case _T('H'):
                    {
                      dt_ostringstream ss;
                      ss << tod.GetHours();
                      rVal += ss.str();
                      break;
                    }
                    case _T('h'):
                    {
                      dt_ostringstream ss;
                      ss << tod.GetAMPMHours();
                      rVal += ss.str();
                      break;
                    }
                    case _T('M'):
                    {
                      dt_ostringstream ss;
                      ss << tod.GetMinutes();
                      rVal += ss.str();
                      break;
                    }
                    case _T('S'):
                    {
                      dt_ostringstream ss;
                      ss << tod.GetSeconds();
                      rVal += ss.str();
                      break;
                    }
                    default:
                    {
                      rVal += c;
                      break;
                    }
                  }
                }
                break;
              }
              default:
              {
                rVal += c;
                break;
              }
            }
          }
        }
        else
        {
          rVal += c;
        }
      }
    }
    return rVal;
  }

//Serialization
#ifdef _AFX
  void Serialize(CArchive& ar)
  {
    if (ar.IsStoring())
    {
      constexpr const WORD wVersion{0x101};
      ar << wVersion;
      ar.Write(&m_nSeconds, sizeof(m_nSeconds));
    }
    else
    {
      WORD wVersion{0};
      ar >> wVersion;
      ASSERT(wVersion == 0x101 || wVersion == 0x100);
      ar.Read(&m_nSeconds, sizeof(m_nSeconds));
      if (wVersion == 0x100)
      {
        bool bOldInGregCalendar{false};
        ar.Read(&bOldInGregCalendar, sizeof(bOldInGregCalendar));
      }
    }
  }
#endif //#ifdef _AFX

#ifdef _AFX
  friend CArchive& operator<<(CArchive& ar, CLDate<TCALENDAR>& ld)
  {
    ASSERT(ar.IsStoring());
    ld.Serialize(ar);
    return ar;
  }
#endif //#ifdef _AFX

#ifdef _AFX
  friend CArchive& operator>>(CArchive& ar, CLDate<TCALENDAR>& ld)
  {
    ASSERT(!ar.IsStoring());
    ld.Serialize(ar);
    return ar;
  }
#endif //#ifdef _AFX

protected:
//Member variables
  int64_t m_nSeconds; //Actual instance data (count of seconds from same epoch as CDate uses)
};

} //namespace DTimePlus

#endif //#ifndef __DTIMEPLUSCLDATE_H__
