/*
Module : DTime+CLLDate.h
Purpose: Defines the interface for the DTime+ CLLDate class
Created: PJN / 01-01-2018

Copyright (c) 2018 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise)
when your product is released in binary form. You are allowed to modify the source code in any way you want
except you cannot modify the copyright details at the top of each module. If you want to distribute source
code with your application, then you are only allowed to distribute versions released by the author. This is
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#if _MSC_VER > 1000
#pragma once
#endif //#if _MSC_VER > 1000

#ifndef __DTIMEPLUSCLLDATE_H__
#define __DTIMEPLUSCLLDATE_H__


//////////////////// Includes /////////////////////////////////////////////////

#include <cmath>
#include <climits>
#include "DTime+CLLTimeOfDay.h"
#include "DTime+CLLTimeSpan.h"
#include "DTime+Helpers.h"


//////////////////// Classes //////////////////////////////////////////////////

namespace DTimePlus
{

//A Date class with a count of seconds stored as a double
template<class TCALENDAR = CGregorianCalendar>
class CLLDate
{
public:
//Constructors
#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLLDate() noexcept
  {
    Set();
  }

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLLDate(int32_t Year, uint16_t month, uint16_t Day, uint16_t Hour, uint16_t Minute, double Second, bool bDoAssert = true) noexcept
  {
    Set(Year, month, Day, Hour, Minute, Second, bDoAssert);
  }

#ifdef _WIN32
#pragma warning(suppress: 26495)
  CLLDate(const SYSTEMTIME& SystemTime, bool bDoAssert = true) noexcept
  {
    Set(SystemTime, bDoAssert);
  }
#endif //#ifdef _WIN32

#ifdef _WIN32
#pragma warning(suppress: 26495)
  CLLDate(const FILETIME& FileTime, bool bDoAssert = true) noexcept
  {
    Set(FileTime, bDoAssert);
  }
#endif //#ifdef _WIN32

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLLDate(int32_t Year, uint16_t month, uint16_t WeekOfMonth, DayOfWeek dow, uint16_t Hour, uint16_t Minute, double Second, bool bDoAssert = true) noexcept
  {
    Set(Year, month, WeekOfMonth, dow, Hour, Minute, Second, bDoAssert);
  }

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLLDate(int32_t Days, DateEpoch e, uint16_t Hour, uint16_t Minute, double Second, bool bDoAssert = true) noexcept
  {
    Set(Days, e, Hour, Minute, Second, bDoAssert);
  }

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLLDate(double Days, DateEpoch e, bool bDoAssert = true) noexcept
  {
    Set(Days, e, bDoAssert);
  }

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLLDate(const CDate<TCALENDAR>& Date, const CLLTimeOfDay<>& Tod) noexcept
  {
    Set(Date, Tod);
  }

  CLLDate(const CLLDate<TCALENDAR>&) = default;
  CLLDate(CLLDate<TCALENDAR>&&) = default;

  CLLDate(const double& Seconds) noexcept : m_fSeconds{Seconds}
  {
  }

#ifdef _AFX
#pragma warning(suppress: 26495)
  CLLDate(const CTime& ctime, bool bDoAssert = true) noexcept
  {
    Set(ctime, bDoAssert);
  }
#endif //#ifdef _AFX

#ifdef _AFX
#pragma warning(suppress: 26495)
  CLLDate(const COleDateTime& oleTime, bool bDoAssert = true) noexcept
  {
    Set(oleTime, bDoAssert);
  }
#endif //#ifdef _AFX

  ~CLLDate() = default;

//Set operators
  CLLDate<TCALENDAR>& Set() noexcept
  {
    //Create an invalid value
#ifdef _MSC_VER
#pragma warning(suppress: 26451)
#endif //#ifdef _MSC_VER
    m_fSeconds = NAN;
    return *this;
  }

  CLLDate<TCALENDAR>& Set(int32_t Year, uint16_t month, uint16_t Day, uint16_t Hour, uint16_t Minute, double Second, bool bDoAssert = true) noexcept
  {
    const CDate<TCALENDAR> Date{Year, month, Day, bDoAssert};
    const CLLTimeOfDay<> Tod{Hour, Minute, Second, bDoAssert};
    if (Date.IsValid() && Tod.IsValid())
      m_fSeconds = (Date.GetValue() * 86400.0) + Tod.GetTotalSeconds();
    else
      Set();
    return *this;
  }

#ifdef _WIN32
  CLLDate<TCALENDAR>& Set(const SYSTEMTIME& st, bool bDoAssert = true) noexcept
  {
    Set(st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond + (st.wMilliseconds/1000.0), bDoAssert);
    return *this;
  }
#endif //#ifdef _WIN32

#ifdef _WIN32
  CLLDate<TCALENDAR>& Set(const FILETIME& FileTime, bool bDoAssert = true) noexcept
  {
    SYSTEMTIME SystemTime{};
    if (::FileTimeToSystemTime(&FileTime, &SystemTime))
      Set(SystemTime, bDoAssert);
    else
    {
    #ifdef _DEBUG
      if (bDoAssert)
        DTIME_ASSERT(false);
    #endif //#ifdef _DEBUG
      Set();
    }
    return *this;
  }
#endif //#ifdef _WIN32

  CLLDate<TCALENDAR>& Set(int32_t Year, uint16_t month, uint16_t WeekOfMonth, DayOfWeek dow, uint16_t Hour, uint16_t Minute, double Second, bool bDoAssert = true) noexcept
  {
    const CDate<TCALENDAR> Date{Year, month, WeekOfMonth, dow, bDoAssert};
    const CLLTimeOfDay<> Tod{Hour, Minute, Second, bDoAssert};
    if (Date.IsValid() && Tod.IsValid())
      m_fSeconds = (Date.GetValue() * 86400.0) + Tod.GetTotalSeconds();
    else
      Set();
    return *this;
  }

  CLLDate<TCALENDAR>& Set(int32_t Days, DateEpoch e, uint16_t Hour, uint16_t Minute, double Second, bool bDoAssert = true) noexcept
  {
    const CDate<TCALENDAR> Date{Days, e, bDoAssert};
    const CLLTimeOfDay<> Tod{Hour, Minute, Second, bDoAssert};
    if (Date.IsValid() && Tod.IsValid())
      m_fSeconds = (Date.GetValue() * 86400.0) + Tod.GetTotalSeconds();
    else
      Set();
    return *this;
  }

  CLLDate<TCALENDAR>& Set(double Days, DateEpoch e, bool bDoAssert = true) noexcept
  {
    CDate<TCALENDAR> Date;
    if (Days < 0)
      Date.Set(static_cast<int32_t>(Days-1), e, bDoAssert);
    else
      Date.Set(static_cast<int32_t>(Days), e, bDoAssert);
    double dInt{0};
    const double dFrac{modf(Days, &dInt)};
    CLLTimeOfDay<> Tod;
    if (dFrac < 0)
      Tod.Set(86400 + (dFrac * 86400), bDoAssert);
    else
      Tod.Set(dFrac * 86400, bDoAssert);
    if (Date.IsValid() && Tod.IsValid())
      m_fSeconds = (Date.GetValue() * 86400.0) + Tod.GetTotalSeconds();
    else
      Set();
    return *this;
  }

  CLLDate<TCALENDAR>& Set(const double& Seconds) noexcept
  {
    m_fSeconds = Seconds;
    return *this;
  }

  CLLDate<TCALENDAR>& Set(const CDate<TCALENDAR>& Date, const CLLTimeOfDay<>& Tod) noexcept
  {
    if (!Date.IsValid() || !Tod.IsValid())
      Set();
    else
      m_fSeconds = (Date.GetValue() * 86400.0) + Tod.GetTotalSeconds();
    return *this;
  }

#ifdef _AFX
  CLLDate<TCALENDAR>& Set(const CTime& ctime, bool bDoAssert = true) noexcept
  {
    const time_t t{ctime.GetTime()};
#pragma warning(suppress: 26472)
    const auto nDays{static_cast<int32_t>(t / 86400)};
#pragma warning(suppress: 26472)
    const time_t nSecondsLeftOver{t - (static_cast<time_t>(nDays) * 86400)};
#pragma warning(suppress: 26472)
    const auto nHours{static_cast<uint16_t>(nSecondsLeftOver / 3600)};
#pragma warning(suppress: 26472)
    const auto nMinutes{static_cast<uint16_t>((nSecondsLeftOver - (static_cast<time_t>(nHours) * 3600)) / 60)};
    const double fSeconds{(nSecondsLeftOver - (nHours * 3600.0) - (nMinutes * 60.0))};
    Set(nDays, DateEpoch::EPOCH_CTIME, nHours, nMinutes, fSeconds, bDoAssert);
    return *this;
  }
#endif //#ifdef _AFX

#ifdef _AFX
  CLLDate<TCALENDAR>& Set(const COleDateTime& oleTime, bool bDoAssert = true) noexcept
  {
    if (oleTime.GetStatus() == COleDateTime::valid)
    {
      SYSTEMTIME st{};
      if (oleTime.GetAsSystemTime(st))
        Set(st, bDoAssert);
      else
        Set();
    }
    else
      Set();
    return *this;
  }
#endif //#ifdef _AFX

//Static constructors
  static CLLDate<TCALENDAR> CurrentLocalTime(bool bDoAssert = true) noexcept
  {
  #if defined(_WIN32)
    //Get the current time from the Win32 API
    SYSTEMTIME st;
    GetLocalTime(&st);
    //PC clock will always return time in Gregorian calendar
    const CLLDate<CGregorianCalendar> date{st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond + (st.wMilliseconds / 1000.0), bDoAssert};
  #elif defined(__DTIMEPLUS_HAS_SYS_TIME)
    //Get the current time from the gettimeofday API
    timeval now;
    if (gettimeofday(&now, 0))
      return CLLDate<TCALENDAR>{};
    tm* pLocalTime{localtime(&now.tv_sec)};
    if (pLocalTime == nullptr)
      return CLLDate<TCALENDAR>{};
    //PC clock will always return time in Gregorian calendar
    CLLDate<CGregorianCalendar> date{static_cast<int32_t>(pLocalTime->tm_year + 1900), static_cast<uint16_t>(pLocalTime->tm_mon + 1), static_cast<uint16_t>(pLocalTime->tm_mday), static_cast<uint16_t>(pLocalTime->tm_hour), static_cast<uint16_t>(pLocalTime->tm_min), pLocalTime->tm_sec + (now.tv_usec / 1000000.0), bDoAssert};
  #else
    //Get the current time from the CRT
    time_t now{0};
    if (time(&now) == -1)
      return CLLDate<TCALENDAR>{};
    //Convert to local time
    tm* pLocalTime{localtime(&now)};
    if (pLocalTime == nullptr)
      return CLLDate<TCALENDAR>{};
    //PC clock will always return time in Gregorian calendar
    CLLDate<CGregorianCalendar> date{static_cast<int32_t>(pLocalTime->tm_year + 1900), static_cast<uint16_t>(pLocalTime->tm_mon + 1), static_cast<uint16_t>(pLocalTime->tm_mday), static_cast<uint16_t>(pLocalTime->tm_hour), static_cast<uint16_t>(pLocalTime->tm_min), static_cast<double>(pLocalTime->tm_sec), bDoAssert};
  #endif //#ifdef _WIN32
    if (date.IsValid())
      return CLLDate<TCALENDAR>{date.GetValue()};
    else
      return CLLDate<TCALENDAR>{};
  }

  static CLLDate<TCALENDAR> CurrentSystemTime(bool bDoAssert = true) noexcept
  {
  #if defined(_WIN32)
    //Get the current time from the Win32 API
    SYSTEMTIME st;
    GetSystemTime(&st);
    //PC clock will always return time in Gregorian calendar
    const CLLDate<CGregorianCalendar> date{st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond + (st.wMilliseconds / 1000.0), bDoAssert};
  #elif defined(__DTIMEPLUS_HAS_SYS_TIME)
    //Get the current time from the gettimeofday API
    timeval now;
    if (gettimeofday(&now, 0))
      return CLLDate<TCALENDAR>();
    auto nDays{static_cast<int>(now.tv_sec / 86400)};
    int nSecondsLeftOver{now.tv_sec - (nDays * 86400)};
    auto nHours{static_cast<uint16_t>(nSecondsLeftOver / 3600)};
    auto nMinutes{static_cast<uint16_t>((nSecondsLeftOver - (nHours * 3600)) / 60)};
    auto nSeconds{static_cast<uint16_t>(nSecondsLeftOver - (nHours * 3600) - (nMinutes * 60))};
    //PC clock will always return time in Gregorian calendar
    CLLDate<CGregorianCalendar> date{nDays, DateEpoch::EPOCH_CTIME, nHours, nMinutes, nSeconds + now.tv_usec / 1000000.0, bDoAssert};
  #else
    //Get the current time from the CRT
    time_t now{0};
    if (time(&now) == -1)
      return CLLDate<TCALENDAR>();
    auto nDays{static_cast<int32_t>(now / 86400)};
    time_t nSecondsLeftOver{now - (nDays * 86400)};
    auto nHours{static_cast<uint16_t>(nSecondsLeftOver / 3600)};
    auto nMinutes{static_cast<uint16_t>((nSecondsLeftOver - (nHours * 3600)) / 60)};
    double fSeconds{nSecondsLeftOver - (nHours * 3600.0) - (nMinutes * 60.0)};
    //PC clock will always return time in Gregorian calendar
    CLLDate<CGregorianCalendar> date{nDays, DateEpoch::EPOCH_CTIME, nHours, nMinutes, fSeconds, bDoAssert};
  #endif //#ifdef _WIN32
    if (date.IsValid())
      return CLLDate<TCALENDAR>{date.GetValue()};
    else
      return CLLDate<TCALENDAR>{};
  }

//Static operations
#ifdef _WIN32
#pragma warning(suppress: 26477)
  static String GetWindowsFormatString(LPCWSTR lpLocaleName = LOCALE_NAME_USER_DEFAULT)
  {
    return CDate<TCALENDAR>::GetWindowsFormatString(lpLocaleName) + _T(" ") + CLTimeOfDay<>::GetWindowsFormatString(lpLocaleName);
  }
#endif //#ifdef _WIN32

//Operations
#ifdef _WIN32
#pragma warning(suppress: 26440)
  DTIME_NODISCARD CDate<TCALENDAR> GetCDate() const
#endif //#ifdef _WIN32
  {
    DTIME_ASSERT(IsValid());

    CDate<TCALENDAR> rVal;
    if (m_fSeconds >= 0)
      rVal.Set(static_cast<int32_t>(m_fSeconds / 86400), DateEpoch::EPOCH_GREG);
    else
      rVal.Set(static_cast<int32_t>((nextafter(m_fSeconds, 0) / 86400) - 1), DateEpoch::EPOCH_GREG);
    return rVal;
  }

  DTIME_NODISCARD CLLTimeOfDay<> GetCLLTimeOfDay() const
  {
    DTIME_ASSERT(IsValid());

    CLLTimeOfDay<> rVal;
    const double fSeconds{m_fSeconds - (GetCDate().GetValue() * 86400.0)};
    rVal.Set(fSeconds);
    return rVal;
  }

  DTIME_NODISCARD double GetValue() const noexcept
  {
    return m_fSeconds;
  }

  DTIME_NODISCARD DateLLS GetDate() const
  {
    DTIME_ASSERT(IsValid());

    //What will be the return value from this function
    DateLLS rVal;

    const DateS ds{GetCDate().GetDate()};
    rVal.Year = ds.Year;
    rVal.Month = ds.Month;
    rVal.Day = ds.Day;
    const CLLTimeOfDay<> tod{GetCLLTimeOfDay()};
    rVal.Hour = tod.GetHours();
    rVal.Minute = tod.GetMinutes();
    rVal.Second = tod.GetSeconds();
    return rVal;
  };

#ifdef _WIN32
  bool GetAsSystemTime(SYSTEMTIME& SystemTime) const
  {
    if (!IsValid())
      return false;
    if (!GetCDate().GetAsSystemTime(SystemTime))
      return false;
    const CLLTimeOfDay<> tod{GetCLLTimeOfDay()};
    SystemTime.wHour = tod.GetHours();
    SystemTime.wMinute = tod.GetMinutes();
    const double fSeconds{tod.GetSeconds()};
#ifdef _MSC_VER
#pragma warning(suppress: 26467)
#endif //#ifdef _MSC_VER
    SystemTime.wSecond = static_cast<WORD>(fSeconds);
    double dInt{0};
#ifdef _MSC_VER
#pragma warning(suppress: 26467)
#endif //#ifdef _MSC_VER
    SystemTime.wMilliseconds = static_cast<WORD>((modf(fSeconds, &dInt) * 1000) + 0.5);
    return true;
  }
#endif //#ifdef _WIN32

#ifdef _WIN32
  bool GetAsFileTime(FILETIME& FileTime) const
  {
    SYSTEMTIME SystemTime{};
    bool bSuccess = GetAsSystemTime(SystemTime);
    bSuccess = bSuccess && ::SystemTimeToFileTime(&SystemTime, &FileTime);
    return bSuccess;
  }
#endif //#ifdef _WIN32

#ifdef _AFX
  bool GetAsCOleDateTime(COleDateTime& OleDateTime) const
  {
    SYSTEMTIME st{};
    const bool rVal = GetAsSystemTime(st);
    if (rVal)
      OleDateTime = COleDateTime(st);
    else
      OleDateTime.SetStatus(COleDateTime::invalid);
    return rVal;
  }
#endif //#ifdef _AFX

  DTIME_NODISCARD CLLDate<CJulianCalendar> AsJulianCLLDate() const noexcept
  {
    CLLDate<CJulianCalendar> rVal;
    rVal.Set(m_fSeconds);
    return rVal;
  }

  DTIME_NODISCARD CLLDate<CGregorianCalendar> AsGregorianCLLDate() const noexcept
  {
    CLLDate<CGregorianCalendar> rVal;
    rVal.Set(m_fSeconds);
    return rVal;
  }

  void AddYear(int Years = 1)
  {
    DTIME_ASSERT(IsValid());

    CDate<TCALENDAR> Date{GetCDate()};
    const CLLTimeOfDay<> Tod{GetCLLTimeOfDay()};
    Date.AddYear(Years);
    Set(Date, Tod);
  }

  void AddMonth(int Months = 1)
  {
    DTIME_ASSERT(IsValid());

    CDate<TCALENDAR> Date{GetCDate()};
    const CLLTimeOfDay<> Tod{GetCLLTimeOfDay()};
    Date.AddMonth(Months);
    Set(Date, Tod);
  }

  void AddWeek(int Weeks = 1)
  {
    DTIME_ASSERT(IsValid());

    CDate<TCALENDAR> Date{GetCDate()};
    const CLLTimeOfDay<> Tod{GetCLLTimeOfDay()};
    Date.AddWeek(Weeks);
    Set(Date, Tod);
  }

  DTIME_NODISCARD bool IsValid() const noexcept
  {
    return !std::isnan(m_fSeconds);
  }

  CLLDate<TCALENDAR>& operator=(const CLLDate<TCALENDAR>&) = default;
  CLLDate<TCALENDAR>& operator=(CLLDate<TCALENDAR>&&) = default;

//Overloaded arithmetic operators
  CLLDate<TCALENDAR> operator+(const CLLTimeSpan& TimeSpan) const noexcept
  {
    if (IsValid() && TimeSpan.IsValid())
    {
      CLLDate<TCALENDAR> rVal;
      rVal.m_fSeconds = m_fSeconds + TimeSpan.GetValue();
      return rVal;
    }
    else
      return CLLDate<TCALENDAR>{};
  }

  CLLTimeSpan operator-(CLLDate<TCALENDAR>& ld) const noexcept
  {
    if (IsValid() && ld.IsValid())
      return CLLTimeSpan{m_fSeconds - ld.m_fSeconds};
    else
      return {};
  }

  CLLDate<TCALENDAR> operator-(const CLLTimeSpan& TimeSpan) const noexcept
  {
    if (IsValid() && TimeSpan.IsValid())
    {
      const double s{m_fSeconds - TimeSpan.GetValue()};
      CLLDate<TCALENDAR> rVal;
      rVal.m_fSeconds = s;
      return rVal;
    }
    else
      return CLLDate<TCALENDAR>{};
  }

  CLLDate<TCALENDAR>& operator+=(const CLLTimeSpan& TimeSpan) noexcept
  {
    *this = *this + TimeSpan;
    return *this;
  }

  CLLDate<TCALENDAR>& operator-=(const CLLTimeSpan& TimeSpan) noexcept
  {
    *this = *this - TimeSpan;
    return *this;
  }

  CLLDate<TCALENDAR>& operator++() noexcept
  {
    *this = *this + CLLTimeSpan{1, 0, 0, 0};
    return *this;
  }

  CLLDate<TCALENDAR>& operator--() noexcept
  {
    *this = *this - CLLTimeSpan{1, 0, 0, 0};
    return *this;
  }

//Overloaded equality operators
  bool operator==(const CLLDate<TCALENDAR>& ld) noexcept
  {
    return (m_fSeconds == ld.m_fSeconds);
  }

  bool operator>(const CLLDate<TCALENDAR>& ld) noexcept
  {
    if (IsValid() && ld.IsValid())
      return (m_fSeconds > ld.m_fSeconds);
    else
      return false; //Only makes sense to compare 2 valid values
  }

  bool operator>=(const CLLDate<TCALENDAR>& ld) noexcept
  {
    return operator==(ld) || operator>(ld);
  }

  bool operator<(const CLLDate<TCALENDAR>& ld) noexcept
  {
    if (IsValid() && ld.IsValid())
      return (m_fSeconds < ld.m_fSeconds);
    else
      return false; //Only makes sense to compare 2 valid values
  }

  bool operator<=(const CLLDate<TCALENDAR>& ld) noexcept
  {
    return operator==(ld) || operator<(ld);
  }

  bool operator!=(const CLLDate<TCALENDAR>& ld) noexcept
  {
    return !operator==(ld);
  }

//Display
  /* The Following Format parameters are supported
  %a  Abbreviated weekday name
  %A  Full weekday name
  %b  Abbreviated month name
  %B  Full month name
  %d  Day of month as decimal number (01 - 31)
  %j  Day of year as decimal number (001 - 366)
  %m  Month as decimal number (01 - 12)
  %P  AM / PM indicator
  %U  Week of year as decimal number
  %w  Weekday as decimal number (1 - 7; Sunday is 1)
  %x  Windows short date representation, namely Short Date representation + " " + Time Representation using GetDateFormatEx() and GetTimeFormatEx() Win32 API
  %y  Year without century, as decimal number (00 - 99)
  %Y  Year with century, as decimal number
  %H  Hours in the day
  %h  12 Hour format Hours in (00 - 12)
  %M  Minutes in the hour
  %S  Seconds in the minute
  %%  Percent sign
  %#x Long date representation, namely Long Date representation + " " + Time Representation using GetDateFormatEx() and GetTimeFormatEx() Win32 API
  %#d, %#j, %#m, %#U, %#w, %#y, %#H, %#h, %#M, %#S  Remove leading zeros (if any)
  */
  DTIME_NODISCARD String Format(const String& sFormat) const
  {
    //What will be the return value from this function
    String rVal;

    //Cache some values for speed optimization
    const CDate<TCALENDAR> date{GetCDate()};
    const CLLTimeOfDay<> tod{GetCLLTimeOfDay()};
    const DateLLS ds{GetDate()};

    if (IsValid())
    {
      const auto nFmtLength{sFormat.length()};
      for (String::size_type i{0}; i<nFmtLength; i++)
      {
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
        TCHAR c{sFormat[i]};
        if (c == _T('%'))
        {
          ++i;
          if (i < nFmtLength)
          {
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
            c = sFormat[i];
            switch (c)
            {
              case _T('H'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << ds.Hour;
                rVal += ss.str();
                break;
              }
              case _T('h'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << tod.GetAMPMHours();
                rVal += ss.str();
                break;
              }
              case _T('M'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << ds.Minute;
                rVal += ss.str();
                break;
              }
            #ifdef _WIN32
              case _T('P'):
              {
                rVal += tod.GetAMPMString();
                break;
              }
            #endif //#ifdef _WIN32
              case _T('S'):
              {
                dt_ostringstream ss;
                const double fSeconds{ds.Second};
                if (fSeconds < 10)
                  ss << _T('0');
                ss << fSeconds;
                rVal += ss.str();
                break;
              }
            #ifdef _WIN32
              case _T('a'):
              {
                rVal += TCALENDAR::GetAbrStringDayOfWeek(date.GetDayOfWeek());
                break;
              }
              case _T('A'):
              {
                rVal += TCALENDAR::GetFullStringDayOfWeek(date.GetDayOfWeek());
                break;
              }
              case _T('b'):
              {
                rVal += TCALENDAR::GetAbrStringMonth(ds.Month);
                break;
              }
              case _T('B'):
              {
                rVal += TCALENDAR::GetFullStringMonth(ds.Month);
                break;
              }
            #endif //#ifdef _WIN32
              case _T('d'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << ds.Day;
                rVal += ss.str();
                break;
              }
              case _T('j'):
              {
                dt_ostringstream ss;
                ss << std::setw(3) << std::setfill(_T('0')) << date.DaysSinceEndPreviousYear();
                rVal += ss.str();
                break;
              }
              case _T('m'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << ds.Month;
                rVal += ss.str();
                break;
              }
              case _T('U'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << date.GetWeekOfYear();
                rVal += ss.str();
                break;
              }
              case _T('w'):
              {
                dt_ostringstream ss;
                ss << static_cast<int>(date.GetDayOfWeek());
                rVal += ss.str();
                break;
              }
            #ifdef _WIN32
              case _T('x'):
              {
                rVal += date.Format(_T("%x")) + _T(" ") + tod.Format(_T("%x"));
                break;
              }
            #endif //#ifdef _WIN32
              case _T('y'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << CDate<TCALENDAR>::Get2DigitYear(ds.Year);
                rVal += ss.str();
                break;
              }
              case _T('Y'):
              {
                dt_ostringstream ss;
                ss << ds.Year;
                rVal += ss.str();
                break;
              }
              case _T('#'):
              {
                if (i < nFmtLength)
                {
                  ++i;
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
                  c = sFormat[i];
                  switch (c)
                  {
                    case _T('d'):
                    {
                      dt_ostringstream ss;
                      ss << ds.Day;
                      rVal += ss.str();
                      break;
                    }
                    case _T('j'):
                    {
                      dt_ostringstream ss;
                      ss << date.DaysSinceEndPreviousYear();
                      rVal += ss.str();
                      break;
                    }
                    case _T('m'):
                    {
                      dt_ostringstream ss;
                      ss << ds.Month;
                      rVal += ss.str();
                      break;
                    }
                    case _T('U'):
                    {
                      dt_ostringstream ss;
                      ss << date.GetWeekOfYear();
                      rVal += ss.str();
                      break;
                    }
                  #ifdef _WIN32
                    case _T('x'):
                    {
                      rVal += date.Format(_T("%#x")) + _T(" ") + tod.Format(_T("%x"));
                      break;
                    }
                  #endif //#ifdef _WIN32
                    case _T('y'):
                    {
                      dt_ostringstream ss;
                      ss << CDate<TCALENDAR>::Get2DigitYear(ds.Year);
                      rVal += ss.str();
                      break;
                    }
                    case _T('H'):
                    {
                      dt_ostringstream ss;
                      ss << tod.GetHours();
                      rVal += ss.str();
                      break;
                    }
                    case _T('h'):
                    {
                      dt_ostringstream ss;
                      ss << tod.GetAMPMHours();
                      rVal += ss.str();
                      break;
                    }
                    case _T('M'):
                    {
                      dt_ostringstream ss;
                      ss << tod.GetMinutes();
                      rVal += ss.str();
                      break;
                    }
                    case _T('S'):
                    {
                      dt_ostringstream ss;
                      ss << tod.GetSeconds();
                      rVal += ss.str();
                      break;
                    }
                    default:
                    {
                      rVal += c;
                      break;
                    }
                  }
                }
                break;
              }
              default:
              {
                rVal += c;
                break;
              }
            }
          }
        }
        else
        {
          rVal += c;
        }
      }
    }
    return rVal;
  }

//Serialization
#ifdef _AFX
  void Serialize(CArchive& ar)
  {
    if (ar.IsStoring())
    {
      constexpr const WORD wVersion{0x100};
      ar << wVersion;
      ar.Write(&m_fSeconds, sizeof(m_fSeconds));
    }
    else
    {
      WORD wVersion{0};
      ar >> wVersion;
      ASSERT(wVersion == 0x100);
      ar.Read(&m_fSeconds, sizeof(m_fSeconds));
    }
  }
#endif //#ifdef _AFX

#ifdef _AFX
  friend CArchive& operator<<(CArchive& ar, CLLDate<TCALENDAR>& ld)
  {
    ASSERT(ar.IsStoring());
    ld.Serialize(ar);
    return ar;
  }
#endif //#ifdef _AFX

#ifdef _AFX
  friend CArchive& operator>>(CArchive& ar, CLLDate<TCALENDAR>& ld)
  {
    ASSERT(!ar.IsStoring());
    ld.Serialize(ar);
    return ar;
  }
#endif //#ifdef _AFX

protected:
//Member variables
  double m_fSeconds; //Actual instance data (count of seconds from same epoch as CDate uses)
};

} //namespace DTimePlus

#endif //#ifndef __DTIMEPLUSCLLDATE_H__
