/*
Module : DTime+CLLTimeOfDay.h
Purpose: Defines the interface for the DTime+ CLLTimeOfDay class
Created: PJN / 01-01-2018

Copyright (c) 2018 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#if _MSC_VER > 1000
#pragma once
#endif //#if _MSC_VER > 1000

#ifndef __DTIMEPLUSCLLTIMEOFDAY_H__
#define __DTIMEPLUSCLLTIMEOFDAY_H__


//////////////////// Includes /////////////////////////////////////////////////

#include <cstdint>
#include <cmath>
#include <ctime>
#include <sstream>
#include <iomanip>
#include "DTime+Defines.h"
#include "DTime+Helpers.h"
#include "DTime+CLTimeOfDay.h"


//////////////////// Classes //////////////////////////////////////////////////

namespace DTimePlus
{

//Forward declaration
class CLLTimeSpan;

//A Time of Day class which is used in conjunction with the time class CDate in the CLLDate class
template<class TS = DTimePlus::CLLTimeSpan>
class CLLTimeOfDay
{
public:
//Constructors
#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLLTimeOfDay() noexcept
  {
    Set();
  }

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLLTimeOfDay(uint16_t Hour, uint16_t Minute, double Second, bool bDoAssert = true) noexcept
  {
    Set(Hour, Minute, Second, bDoAssert);
  }

#ifdef _WIN32
#pragma warning(suppress: 26495)
  CLLTimeOfDay(const SYSTEMTIME& SystemTime, bool bDoAssert = true) noexcept
  {
    Set(SystemTime, bDoAssert);
  }
#endif //#ifdef _WIN32

  CLLTimeOfDay(const CLLTimeOfDay&) = default;
  CLLTimeOfDay(CLLTimeOfDay&&) = default;

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLLTimeOfDay(double TotalSeconds, bool bDoAssert = true) noexcept
  {
    Set(TotalSeconds, bDoAssert);
  }

  ~CLLTimeOfDay() = default;

//Set operators
  CLLTimeOfDay& Set() noexcept
  {
    //Create an invalid value
#ifdef _MSC_VER
#pragma warning(suppress: 26451)
#endif //#ifdef _MSC_VER
    m_fTotalSeconds = NAN;
    return *this;
  }

#ifdef _DEBUG
  CLLTimeOfDay& Set(uint16_t Hour, uint16_t Minute, double Second, bool bDoAssert = true) noexcept
#else
  CLLTimeOfDay& Set(uint16_t Hour, uint16_t Minute, double Second, bool /*bDoAssert*/ = true) noexcept
#endif
  {
    m_fTotalSeconds = (Hour * 3600.0) + (Minute * 60.0) + Second;
    if (m_fTotalSeconds > 86400) //Note we allow 23:59:60 to cater for leap seconds
    {
    #ifdef _DEBUG
      if (bDoAssert)
        DTIME_ASSERT(false);
    #endif //#ifdef _DEBUG
      Set();
      return *this;
    }
    return *this;
  }

#ifdef _WIN32
  CLLTimeOfDay& Set(const SYSTEMTIME& SystemTime, bool bDoAssert = true) noexcept
  {
    return Set(SystemTime.wHour, SystemTime.wMinute, SystemTime.wSecond + (SystemTime.wMilliseconds / 1000.0), bDoAssert);
  }
#endif //#ifdef _WIN32

#ifdef _DEBUG
  CLLTimeOfDay& Set(double TotalSeconds, bool bDoAssert = true) noexcept
#else
  CLLTimeOfDay& Set(double TotalSeconds, bool /*bDoAssert*/ = true) noexcept
#endif //#ifdef _DEBUG
  {
    m_fTotalSeconds = TotalSeconds;
    if ((m_fTotalSeconds > 86400) || (m_fTotalSeconds < 0)) //Note we allow 23:59:60 to cater for leap seconds
    {
    #ifdef _DEBUG
      if (bDoAssert)
        DTIME_ASSERT(false);
    #endif //#ifdef _DEBUG
      Set();
      return *this;
    }
    return *this;
  }

//Static constructors
  static CLLTimeOfDay CurrentSystemTime(bool bDoAssert = true) noexcept
  {
  #if defined(_WIN32)
    //Get the current time from the Win32 API
    SYSTEMTIME st;
    GetSystemTime(&st);
    return CLLTimeOfDay<TS>{st.wHour, st.wMinute, st.wSecond + (st.wMilliseconds / 1000.0), bDoAssert};
  #elif defined(__DTIMEPLUS_HAS_SYS_TIME)
    //Get the current time from the gettimeofday API
    timeval now;
    if (gettimeofday(&now, 0))
      return CLLTimeOfDay<TS>{};
    auto nDays{static_cast<int>(now.tv_sec / 86400)};
    int nSecondsLeftOver{now.tv_sec - (nDays * 86400)};
    return CLLTimeOfDay<TS>{nSecondsLeftOver + (now.tv_usec / 1000000.0), bDoAssert};
  #else
    //Get the current time from the CRT (by delegating to CLTimeOfDay)
    CLTimeOfDay<TS> tod{CLTimeOfDay<TS>::CurrentSystemTime(bDoAssert)};
    return CLLTimeOfDay<TS>{tod.GetTotalSeconds(), bDoAssert};
  #endif //#ifdef _WIN32
  }

  static CLLTimeOfDay CurrentLocalTime(bool bDoAssert = true) noexcept
  {
  #if defined(_WIN32)
    //Get the current time from the Win32 API
    SYSTEMTIME st;
    GetLocalTime(&st);
    return CLLTimeOfDay<TS>{st.wHour, st.wMinute, st.wSecond + (st.wMilliseconds / 1000.0), bDoAssert};
  #elif defined(__DTIMEPLUS_HAS_SYS_TIME)
    //Get the current time from the gettimeofday API
    timeval now;
    if (gettimeofday(&now, 0))
      return CLLTimeOfDay<TS>{};
    tm* pLocalTime = localtime(&now.tv_sec);
    if (pLocalTime == nullptr)
      return CLLTimeOfDay<TS>{};
    else
      return CLLTimeOfDay<TS>{static_cast<uint16_t>(pLocalTime->tm_hour), static_cast<uint16_t>(pLocalTime->tm_min), pLocalTime->tm_sec + (now.tv_usec / 1000000.0), bDoAssert};
  #else
    //Get the current time from the CRT (by delegating to CLTimeOfDay)
    CLTimeOfDay<TS> tod{CLTimeOfDay<TS>::CurrentLocalTime(bDoAssert)};
    return CLLTimeOfDay<TS>{tod.GetTotalSeconds(), bDoAssert};
  #endif //#ifdef _WIN32
  }

  static CLLTimeOfDay Midnight() noexcept
  {
    return CLLTimeOfDay<TS>{0, 0, 0};
  }

  static CLLTimeOfDay Midday() noexcept
  {
    return CLLTimeOfDay<TS>{12, 0, 0};
  }

#ifdef _WIN32
  static String GetAMString()
  {
#pragma warning(suppress: 26477)
    return SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_S1159);
  }
#endif //#ifdef _WIN32

#ifdef _WIN32
  static String GetPMString()
  {
#pragma warning(suppress: 26477)
    return SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_S2359);
  }
#endif //#ifdef _WIN32

//Operations
  DTIME_NODISCARD uint16_t GetHours() const noexcept
  {
    DTIME_ASSERT(IsValid());

#ifdef _MSC_VER
#pragma warning(suppress: 26467)
#endif //#ifdef _MSC_VER
    return static_cast<uint16_t>(m_fTotalSeconds/3600.0);
  }

  DTIME_NODISCARD uint16_t GetMinutes() const noexcept
  {
    DTIME_ASSERT(IsValid());

#ifdef _MSC_VER
#pragma warning(suppress: 26467)
#endif //#ifdef _MSC_VER
    return static_cast<uint16_t>((m_fTotalSeconds - (GetHours() * 3600.0))/60);
  }

  DTIME_NODISCARD double GetTotalSeconds() const noexcept
  {
    DTIME_ASSERT(IsValid());

    return m_fTotalSeconds;
  }

  DTIME_NODISCARD double GetSeconds() const noexcept
  {
    DTIME_ASSERT(IsValid());

    return m_fTotalSeconds - (60.0 * (GetMinutes() + (60.0 * GetHours())));
  }

  DTIME_NODISCARD uint16_t GetAMPMHours() const noexcept
  {
    DTIME_ASSERT(IsValid());

    uint16_t wHour{GetHours()};
    if (wHour > 12)
      wHour -= 12;
    return wHour;
  }

#ifdef _WIN32
  DTIME_NODISCARD String GetAMPMString() const
  {
    DTIME_ASSERT(IsValid());

    if (GetHours() >= 12)
      return GetPMString();
    else
      return GetAMString();
  }
#endif //#ifdef _WIN32

  DTIME_NODISCARD bool IsValid() const noexcept
  {
    return !std::isnan(m_fTotalSeconds);
  }

#ifdef _WIN32
  bool GetAsSystemTime(SYSTEMTIME& SystemTime) const noexcept
  {
    if (!IsValid())
      return false;

    SystemTime.wYear = 0;
    SystemTime.wMonth = 0;
    SystemTime.wDayOfWeek = 0;
    SystemTime.wDay = 0;
    SystemTime.wHour = GetHours();
    SystemTime.wMinute = GetMinutes();
#ifdef _MSC_VER
#pragma warning(suppress: 26467)
#endif //#ifdef _MSC_VER
    SystemTime.wSecond = static_cast<WORD>(GetSeconds());
    double dInt{0};
#ifdef _MSC_VER
#pragma warning(suppress: 26467)
#endif //#ifdef _MSC_VER
    SystemTime.wMilliseconds = static_cast<WORD>((modf(m_fTotalSeconds, &dInt) * 1000.0) + 0.5);
    return true;
  }
#endif //#ifdef _WIN32

  CLLTimeOfDay& operator=(const CLLTimeOfDay&) = default;
  CLLTimeOfDay& operator=(CLLTimeOfDay&&) = default;

//Overloaded arithmetic operators
  CLLTimeOfDay operator+(const TS& TimeSpan) const noexcept
  {
    if (IsValid() && TimeSpan.IsValid())
    {
      TS t{TS{m_fTotalSeconds} + TS{0, TimeSpan.GetHours(), TimeSpan.GetMinutes(), TimeSpan.GetSeconds()}};
      return CLLTimeOfDay<TS>{t.GetHours(), t.GetMinutes(), t.GetSeconds()};
    }
    else
      return CLLTimeOfDay<TS>{};
  }

  CLLTimeOfDay operator-(const TS& TimeSpan) const noexcept
  {
    if (IsValid() && TimeSpan.IsValid())
    {
      //Addition of 1 Day is used to ensure result "t" does not go negative
      TS t{TS::OneDay() + TS{m_fTotalSeconds} - TS{0, TimeSpan.GetHours(), TimeSpan.GetMinutes(), TimeSpan.GetSeconds()}};
      return CLLTimeOfDay<TS>{t.GetHours(), t.GetMinutes(), t.GetSeconds()};
    }
    else
      return CLLTimeOfDay<TS>{};
  }

  CLLTimeOfDay& operator+=(const TS& TimeSpan) noexcept
  {
    *this = *this + TimeSpan;
    return *this;
  }

  CLLTimeOfDay& operator-=(const TS& TimeSpan) noexcept
  {
    *this = *this - TimeSpan;
    return *this;
  }

//Overloaded equality operators
  bool operator==(const CLLTimeOfDay& Tod) const noexcept
  {
    return (m_fTotalSeconds == Tod.m_fTotalSeconds);
  }

  bool operator>(const CLLTimeOfDay& Tod) const noexcept
  {
    if (IsValid() && Tod.IsValid())
      return (m_fTotalSeconds > Tod.m_fTotalSeconds);
    else
      return false; //Only makes sense to compare 2 valid values
  }

  bool operator>=(const CLLTimeOfDay& Tod) const noexcept
  {
    return operator==(Tod) || operator>(Tod);
  }

  bool operator<(const CLLTimeOfDay& Tod) const noexcept
  {
    if (IsValid() && Tod.IsValid())
      return (m_fTotalSeconds < Tod.m_fTotalSeconds);
    else
      return false; //Only makes sense to compare 2 valid values
  }

  bool operator<=(const CLLTimeOfDay& Tod) const noexcept
  {
    return operator==(Tod) || operator<(Tod);
  }

  bool operator!=(const CLLTimeOfDay& Tod) const noexcept
  {
    return !operator==(Tod);
  }

//Display
  /* The Following Format parameters are supported
  %H    Hours in the current day
  %M    Minutes in the current hour
  %h    12 Hour format Hours in this CLTimeSpan (00 - 12)
  %P    AM / PM indicator
  %S    Seconds in the current minute
  %%    Percent sign
  %x    Windows Time Of Day representation using GetTimeFormatEx() Win32 API 
  %#H, %#h, %#M, %#S  Remove leading zeros (if any)
  */
  DTIME_NODISCARD String Format(const String& sFormat) const
  { 
    String rVal;
    if (IsValid())
    {
      const auto nFmtLength{sFormat.length()};
      for (String::size_type i{0}; i<nFmtLength; i++)
      {
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
        TCHAR c{sFormat[i]};
        if (c == _T('%'))
        {
          ++i;
          if (i < nFmtLength)
          {
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
            c = sFormat[i];
            switch (c)
            {
              case _T('H'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << GetHours();
                rVal += ss.str();
                break;
              }
              case _T('h'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << GetAMPMHours();
                rVal += ss.str();
                break;
              }
              case _T('M'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << GetMinutes();
                rVal += ss.str();
                break;
              }
            #ifdef _WIN32
              case _T('P'):
              {
                rVal += GetAMPMString();
                break;
              }
            #endif //#ifdef _WIN32
              case _T('S'):
              {
                dt_ostringstream ss;
                const double fSeconds{GetSeconds()};
                if (fSeconds < 10)
                  ss << _T('0');
                ss << fSeconds;
                rVal += ss.str();
                break;
              }
            #ifdef _WIN32
              case _T('x'):
              {
                SYSTEMTIME st{};
                if (GetAsSystemTime(st))
#pragma warning(suppress: 26477)
                  rVal += SafeGetTimeFormat(LOCALE_NAME_USER_DEFAULT, 0, &st);
                break;
              }
            #endif //#ifdef _WIN32
              case _T('#'):
              {
                if (i < nFmtLength)
                {
                  ++i;
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
                  c = sFormat[i];
                  switch (c)
                  {
                    case _T('H'):
                    {
                      dt_ostringstream ss;
                      ss << GetHours();
                      rVal += ss.str();
                      break;
                    }
                    case _T('h'):
                    {
                      dt_ostringstream ss;
                      ss << GetAMPMHours();
                      rVal += ss.str();
                      break;
                    }
                    case _T('M'):
                    {
                      dt_ostringstream ss;
                      ss << GetMinutes();
                      rVal += ss.str();
                      break;
                    }
                    case _T('S'):
                    {
                      dt_ostringstream ss;
                      ss << GetSeconds();
                      rVal += ss.str();
                      break;
                    }
                    default:
                    {
                      rVal += c;
                      break;
                    }
                  }
                }
                break;
              }
              default:
              {
                rVal += c;
                break;
              }
            }
          }
        }
        else
        {
          rVal += c;
        }
      }
    }
    return rVal;
  }

//Serialization
#ifdef _AFX
  void Serialize(CArchive& ar)
  {
    if (ar.IsStoring())
    {
      constexpr const WORD wVersion{0x100};
      ar << wVersion;
      ar << m_fTotalSeconds;
    }
    else
    {
      WORD wVersion{0};
      ar >> wVersion;
      ASSERT(wVersion == 0x100);
      ar >> m_fTotalSeconds;
    }
  }

  friend CArchive& operator<<(CArchive& ar, CLLTimeOfDay& tod)
  {
    ASSERT(ar.IsStoring());
    tod.Serialize(ar);
    return ar;
  }

  friend CArchive& operator>>(CArchive& ar, CLLTimeOfDay& tod)
  {
    ASSERT(!ar.IsStoring());
    tod.Serialize(ar);
    return ar;
  }
#endif //#ifdef _AFX

protected:
//Member variables
  double m_fTotalSeconds; //Actual instance data
};

} //namespace DTimePlus

#endif //#ifndef __DTIMEPLUSCLLTIMEOFDAY_H__
