/*
Module : DTime+CLLTimeSpan.h
Purpose: Defines the interface for the DTime+ class library
Created: PJN / 01-01-2018

Copyright (c) 2018 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#if _MSC_VER > 1000
#pragma once
#endif //#if _MSC_VER > 1000

#ifndef __DTIMEPLUSCLLTIMESPAN_H__
#define __DTIMEPLUSCLLTIMESPAN_H__


//////////////////// Includes /////////////////////////////////////////////////

#include <ctime>
#include <cmath>
#include <cstdint>
#include <sstream>
#include <iomanip>
#include <climits>
#include "DTime+Defines.h"
#include "DTime+Helpers.h"


//////////////////// Classes //////////////////////////////////////////////////

namespace DTimePlus 
{

//A Time span class which is used in conjunction with the time class CLLDate
class CLLTimeSpan
{
public:
//Constructors
#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLLTimeSpan() noexcept
  {
    Set();
  }

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLLTimeSpan(int32_t Day, uint16_t Hour, uint16_t Minute, double Second) noexcept
  {
    Set(Day, Hour, Minute, Second);
  }

  CLLTimeSpan(const CLLTimeSpan&) = default;
  CLLTimeSpan(CLLTimeSpan&&) = default;

#ifdef _AFX
#pragma warning(suppress: 26495)
  CLLTimeSpan(const CTimeSpan& ts) noexcept
  {
    Set(ts);
  }
#endif //#ifdef _AFX

#ifdef _AFX
#pragma warning(suppress: 26495)
  CLLTimeSpan(const COleDateTimeSpan& oleTimeSpan) noexcept
  {
    Set(oleTimeSpan);
  }
#endif //#ifdef _AFX

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLLTimeSpan(const double& Seconds) noexcept
  {
    Set(Seconds);
  }

  ~CLLTimeSpan() = default;

//Set operators
  CLLTimeSpan& Set() noexcept
  {
    //Create an invalid value
#ifdef _MSC_VER
#pragma warning(suppress: 26451)
#endif //#ifdef _MSC_VER
    m_fSeconds = NAN;
    return *this;
  }

  CLLTimeSpan& Set(int32_t Day, uint16_t Hour, uint16_t Minute, double Second) noexcept
  {
    if (Day >= 0)
      m_fSeconds = (Day*86400.0) + (Hour*3600.0) + (Minute*60.0) + Second;
    else
      m_fSeconds = (Day*86400.0) - (Hour*3600.0) - (Minute*60.0) - Second;
    return *this;
  }

#ifdef _AFX
  CLLTimeSpan& Set(const CTimeSpan& ts) noexcept
  {
    m_fSeconds = static_cast<double>(ts.GetTotalSeconds());
    return *this;
  }
#endif //#ifdef _AFX

#ifdef _AFX
  CLLTimeSpan& Set(const COleDateTimeSpan& oleTimeSpan) noexcept
  {
    if (oleTimeSpan.GetStatus() == COleDateTimeSpan::valid)
      Set(static_cast<double>(oleTimeSpan.GetTotalSeconds()));
    else
      Set();
    return *this;
  }
#endif //#ifdef _AFX

  CLLTimeSpan& Set(const double& Seconds) noexcept
  {
    m_fSeconds = Seconds;
    return *this;
  }

//Static constructors
  static CLLTimeSpan OneDay() noexcept
  {
    return {1, 0, 0, 0};
  }

  static CLLTimeSpan OneHour() noexcept
  {
    return {0, 1, 0, 0};
  }

  static CLLTimeSpan OneMinute() noexcept
  {
    return {0, 0, 1, 0};
  }

  static CLLTimeSpan OneSecond() noexcept
  {
    return {0, 0, 0, 1};
  }

//Operations
  DTIME_NODISCARD double GetValue() const noexcept
  {
    return m_fSeconds;
  }

  DTIME_NODISCARD int32_t GetTotalDays() const noexcept
  {
    DTIME_ASSERT(IsValid());

    return static_cast<int32_t>(m_fSeconds/86400);
  }

  DTIME_NODISCARD uint16_t GetHours() const noexcept
  {
    DTIME_ASSERT(IsValid());

    if (IsPositiveSpan())
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26467)
#endif //#ifdef _MSC_VER
      return static_cast<uint16_t>((m_fSeconds - 86400.0*GetTotalDays())/3600.0);
    }
    else
    {
      CLLTimeSpan n{*this};
      n.Negate();
      return n.GetHours();
    }
  }

  DTIME_NODISCARD uint16_t GetMinutes() const noexcept
  {
    DTIME_ASSERT(IsValid());

    if (IsPositiveSpan())
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26467)
#endif //#ifdef _MSC_VER
      return static_cast<uint16_t>((m_fSeconds - (86400.0*GetTotalDays()) - (3600.0*GetHours()))/60);
    }
    else
    {
      CLLTimeSpan n{*this};
      n.Negate();
      return n.GetMinutes();
    }
  }

  DTIME_NODISCARD double GetSeconds() const noexcept
  {
    DTIME_ASSERT(IsValid());

    if (IsPositiveSpan())
      return m_fSeconds - (86400.0*GetTotalDays()) - (3600.0*GetHours()) - (60.0*GetMinutes());
    else
    {
      CLLTimeSpan n{*this};
      n.Negate();
      return n.GetSeconds();
    }
  }

  DTIME_NODISCARD bool IsValid() const noexcept
  {
    return !std::isnan(m_fSeconds);
  }

  CLLTimeSpan& Negate() noexcept
  {
    if (IsValid())
      m_fSeconds = (-m_fSeconds);
    return *this;
  }

  DTIME_NODISCARD bool IsPositiveSpan() const noexcept
  {
    DTIME_ASSERT(IsValid());

    return (m_fSeconds >= 0);
  }

  CLLTimeSpan& operator=(const CLLTimeSpan&) = default;
  CLLTimeSpan& operator=(CLLTimeSpan&&) = default;

//Overloaded arithmetic operators
  CLLTimeSpan operator+(const CLLTimeSpan& TimeSpan) const noexcept
  {
    if (IsValid() && TimeSpan.IsValid())
    {
      const double fSeconds{m_fSeconds + TimeSpan.m_fSeconds};
      return {fSeconds};
    }
    else
      return {};
  }

  CLLTimeSpan operator-(const CLLTimeSpan& TimeSpan) const noexcept
  {
    if (IsValid() && TimeSpan.IsValid())
    {
      const double fSeconds{m_fSeconds - TimeSpan.m_fSeconds};
      return {fSeconds};
    }
    else
      return {};
  }

  CLLTimeSpan& operator+=(const CLLTimeSpan& TimeSpan) noexcept
  {
    *this = *this + TimeSpan;
    return *this;
  }

  CLLTimeSpan& operator-=(const CLLTimeSpan& TimeSpan) noexcept
  {
    *this = *this - TimeSpan;
    return *this;
  }

  friend CLLTimeSpan operator-(const CLLTimeSpan& TimeSpan) noexcept
  {
    CLLTimeSpan rVal{TimeSpan};
    rVal.Negate();
    return rVal;
  }

  CLLTimeSpan operator*(double Multiplier) const noexcept
  {
    if (IsValid())
    {
      const double rVal{m_fSeconds * Multiplier};
      return {rVal};
    }
    else
      return {};
  }

  CLLTimeSpan operator/(double divisor) const noexcept
  {
    if (IsValid())
    {
      const double rVal{m_fSeconds / divisor};
      return {rVal};
    }
    else
      return {};
  }

  CLLTimeSpan& operator*=(double Multiplier) noexcept
  {
    *this = operator*(Multiplier);
    return *this;
  }

  CLLTimeSpan& operator/=(double Divisor) noexcept
  {
    *this = operator/(Divisor);
    return *this;
  }

//Overloaded equality operators
  bool operator==(const CLLTimeSpan& TimeSpan) const noexcept
  {
    const bool bValid{IsValid()};
    const bool bOtherValid{TimeSpan.IsValid()};
    if (bValid && bOtherValid)
      return (m_fSeconds == TimeSpan.m_fSeconds);
    else
      return (bValid == bOtherValid);
  }

  bool operator>(const CLLTimeSpan& TimeSpan) const noexcept
  {
    if (IsValid() && TimeSpan.IsValid())
      return (m_fSeconds > TimeSpan.m_fSeconds);
    else
      return false; //Only makes sense to compare 2 valid values
  }

  bool operator>=(const CLLTimeSpan& TimeSpan) const noexcept
  {
    return operator==(TimeSpan) || operator>(TimeSpan);
  }

  bool operator<(const CLLTimeSpan& TimeSpan) const noexcept
  {
    if (IsValid() && TimeSpan.IsValid())
      return (m_fSeconds < TimeSpan.m_fSeconds);
    else
      return false; //Only makes sense to compare 2 valid values
  }

  bool operator<=(const CLLTimeSpan& TimeSpan) const noexcept
  {
    return operator==(TimeSpan) || operator<(TimeSpan);
  }

  bool operator!=(const CLLTimeSpan& TimeSpan) const noexcept
  {
    return !operator==(TimeSpan);
  }

//Display
  /*  The Following Format parameters are supported
  %D    Total days in this CLLTimeSpan
  %H    Hours in this CLLTimeSpan (00 - 23)
  %M    Minutes in the current hour in this CLLTimeSpan (00 - 59)
  %S    Seconds in the current minute in this CLLTimeSpan (00 - 59)
  %%    Percent sign
  %#H, %#M, %#S  Remove leading zeros (if any).
  */
  DTIME_NODISCARD String Format(const String& sFormat) const
  {
    //What will be the return value from this function
    String rVal;

    if (IsValid())
    {
      const auto nFmtLength{sFormat.length()};
      for (String::size_type i{0}; i<nFmtLength; i++)
      {
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
        TCHAR c{sFormat[i]};
        if (c == _T('%'))
        {
          ++i;
          if (i < nFmtLength)
          {
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
            c = sFormat[i];
            switch (c)
            {
              case _T('D'):
              {
                dt_ostringstream ss;
                const int32_t Days{GetTotalDays()};
                if (IsPositiveSpan())
                  ss << _T("+");
                else if (Days == 0) //handle negative 0
                  ss << _T("-");
                ss << Days;
                rVal += ss.str();
                break;
              }
              case _T('H'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << GetHours();
                rVal += ss.str();
                break;
              }
              case _T('M'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << GetMinutes();
                rVal += ss.str();
                break;
              }
              case _T('S'):
              {
                dt_ostringstream ss;
                const double fSeconds{GetSeconds()};
                if (fSeconds < 10)
                  ss << _T('0');
                ss << fSeconds;
                rVal += ss.str();
                break;
              }
              case _T('#'):
              {
                if (i < nFmtLength)
                {
                  ++i;
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
                  c = sFormat[i];
                  switch (c)
                  {
                    case _T('H'):
                    {
                      dt_ostringstream ss;
                      ss << GetHours();
                      rVal += ss.str();
                      break;
                    }
                    case _T('M'):
                    {
                      dt_ostringstream ss;
                      ss << GetMinutes();
                      rVal += ss.str();
                      break;
                    }
                    case _T('S'):
                    {
                      dt_ostringstream ss;
                      ss << GetSeconds();
                      rVal += ss.str();
                      break;
                    }
                    default:
                    {
                      rVal += c;
                      break;
                    }
                  }
                }
                break;
              }
              default:
              {
                rVal += c;
                break;
              }
            }
          }
        }
        else
        {
          rVal += c;
        }
      }
    }
    return rVal;
  }

//Serialization
#ifdef _AFX
  void Serialize(CArchive& ar)
  {
    if (ar.IsStoring())
    {
      constexpr const WORD wVersion{0x100};
      ar << wVersion;
      ar.Write(&m_fSeconds, sizeof(m_fSeconds));
    }
    else
    {
      WORD wVersion{0};
      ar >> wVersion;
      ASSERT(wVersion == 0x100);
      ar.Read(&m_fSeconds, sizeof(m_fSeconds));
    }
  }
#endif //#ifdef _AFX

#ifdef _AFX
  friend CArchive& operator<<(CArchive& ar, CLLTimeSpan& TimeSpan)
  {
    ASSERT(ar.IsStoring());
    TimeSpan.Serialize(ar);
    return ar;
  }
#endif //#ifdef _AFX

#ifdef _AFX
  friend CArchive& operator>>(CArchive& ar, CLLTimeSpan& TimeSpan)
  {
    ASSERT(!ar.IsStoring());
    TimeSpan.Serialize(ar);
    return ar;
  }
#endif //#ifdef _AFX

protected:
//Member variables
  double m_fSeconds; //Actual instance data
};

inline CLLTimeSpan operator*(double Multiplier, const DTimePlus::CLLTimeSpan& TimeSpan) noexcept
{
  return TimeSpan.operator*(Multiplier);
}

} //namespace DTimePlus


#endif //#ifndef __DTIMEPLUSCLLTimeSpan_H__
