/*
Module : DTime+CLTimeOfDay.h
Purpose: Defines the interface for the DTime+ CLTimeOfDay class
Created: PJN / 01-05-2006
History: PJN / 29-12-2017 1. Updated copyright details.
                          2. Fix compile problems in CLTimeOfDay::GetWindowsFormatString in VC 2017

Copyright (c) 2006 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#if _MSC_VER > 1000
#pragma once
#endif //#if _MSC_VER > 1000

#ifndef __DTIMEPLUSCLTIMEOFDAY_H__
#define __DTIMEPLUSCLTIMEOFDAY_H__


//////////////////// Includes /////////////////////////////////////////////////

#include <cstdint>
#include <ctime>
#include <sstream>
#include <iomanip>
#include "DTime+Defines.h"
#include "DTime+Helpers.h"


//////////////////// Classes //////////////////////////////////////////////////

namespace DTimePlus
{

//Forward declaration
class CLTimeSpan;

//A Time of Day class with a granularity of 1 second which
//is used in conjunction with the time class CDate in the CLDate class
template<class TS = DTimePlus::CLTimeSpan>
class CLTimeOfDay
{
public:
//Constructors
#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLTimeOfDay() noexcept
  {
    Set();
  }

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLTimeOfDay(uint16_t Hour, uint16_t Minute, uint16_t Second, bool bDoAssert = true) noexcept
  {
    Set(Hour, Minute, Second, bDoAssert);
  }

#ifdef _WIN32
#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLTimeOfDay(const SYSTEMTIME& SystemTime, bool bDoAssert = true) noexcept
  {
    Set(SystemTime, bDoAssert);
  }
#endif //#ifdef _WIN32

  CLTimeOfDay(const CLTimeOfDay&) = default;
  CLTimeOfDay(CLTimeOfDay&&) = default;

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLTimeOfDay(uint32_t TotalSeconds, bool bDoAssert = true) noexcept
  {
    Set(TotalSeconds, bDoAssert);
  }

  ~CLTimeOfDay() = default;

//Set operators
  CLTimeOfDay& Set() noexcept
  {
    //Create an invalid value
    m_dwTotalSeconds = 0xFFFFFFFF;
    return *this;
  }

#ifdef _DEBUG
  CLTimeOfDay& Set(uint16_t Hour, uint16_t Minute, uint16_t Second, bool bDoAssert = true) noexcept
#else
  CLTimeOfDay& Set(uint16_t Hour, uint16_t Minute, uint16_t Second, bool /*bDoAssert*/ = true) noexcept
#endif
  {
    m_dwTotalSeconds = (Hour * 3600) + (Minute * 60) + Second;
    if (m_dwTotalSeconds > 86400L) //Note we allow 23:59:60 to cater for leap seconds
    {
    #ifdef _DEBUG
      if (bDoAssert)
        DTIME_ASSERT(false);
    #endif //#ifdef _DEBUG
      Set();
      return *this;
    }
    return *this;
  }

#ifdef _WIN32
  CLTimeOfDay& Set(const SYSTEMTIME& SystemTime, bool bDoAssert = true) noexcept
  {
    return Set(SystemTime.wHour, SystemTime.wMinute, SystemTime.wSecond, bDoAssert);
  }
#endif //#ifdef _WIN32

#ifdef _DEBUG
  CLTimeOfDay& Set(uint32_t TotalSeconds, bool bDoAssert = true) noexcept
#else
  CLTimeOfDay& Set(uint32_t TotalSeconds, bool /*bDoAssert*/ = true) noexcept
#endif //#ifdef _DEBUG
  {
    m_dwTotalSeconds = TotalSeconds;
    if (m_dwTotalSeconds > 86400L) //Note we allow 23:59:60 to cater for leap seconds
    {
    #ifdef _DEBUG
      if (bDoAssert)
        DTIME_ASSERT(false);
    #endif //#ifdef _DEBUG
      Set();
      return *this;
    }
    return *this;
  }

  CLTimeOfDay& SetFromCollate(uint32_t Time, bool bDoAssert = true)
  {
    dt_ostringstream ss;
    ss << Time;
    String sTime(ss.str());

    const auto nLen{sTime.length()};
    String sSecond{sTime.substr(nLen - 2, 2)};
    String sMinute{sTime.substr(nLen - 4, 2)};
    String sHour{sTime.substr(0, nLen - 4)};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    const auto Second{static_cast<uint16_t>(_ttoi(sSecond.c_str()))};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    const auto Minute{static_cast<uint16_t>(_ttoi(sMinute.c_str()))};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    const auto Hour{static_cast<uint16_t>(_ttoi(sHour.c_str()))};
#ifdef _MSC_VER
#pragma warning(suppress: 26487 26489)
#endif //#ifdef _MSC_VER
    return Set(Hour, Minute, Second, bDoAssert);
  }

//Static constructors
  static CLTimeOfDay CurrentSystemTime(bool bDoAssert = true) noexcept
  { 
    //Get the current time from the CRT
    time_t now{0};
    if (time(&now) != -1)
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      const auto nDays{static_cast<int32_t>(now / 86400)};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      const auto nSecondsLeftOver{static_cast<uint32_t>(now - (static_cast<time_t>(nDays) * 86400))};
      return CLTimeOfDay<TS>{nSecondsLeftOver, bDoAssert};
    }
    else
      return CLTimeOfDay<TS>{};
  }

  static CLTimeOfDay CurrentLocalTime(bool bDoAssert = true) noexcept
  {
    //Get the current time from the CRT
    time_t now{0};
    if (time(&now) != -1)
    {
      //Convert to local time
    #ifdef _MSC_VER
      tm lt{};
      const errno_t e{localtime_s(&lt, &now)};
      if (e != 0)
        return CLTimeOfDay<TS>{};
      const tm* pLocalTime{&lt};
    #else
      const tm* pLocalTime{localtime(&now)};
    #endif //#ifdef _MSC_VER
      if (pLocalTime == nullptr)
        return CLTimeOfDay<TS>{};
      else
      {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
        return CLTimeOfDay<TS>{static_cast<uint16_t>(pLocalTime->tm_hour), static_cast<uint16_t>(pLocalTime->tm_min), static_cast<uint16_t>(pLocalTime->tm_sec), bDoAssert};
      }
    }
    else
      return CLTimeOfDay<TS>{};
  }

  static CLTimeOfDay Midnight() noexcept
  {
    return CLTimeOfDay<TS>{0, 0, 0};
  }

  static CLTimeOfDay Midday() noexcept
  {
    return CLTimeOfDay<TS>{12, 0, 0};
  }

//static operations
#ifdef _WIN32
#pragma warning(suppress: 26477)
  static String GetWindowsFormatString(LPCWSTR lpLocaleName = LOCALE_NAME_USER_DEFAULT)
  {
    //Is there to be a leading 0 on the Hour?
    bool bHourLZ{false};
    String sBuf{SafeGetLocaleInfo(lpLocaleName, LOCALE_ITLZERO)};
    if (sBuf == _T("0"))  //No leading zeros for hours
      bHourLZ = false;
    else if (sBuf == _T("1"))  //Leading zeros for hours
      bHourLZ = true;
    else
      DTIME_ASSERT(false);

    //Get the time separator character
    String sSep{_T(":")};
    sBuf = SafeGetLocaleInfo(lpLocaleName, LOCALE_STIME);
    if (sBuf.length())
      sSep = sBuf;

    //Get the 24 Hour / 12 Hour indicator
    bool b24Hour{false};
    sBuf = SafeGetLocaleInfo(lpLocaleName, LOCALE_ITIME);
    if (sBuf == _T("0")) //AM / PM Style
      b24Hour = false;
    else if (sBuf == _T("1")) //24 Hour Style
      b24Hour = true;
    else
      DTIME_ASSERT(false);

    //What will be the return value from this function
    String sFormatString;

    //Modify behaviour depending on the various flags
    if (bHourLZ)
    {
      if (b24Hour)
        sFormatString += _T("%H");
      else
        sFormatString += _T("%h");
    }
    else
    {
      if (b24Hour)
        sFormatString += _T("%#H");
      else
        sFormatString += _T("%#h");
    }
    sFormatString += sSep;
    sFormatString += _T("%M");
    sFormatString += sSep;
    sFormatString += _T("%S");
    if (!b24Hour)
      sFormatString += _T(" %P");
    return sFormatString;
  }
#endif //#ifdef _WIN32

#ifdef _WIN32
  static String GetAMString()
  {
#pragma warning(suppress: 26477)
    return SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_S1159);
  }
#endif //#ifdef _WIN32

#ifdef _WIN32
  static String GetPMString()
  {
#pragma warning(suppress: 26477)
    return SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_S2359);
  }
#endif //#ifdef _WIN32

//Operations
  DTIME_NODISCARD uint16_t GetHours() const noexcept
  {
    DTIME_ASSERT(IsValid());

#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    return static_cast<uint16_t>(m_dwTotalSeconds/3600);
  }

  DTIME_NODISCARD uint16_t GetMinutes() const noexcept
  {
    DTIME_ASSERT(IsValid());

#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    return static_cast<uint16_t>((m_dwTotalSeconds - (GetHours() * 3600))/60);
  }

  DTIME_NODISCARD uint32_t GetTotalSeconds() const noexcept
  {
    DTIME_ASSERT(IsValid());

    return m_dwTotalSeconds;
  }

  DTIME_NODISCARD uint16_t GetSeconds() const noexcept
  {
    DTIME_ASSERT(IsValid());

#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    return static_cast<uint16_t>(m_dwTotalSeconds - (60 * (GetMinutes() + (60 * GetHours()))));
  }

  DTIME_NODISCARD uint16_t GetAMPMHours() const noexcept
  {
    DTIME_ASSERT(IsValid());

    uint16_t wHour{GetHours()};
    if (wHour > 12)
      wHour -= 12;
    return wHour;
  }

#ifdef _WIN32
  DTIME_NODISCARD String GetAMPMString() const
  {
    DTIME_ASSERT(IsValid());

    if (GetHours() >= 12)
      return GetPMString();
    else
      return GetAMString();
  }
#endif //#ifdef _WIN32

  DTIME_NODISCARD bool IsValid() const noexcept
  {
    return (m_dwTotalSeconds != 0xFFFFFFFF);
  }

  DTIME_NODISCARD uint32_t Collate() const noexcept
  {
    const uint32_t rVal = (GetHours() * 10000L) + (GetMinutes() * 100L) + GetSeconds();
    return rVal;
  }

#ifdef _WIN32
  bool GetAsSystemTime(SYSTEMTIME& SystemTime) const noexcept
  {
    if (!IsValid())
      return false;
    SystemTime.wYear = 0;
    SystemTime.wMonth = 0;
    SystemTime.wDayOfWeek = 0;
    SystemTime.wDay = 0;
    SystemTime.wHour = GetHours();
    SystemTime.wMinute = GetMinutes();
    SystemTime.wSecond = GetSeconds();
    SystemTime.wMilliseconds = 0;
    return true;
  }
#endif //#ifdef _WIN32

  bool GetAsTM(tm& time) const noexcept
  {
    if (!IsValid())
      return false;
    time.tm_hour = GetHours();
    time.tm_min = GetMinutes();
    time.tm_sec = GetSeconds();
    time.tm_mday = 0;
    time.tm_mon = 0;
    time.tm_year = 0;
    time.tm_wday = 0;
    time.tm_yday = 0;
    time.tm_isdst = 0;
    return true;
  }

  CLTimeOfDay& operator=(const CLTimeOfDay&) = default;
  CLTimeOfDay& operator=(CLTimeOfDay&&) = default;

//Overloaded arithmetic operators
  CLTimeOfDay operator+(const TS& TimeSpan) const noexcept
  {
    if (IsValid() && TimeSpan.IsValid())
    {
      TS t{TS{m_dwTotalSeconds} + TS{0, TimeSpan.GetHours(), TimeSpan.GetMinutes(), TimeSpan.GetSeconds()}};
      return CLTimeOfDay<TS>{t.GetHours(), t.GetMinutes(), t.GetSeconds()};
    }
    else
      return CLTimeOfDay<TS>{};
  }

  CLTimeOfDay operator-(const TS& TimeSpan) const noexcept
  {
    if (IsValid() && TimeSpan.IsValid())
    {
      //Addition of 1 Day is used to ensure result "t" does not go negative
      TS t{TS::OneDay() + TS{m_dwTotalSeconds} - TS{0, TimeSpan.GetHours(), TimeSpan.GetMinutes(), TimeSpan.GetSeconds()}};
      return CLTimeOfDay<TS>{t.GetHours(), t.GetMinutes(), t.GetSeconds()};
    }
    else
      return CLTimeOfDay<TS>{};
  }

  CLTimeOfDay& operator+=(const TS& TimeSpan) noexcept
  {
    *this = *this + TimeSpan;
    return *this;
  }

  CLTimeOfDay& operator-=(const TS& TimeSpan) noexcept
  {
    *this = *this - TimeSpan;
    return *this;
  }

//Overloaded equality operators
  bool operator==(const CLTimeOfDay& Tod) const noexcept
  {
    return (m_dwTotalSeconds == Tod.m_dwTotalSeconds);
  }

  bool operator>(const CLTimeOfDay& Tod) const noexcept
  {
    if (IsValid() && Tod.IsValid())
      return (m_dwTotalSeconds > Tod.m_dwTotalSeconds);
    else
      return false; //Only makes sense to compare 2 valid values
  }

  bool operator>=(const CLTimeOfDay& Tod) const noexcept
  {
    return operator==(Tod) || operator>(Tod);
  }

  bool operator<(const CLTimeOfDay& Tod) const noexcept
  {
    if (IsValid() && Tod.IsValid())
      return (m_dwTotalSeconds < Tod.m_dwTotalSeconds);
    else
      return false; //Only makes sense to compare 2 valid values
  }

  bool operator<=(const CLTimeOfDay& Tod) const noexcept
  {
    return operator==(Tod) || operator<(Tod);
  }

  bool operator!=(const CLTimeOfDay& Tod) const noexcept
  {
    return !operator==(Tod);
  }

//Display
  /* The Following Format parameters are supported
  %H Hours in the current day
  %M Minutes in the current hour
  %h 12 Hour format Hours in this CLTimeSpan (00 - 12)
  %P AM / PM indicator
  %S Seconds in the current minute
  %% Percent sign
  %x Windows Time Of Day representation using GetTimeFormatEx() Win32 API 
  %#H, %#h, %#M, %#S  Remove leading zeros (if any)
  */
  DTIME_NODISCARD String Format(const String& sFormat) const
  {
    String rVal;
    if (IsValid())
    {
      const auto nFmtLength{sFormat.length()};
      for (String::size_type i{0}; i<nFmtLength; i++)
      {
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
        TCHAR c{sFormat[i]};
        if (c == _T('%'))
        {
          ++i;
          if (i < nFmtLength)
          {
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
            c = sFormat[i];
            switch (c)
            {
              case _T('H'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << GetHours();
                rVal += ss.str();
                break;
              }
              case _T('h'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << GetAMPMHours();
                rVal += ss.str();
                break;
              }
              case _T('M'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << GetMinutes();
                rVal += ss.str();
                break;
              }
            #ifdef _WIN32
              case _T('P'):
              {
                rVal += GetAMPMString();
                break;
              }
            #endif //#ifdef _WIN32
              case _T('S'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << GetSeconds();
                rVal += ss.str();
                break;
              }
            #ifdef _WIN32
              case _T('x'):
              {
                SYSTEMTIME st{};
                if (GetAsSystemTime(st))
#pragma warning(suppress: 26477)
                  rVal += SafeGetTimeFormat(LOCALE_NAME_USER_DEFAULT, 0, &st);
                break;
              }
            #endif //#ifdef _WIN32
              case _T('#'):
              {
                if (i < nFmtLength)
                {
                  ++i;
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
                  c = sFormat[i];
                  switch (c)
                  {
                    case _T('H'):
                    {
                      dt_ostringstream ss;
                      ss << GetHours();
                      rVal += ss.str();
                      break;
                    }
                    case _T('h'):
                    {
                      dt_ostringstream ss;
                      ss << GetAMPMHours();
                      rVal += ss.str();
                      break;
                    }
                    case _T('M'):
                    {
                      dt_ostringstream ss;
                      ss << GetMinutes();
                      rVal += ss.str();
                      break;
                    }
                    case _T('S'):
                    {
                      dt_ostringstream ss;
                      ss << GetSeconds();
                      rVal += ss.str();
                      break;
                    }
                    default:
                    {
                      rVal += c;
                      break;
                    }
                  }
                }
                break;
              }
              default:
              {
                rVal += c;
                break;
              }
            }
          }
        }
        else
        {
          rVal += c;
        }
      }
    }
    return rVal;
  }

//Serialization
#ifdef _AFX
  void Serialize(CArchive& ar)
  {
    if (ar.IsStoring())
    {
      constexpr const WORD wVersion{0x100};
      ar << wVersion;
      ar << m_dwTotalSeconds;
    }
    else
    {
      WORD wVersion{0};
      ar >> wVersion;
      ASSERT(wVersion == 0x100);
      ar >> m_dwTotalSeconds;
    }
  }

  friend CArchive& operator<<(CArchive& ar, CLTimeOfDay& tod)
  {
    ASSERT(ar.IsStoring());
    tod.Serialize(ar);
    return ar;
  }

  friend CArchive& operator>>(CArchive& ar, CLTimeOfDay& tod)
  {
    ASSERT(!ar.IsStoring());
    tod.Serialize(ar);
    return ar;
  }
#endif //#ifdef _AFX

protected:
//Member variables
  uint32_t m_dwTotalSeconds; //Actual instance data 
};

} //namespace DTimePlus

#endif //#ifndef __DTIMEPLUSCLTIMEOFDAY_H__
