/*
Module : DTime+CLTimeSpan.h
Purpose: Defines the interface for the DTime+ class library
Created: PJN / 01-05-2006
History: PJN / 01-01-2018 1. Made layout of CLTimeSpan::GetWindowsFormatString parameters consistent with 
                          the other GetWindowsFormatString implementations.
                          2. Fixed a bug in CLTimeSpan::GetWindowsFormatString related to leading zeros
                          for the hour value.

Copyright (c) 2006 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#if _MSC_VER > 1000
#pragma once
#endif //#if _MSC_VER > 1000

#ifndef __DTIMEPLUSCLTIMESPAN_H__
#define __DTIMEPLUSCLTIMESPAN_H__


//////////////////// Includes /////////////////////////////////////////////////

#include <ctime>
#include <cstdint>
#include <sstream>
#include <iomanip>
#include <climits>
#include "DTime+Defines.h"
#include "DTime+Helpers.h"


//////////////////// Classes //////////////////////////////////////////////////

namespace DTimePlus
{

//A Time span class with a granularity of 1 second which
//is used in conjunction with the time class CLDate
class CLTimeSpan
{
public:
//Constructors
#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLTimeSpan() noexcept
  {
    Set();
  }

#ifdef _MSC_VER
#pragma warning(suppress: 26495)
#endif //#ifdef _MSC_VER
  CLTimeSpan(int32_t Day, uint16_t Hour, uint16_t Minute, uint16_t Second) noexcept
  {
    Set(Day, Hour, Minute, Second);
  }

  CLTimeSpan(const CLTimeSpan&) = default;
  CLTimeSpan(CLTimeSpan&&) = default;

#ifdef _AFX
#pragma warning(suppress: 26495)
  CLTimeSpan(const CTimeSpan& ts) noexcept
  {
    Set(ts);
  }
#endif //#ifdef _AFX

#ifdef _AFX
#pragma warning(suppress: 26495)
  CLTimeSpan(const COleDateTimeSpan& oleTimeSpan) noexcept
  {
    Set(oleTimeSpan);
  }
#endif //#ifdef _AFX

  CLTimeSpan(const int64_t& Seconds) noexcept : m_nSeconds(Seconds)
  {
  }

  ~CLTimeSpan() = default;

//Set operators
  CLTimeSpan& Set() noexcept
  {
    //Create an invalid value
    m_nSeconds = INT64_MIN;
    return *this;
  }

  CLTimeSpan& Set(int32_t Day, uint16_t Hour, uint16_t Minute, uint16_t Second) noexcept
  {
    if (Day >= 0)
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      m_nSeconds = (static_cast<int64_t>(Day)*86400L) + (static_cast<int64_t>(Hour)*3600L) + (static_cast<int64_t>(Minute)*60L) + Second;
    }
    else
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      m_nSeconds = (static_cast<int64_t>(Day)*86400L) - (static_cast<int64_t>(Hour)*3600L) - (static_cast<int64_t>(Minute)*60L) - Second;
    }
    return *this;
  }

#ifdef _AFX
  CLTimeSpan& Set(const CTimeSpan& ts) noexcept
  {
    m_nSeconds = ts.GetTotalSeconds();
    return *this;
  }
#endif //#ifdef _AFX

#ifdef _AFX
  CLTimeSpan& Set(const COleDateTimeSpan& oleTimeSpan) noexcept
  {
    if (oleTimeSpan.GetStatus() == COleDateTimeSpan::valid)
      Set(static_cast<int64_t>(oleTimeSpan.GetTotalSeconds()));
    else
      Set();
    return *this;
  }
#endif //#ifdef _AFX

  CLTimeSpan& Set(const int64_t& Seconds) noexcept
  {
    m_nSeconds = Seconds;
    return *this;
  }

//Static constructors
  static CLTimeSpan OneDay() noexcept
  {
    return CLTimeSpan{1, 0, 0, 0};
  }

  static CLTimeSpan OneHour() noexcept
  {
    return CLTimeSpan{0, 1, 0, 0};
  }

  static CLTimeSpan OneMinute() noexcept
  {
    return CLTimeSpan{0, 0, 1, 0};
  }

  static CLTimeSpan OneSecond() noexcept
  {
    return CLTimeSpan{0, 0, 0, 1};
  }

//Static operations
#ifdef _WIN32
#pragma warning(suppress: 26477)
  static String GetWindowsFormatString(LPCWSTR lpLocaleName = LOCALE_NAME_USER_DEFAULT)
  {
    //Is there to be a leading 0 on the Hour?
    bool bHourLZ{false};
    String sBuf{DTimePlus::SafeGetLocaleInfo(lpLocaleName, LOCALE_ITLZERO)};
    if (sBuf == _T("0")) //No leading zeros for hours
      bHourLZ = false;
    else if (sBuf == _T("1")) //Leading zeros for hours
      bHourLZ = true;
    else
      DTIME_ASSERT(false);

    //Get the time separator character
    String sSep{_T(":")};
    sBuf = DTimePlus::SafeGetLocaleInfo(lpLocaleName, LOCALE_STIME);
    if (sBuf.length())
      sSep = sBuf;

    //Form the return string
    String sFormatString{_T("%D ")};
    if (!bHourLZ)
      sFormatString += _T("%#H");
    else
      sFormatString += _T("%H");
    sFormatString += sSep;
    sFormatString += _T("%M");
    sFormatString += sSep;
    sFormatString += _T("%S");
    return sFormatString;
  }
#endif //#ifdef _WIN32

//Operations
  DTIME_NODISCARD int64_t GetValue() const noexcept
  {
    DTIME_ASSERT(IsValid());

    return m_nSeconds;
  }

  DTIME_NODISCARD int32_t GetTotalDays() const noexcept
  {
    DTIME_ASSERT(IsValid());

#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    return static_cast<int32_t>(m_nSeconds/86400);
  }

  DTIME_NODISCARD uint16_t GetHours() const noexcept
  {
    DTIME_ASSERT(IsValid());

    if (IsPositiveSpan())
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      return static_cast<uint16_t>((m_nSeconds - 86400 * static_cast<int64_t>(GetTotalDays())) / 3600L);
    }
    else
    {
      CLTimeSpan n{*this};
      n.Negate();
      return n.GetHours();
    }
  }

  DTIME_NODISCARD uint16_t GetMinutes() const noexcept
  {
    DTIME_ASSERT(IsValid());

    if (IsPositiveSpan())
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      return static_cast<uint16_t>((m_nSeconds - 86400*static_cast<int64_t>(GetTotalDays()) - (3600*static_cast<int64_t>(GetHours()))) / 60);
    }
    else
    {
      CLTimeSpan n{*this};
      n.Negate();
      return n.GetMinutes();
    }
  }

  DTIME_NODISCARD uint16_t GetSeconds() const noexcept
  {
    DTIME_ASSERT(IsValid());

    if (IsPositiveSpan())
    {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
      return static_cast<uint16_t>(m_nSeconds - 86400 * static_cast<int64_t>(GetTotalDays()) - (3600*static_cast<int64_t>(GetHours())) - (60* static_cast<int64_t>(GetMinutes())));
    }
    else
    {
      CLTimeSpan n{*this};
      n.Negate();
      return n.GetSeconds();
    }
  }

  DTIME_NODISCARD bool IsValid() const noexcept
  {
    return (m_nSeconds != INT64_MIN);
  }

  CLTimeSpan& Negate() noexcept
  {
    if (IsValid())
      m_nSeconds = (-m_nSeconds);
    return *this;
  }

  DTIME_NODISCARD bool IsPositiveSpan() const noexcept
  {
    DTIME_ASSERT(IsValid());

    return (m_nSeconds >= 0);
  }

  DTIME_NODISCARD double SecondsAsDouble() const noexcept
  {
    DTIME_ASSERT(IsValid());

    return static_cast<double>(m_nSeconds);
  }

  CLTimeSpan& operator=(const CLTimeSpan&) = default;
  CLTimeSpan& operator=(CLTimeSpan&&) = default;

//Overloaded arithmetic operators
  CLTimeSpan operator+(const CLTimeSpan& TimeSpan) const noexcept
  {
    if (IsValid() && TimeSpan.IsValid())
    {
      const int64_t nSeconds{m_nSeconds + TimeSpan.m_nSeconds};
      return CLTimeSpan{nSeconds};
    }
    else
      return CLTimeSpan{};
  }

  CLTimeSpan operator-(const CLTimeSpan& TimeSpan) const noexcept
  {
    if (IsValid() && TimeSpan.IsValid())
    {
      const int64_t nSeconds{m_nSeconds - TimeSpan.m_nSeconds};
      return CLTimeSpan{nSeconds};
    }
    else
      return CLTimeSpan{};
  }

  CLTimeSpan& operator+=(const CLTimeSpan& TimeSpan) noexcept
  {
    *this = *this + TimeSpan;
    return *this;
  }

  CLTimeSpan& operator-=(const CLTimeSpan& TimeSpan) noexcept
  {
    *this = *this - TimeSpan;
    return *this;
  }

  friend CLTimeSpan operator-(const CLTimeSpan& TimeSpan) noexcept
  {
    CLTimeSpan rVal{TimeSpan};
    rVal.Negate();
    return rVal;
  }

  CLTimeSpan operator*(uint16_t Multiplier) const noexcept
  {
    if (IsValid())
    {
      const int64_t rVal{m_nSeconds * Multiplier};
      return CLTimeSpan{rVal};
    }
    else
      return CLTimeSpan{};
  }

  CLTimeSpan operator/(uint16_t divisor) const noexcept
  {
    if (IsValid())
    {
      const int64_t rVal{m_nSeconds / divisor};
      return CLTimeSpan{rVal};
    }
    else
      return CLTimeSpan{};
  }

  CLTimeSpan& operator*=(uint16_t Multiplier) noexcept
  {
    *this = operator*(Multiplier);
    return *this;
  }

  CLTimeSpan& operator/=(uint16_t Divisor) noexcept
  {
    *this = operator/(Divisor);
    return *this;
  }

//Overloaded equality operators
  bool operator==(const CLTimeSpan& TimeSpan) const noexcept
  {
    const bool bValid{IsValid()};
    const bool bOtherValid{TimeSpan.IsValid()};
    if (bValid && bOtherValid)
      return (m_nSeconds == TimeSpan.m_nSeconds);
    else
      return (bValid == bOtherValid);
  }

  bool operator>(const CLTimeSpan& TimeSpan) const noexcept
  {
    if (IsValid() && TimeSpan.IsValid())
      return (m_nSeconds > TimeSpan.m_nSeconds);
    else
      return false; //Only makes sense to compare 2 valid values
  }

  bool operator>=(const CLTimeSpan& TimeSpan) const noexcept
  {
    return operator==(TimeSpan) || operator>(TimeSpan);
  }

  bool operator<(const CLTimeSpan& TimeSpan) const noexcept
  {
    if (IsValid() && TimeSpan.IsValid())
      return (m_nSeconds < TimeSpan.m_nSeconds);
    else
      return false; //Only makes sense to compare 2 valid values
  }

  bool operator<=(const CLTimeSpan& TimeSpan) const noexcept
  {
    return operator==(TimeSpan) || operator<(TimeSpan);
  }

  bool operator!=(const CLTimeSpan& TimeSpan) const noexcept
  {
    return !operator==(TimeSpan);
  }

//Display
  /* The Following Format parameters are supported
  %D Total days in this CLTimeSpan
  %H Hours in this CLTimeSpan (00 - 23)
  %M Minutes in the current hour in this CLTimeSpan (00 - 59)
  %S Seconds in the current minute in this CLTimeSpan (00 - 59)
  %% Percent sign
  %#H, %#M, %#S  Remove leading zeros (if any)
  */
  DTIME_NODISCARD String Format(const String& sFormat) const
  {
    //What will be the return value from this function
    String rVal;

    if (IsValid())
    {
      const auto nFmtLength{sFormat.length()};
      for (String::size_type i{0}; i<nFmtLength; i++)
      {
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
        TCHAR c{sFormat[i]};
        if (c == _T('%'))
        {
          ++i;
          if (i < nFmtLength)
          {
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
            c = sFormat[i];
            switch (c)
            {
              case _T('D'):
              {
                dt_ostringstream ss;
                const int32_t Days{GetTotalDays()};
                if (IsPositiveSpan())
                  ss << _T("+");
                else if (Days == 0) //handle negative 0
                  ss << _T("-");
                ss << Days;
                rVal += ss.str();
                break;
              }
              case _T('H'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << GetHours();
                rVal += ss.str();
                break;
              }
              case _T('M'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << GetMinutes();
                rVal += ss.str();
                break;
              }
              case _T('S'):
              {
                dt_ostringstream ss;
                ss << std::setw(2) << std::setfill(_T('0')) << GetSeconds();
                rVal += ss.str();
                break;
              }
              case _T('#'):
              {
                if (i < nFmtLength)
                {
                  ++i;
#ifdef _MSC_VER
#pragma warning(suppress: 26446)
#endif //#ifdef _MSC_VER
                  c = sFormat[i];
                  switch (c)
                  {
                    case _T('H'):
                    {
                      dt_ostringstream ss;
                      ss << GetHours();
                      rVal += ss.str();
                      break;
                    }
                    case _T('M'):
                    {
                      dt_ostringstream ss;
                      ss << GetMinutes();
                      rVal += ss.str();
                      break;
                    }
                    case _T('S'):
                    {
                      dt_ostringstream ss;
                      ss << GetSeconds();
                      rVal += ss.str();
                      break;
                    }
                    default:
                    {
                      rVal += c;
                      break;
                    }
                  }
                }
                break;
              }
              default:
              {
                rVal += c;
                break;
              }
            }
          }
        }
        else
        {
          rVal += c;
        }
      }
    }
    return rVal;
  }

//Serialization
#ifdef _AFX
  void Serialize(CArchive& ar)
  {
    if (ar.IsStoring())
    {
      constexpr const WORD wVersion{0x100};
      ar << wVersion;
      ar.Write(&m_nSeconds, sizeof(m_nSeconds));
    }
    else
    {
      WORD wVersion{0};
      ar >> wVersion;
      ASSERT(wVersion == 0x100);
      ar.Read(&m_nSeconds, sizeof(m_nSeconds));
    }
  }
#endif //#ifdef _AFX

#ifdef _AFX
  friend CArchive& operator<<(CArchive& ar, CLTimeSpan& TimeSpan)
  {
    ASSERT(ar.IsStoring());
    TimeSpan.Serialize(ar);
    return ar;
  }
#endif //#ifdef _AFX

#ifdef _AFX
  friend CArchive& operator>>(CArchive& ar, CLTimeSpan& TimeSpan)
  {
    ASSERT(!ar.IsStoring());
    TimeSpan.Serialize(ar);
    return ar;
  }
#endif //#ifdef _AFX

protected:
//Member variables
  int64_t m_nSeconds; //Actual instance data
};

inline CLTimeSpan operator*(uint16_t Multiplier, const DTimePlus::CLTimeSpan& TimeSpan) noexcept
{
  return TimeSpan.operator*(Multiplier);
}

} //namespace DTimePlus


#endif //#ifndef __DTIMEPLUSCLTIMESPAN_H__
