/*
Module : DTime+HebrewCalendar.h
Purpose: Defines the interface for the DTime+ CHebrewCalendar class
Created: PJN / 01-05-2006

Copyright (c) 2006 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise)
when your product is released in binary form. You are allowed to modify the source code in any way you want
except you cannot modify the copyright details at the top of each module. If you want to distribute source
code with your application, then you are only allowed to distribute versions released by the author. This is
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#if _MSC_VER > 1000
#pragma once
#endif //#if _MSC_VER > 1000

#ifndef __DTIMEPLUSHEBREWCALENDAR_H__
#define __DTIMEPLUSHEBREWCALENDAR_H__


//////////////////// Includes /////////////////////////////////////////////////

#include <array>
#include "DTime+Defines.h"
#include "DTime+Helpers.h"


//////////////////// Classes //////////////////////////////////////////////////

namespace DTimePlus
{

//Enums for the Hebrew Month names
enum class HebrewMonth
{
  Nissan   = 1,
  Iyyar    = 2,
  Sivan    = 3,
  Tammuz   = 4,
  Av       = 5,
  Elul     = 6,
  Tishrei  = 7,
  Cheshvan = 8,
  Kislev   = 9,
  Tevet    = 10,
  Shevat   = 11,
  AdarI    = 12,
  AdarII   = 13
};

//Class which implements the Hebrew calendar
class CHebrewCalendar
{
public:
  inline static int32_t HebrewNewYear(int32_t Year) noexcept
  {
    return HEBREW_EPOCH + HebrewCalendarElapsedDays(Year) + HebrewNewYearDelay(Year);
  }

  inline static uint16_t HebrewNewYearDelay(int32_t Year) noexcept
  {
    const int32_t ny0{HebrewCalendarElapsedDays(Year - 1)};
    const int32_t ny1{HebrewCalendarElapsedDays(Year)};
    const int32_t ny2{HebrewCalendarElapsedDays(Year + 1)};

    if ((ny2 - ny1) == 356)
      return 2;
    else if ((ny1 - ny0) == 382)
      return 1;
    else
      return 0;
  }

  inline static int32_t HebrewCalendarElapsedDays(int32_t Year) noexcept
  {
    //This is the approach we would use if we wanted to avoid the use of floating point arithmetic below
    /*
    int32_t nMonthsElapsed{lquotient(235*Year - 234, 19)};
    int64_t nPartsElapsed{12084 + 13753*(static_cast<int64_t>(nMonthsElapsed))};
    int32_t nDay{static_cast<int32_t>((29*nMonthsElapsed) + lquotient(nPartsElapsed, 25920))};
    */

    const int nDay{static_cast<int>(Molad(Year, static_cast<uint16_t>(HebrewMonth::Tishrei)) - HEBREW_EPOCH + 0.5)};
    return (lmod(3*(nDay+1), 7) < 3) ? nDay + 1 : nDay;
  }

  inline static bool LongCheshvan(int32_t Year) noexcept
  {
    const uint16_t wDaysInYear{DaysInYear(Year)};
    return (wDaysInYear == 355 || wDaysInYear == 385) ? true : false;
  }

  inline static bool ShortCheshvan(int32_t Year) noexcept
  {
    return !LongCheshvan(Year);
  }

  inline static bool LongKislev(int32_t Year) noexcept
  {
    return !ShortKislev(Year);
  }

  inline static bool ShortKislev(int32_t Year) noexcept
  {
    const uint16_t wDaysInYear{DaysInYear(Year)};
    return (wDaysInYear == 353 || wDaysInYear == 383) ? true : false;
  }

  inline static double Molad(int32_t Year, uint16_t month) noexcept
  {
    const int32_t nMonthsElapsed{month - static_cast<int32_t>(HebrewMonth::Tishrei) + lquotient(235 * Year - 234, 19)};
    return HEBREW_EPOCH - 876.0/25920 + nMonthsElapsed*29.53059413580246913580246913580247;
  }

  inline static uint16_t DaysSinceEndPreviousYear(int32_t Year, uint16_t month, uint16_t Day) noexcept
  {
    //Validate our parameters
    DTIME_ASSERT((month >= static_cast<uint16_t>(HebrewMonth::Nissan)) && (month <= LastMonthOfYear(IsLeap(Year))));
    DTIME_ASSERT((Day >= 1) && (Day <= DaysInMonth(Year, month)));

    int nDays{0};
    if (month < static_cast<uint16_t>(HebrewMonth::Tishrei))
    {
      for (auto i{static_cast<uint16_t>(HebrewMonth::Tishrei)}; i<=LastMonthOfYear(IsLeap(Year)); i++)
        nDays += DaysInMonth(Year, i);
      for (auto i{static_cast<uint16_t>(HebrewMonth::Nissan)}; i<month; i++)
        nDays += DaysInMonth(Year, i);
    }
    else
    {
      for (auto i{static_cast<uint16_t>(HebrewMonth::Tishrei)}; i<month; i++)
        nDays += DaysInMonth(Year, i);
    }
    nDays += Day;

#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    return static_cast<uint16_t>(nDays);
  }

  inline static uint16_t DaysSinceFirstOfYear(int32_t Year, uint16_t month, uint16_t Day) noexcept
  {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    return static_cast<uint16_t>(DaysSinceEndPreviousYear(Year, month, Day) - 1);
  }

  inline static uint8_t DaysInMonth(int32_t Year, uint16_t month) noexcept
  {
    if ((month == static_cast<uint16_t>(HebrewMonth::Iyyar)) || (month == static_cast<uint16_t>(HebrewMonth::Tammuz)) || (month == static_cast<uint16_t>(HebrewMonth::Elul)) || (month == static_cast<uint16_t>(HebrewMonth::Tevet)) || (month == static_cast<uint16_t>(HebrewMonth::AdarII)) || ((month == static_cast<uint16_t>(HebrewMonth::AdarI)) && !IsLeap(Year)) ||
        ((month == static_cast<uint16_t>(HebrewMonth::Cheshvan)) && !LongCheshvan(Year)) || ((month == static_cast<uint16_t>(HebrewMonth::Kislev)) && ShortKislev(Year)))
      return 29;
    else
      return 30;
  }

  inline static uint16_t DaysInYear(int32_t Year) noexcept
  {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    return static_cast<uint16_t>(HebrewNewYear(Year+1) - HebrewNewYear(Year));
  }

  constexpr inline static uint16_t LastMonthOfYear(bool IsLeap) noexcept
  {
    return IsLeap ? static_cast<uint16_t>(13) : static_cast<uint16_t>(12);
  }

  static void NumToDMY(int32_t Days, int32_t& Year, uint16_t& Month, uint16_t& Day) noexcept
  {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    auto nApproxYear{static_cast<int32_t>(1 + dquotient((Days + 0.0 - HEBREW_EPOCH), 35975351.0 / 98496))};

    while (HebrewNewYear(nApproxYear) <= Days)
      ++nApproxYear;
    Year = nApproxYear - 1;

    int32_t nTempDays{0};
    DMYToNum(Year, static_cast<uint16_t>(HebrewMonth::Nissan), 1, nTempDays, true);
    Month = static_cast<uint16_t>((Days < nTempDays) ? HebrewMonth::Tishrei : HebrewMonth::Nissan);

    DMYToNum(Year, Month, DaysInMonth(Year, Month), nTempDays, true);
    while (nTempDays < Days)
    {
      ++Month;
      DMYToNum(Year, Month, DaysInMonth(Year, Month), nTempDays, true);
    }

    DMYToNum(Year, Month, 1, nTempDays, true);
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Day = static_cast<uint16_t>(Days - nTempDays + 1);

    int32_t lFirstOfMonth{0};
    DMYToNum(Year, Month, 1, lFirstOfMonth, true);
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Day = static_cast<uint16_t>(Days - lFirstOfMonth + 1);
  }

#ifdef _DEBUG
  static bool DMYToNum(int32_t Year, uint16_t month, uint16_t Day, int32_t& lDays, bool bDoAssert) noexcept
#else
  static bool DMYToNum(int32_t Year, uint16_t month, uint16_t Day, int32_t& lDays, bool /*bDoAssert*/) noexcept
#endif //#ifdef _DEBUG
  {
    //Validate our parameters
    const bool bLeap{IsLeap(Year)};
    if ((month < static_cast<uint16_t>(HebrewMonth::Nissan)) || (month > LastMonthOfYear(bLeap)) || (Day > DaysInMonth(Year, month)))
    {
    #ifdef _DEBUG
      if (bDoAssert)
        DTIME_ASSERT(false);
    #endif //#ifdef _DEBUG
      return false;
    }

    lDays = HebrewNewYear(Year) + Day - 1;
    if (month < static_cast<uint16_t>(HebrewMonth::Tishrei))
    {
      for (auto i{static_cast<uint16_t>(HebrewMonth::Tishrei)}; i<=LastMonthOfYear(bLeap); i++)
        lDays += DaysInMonth(Year, i);

      for (auto i{static_cast<uint16_t>(HebrewMonth::Nissan)}; i<month; i++)
        lDays += DaysInMonth(Year, i);
    }
    else
    {
      for (auto i{static_cast<uint16_t>(HebrewMonth::Tishrei)}; i<month; i++)
        lDays += DaysInMonth(Year, i);
    }
    return true;
  }

  inline static bool IsLeap(int32_t Year) noexcept
  {
    return (lmod((7*Year + 1),19) < 7) ? true : false;
  }

  inline static String GetFullStringDayOfWeek(DayOfWeek dow)
  {
    //Validate our parameters
    DTIME_ASSUME((dow >= DayOfWeek::FirstDayOfWeek) && (dow <= DayOfWeek::LastDayOfWeek));

    static std::array<const TCHAR*, 7> s_HebrewDOW
    {
      _T("yom rishon"),
      _T("yom sheni"),
      _T("yom shelishi"),
      _T("yom revi'i"),
      _T("yom hamishi"),
      _T("yom shishi"),
      _T("yom shabbat")
    };

#ifdef _MSC_VER
#pragma warning(suppress: 26446 26482)
#endif //#ifdef _MSC_VER
    return String{s_HebrewDOW[static_cast<size_t>(dow) - static_cast<size_t>(DayOfWeek::FirstDayOfWeek)]};
  }

  inline static String GetAbrStringDayOfWeek(DayOfWeek dow)
  {
    //There is no concept of an abbreviated day of week name in the Hebrew calendar (That I
    //know of), hence just delegate the work to GetFullStringDayOfWeek
    return GetFullStringDayOfWeek(dow);
  }

  inline static DayOfWeek GetBeginningDayOfWeek() noexcept
  {
    //We do not use the Win32 NLS API's here, instead we just return the hardcoded value of Sunday
    return DayOfWeek::Sunday;
  }

  static String GetFullStringMonth(uint16_t month)
  {
    //Validate our parameters
    DTIME_ASSUME((month >= static_cast<uint16_t>(HebrewMonth::Nissan)) && (month <= static_cast<uint16_t>(HebrewMonth::AdarII)));

    static std::array<const TCHAR*, 13> s_HebrewMonths
    {
      _T("Nissan"),
      _T("Iyyar"),
      _T("Sivan"),
      _T("Tammuz"),
      _T("Av"),
      _T("Elul"),
      _T("Tishrei"),
      _T("Cheshvan"),
      _T("Kislev"),
      _T("Tevet"),
      _T("Shevat"),
      _T("Adar I"),
      _T("Adar II")
    };

#ifdef _MSC_VER
#pragma warning(suppress: 26446 26482)
#endif //#ifdef _MSC_VER
    return String{s_HebrewMonths[month - static_cast<int>(HebrewMonth::Nissan)]};
  }

  inline static String GetAbrStringMonth(uint16_t month)
  {
    //There is no concept of an abbreviated Month name in the Hebrew calendar (That I
    //know of), hence just delegate the work to GetFullStringMonth
    return GetFullStringMonth(month);
  }

protected:
//Enums
  enum
  {
    HEBREW_EPOCH = -1951163 //Corresponds to 7 October 3761 B.C.E. O.S (JD 347997.5)
  };

};

} //namespace DTimePlus

#endif //#ifndef __DTIMEPLUSHEBREWCALENDAR_H__
