/*
Module : DTime+Helpers.h
Purpose: Defines the interface for the various helper functions for DTime+
Created: PJN / 01-05-2006
History: PJN / 01-01-2018 1. Removed defunct dmod method.
                          2. Replaced GetLocaleInfo Win32 API call with GetLocaleInfoEx.
                          3. Replaced GetDateFormat Win32 API call with GetDateFormatEx.
                          4. Replaced GetTimeFormat Win32 API call with GetTimeFormatEx.

Copyright (c) 2006 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise)
when your product is released in binary form. You are allowed to modify the source code in any way you want
except you cannot modify the copyright details at the top of each module. If you want to distribute source
code with your application, then you are only allowed to distribute versions released by the author. This is
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#if _MSC_VER > 1000
#pragma once
#endif //#if _MSC_VER > 1000

#ifndef __DTIMEPLUSHELPERS_H__
#define __DTIMEPLUSHELPERS_H__


//////////////////// Includes /////////////////////////////////////////////////

#include <cmath>
#include <cassert>
#include "DTime+Defines.h"


//////////////////// Classes //////////////////////////////////////////////////

namespace DTimePlus
{
#ifdef _WIN32
  static String SafeGetLocaleInfo(LPCWSTR lpLocaleName, LCTYPE LCType)
  {
    //What will be the return value from this function
    String sInfo;

    //First ask the API to give us the buffer size
    int nChars{GetLocaleInfoEx(lpLocaleName, LCType, nullptr, 0)};
    if (nChars)
    {
      std::wstring sTemp;
      sTemp.resize(nChars);
#pragma warning(suppress: 26446)
      LPWSTR pszBuffer{sTemp.data()};
      nChars = GetLocaleInfoEx(lpLocaleName, LCType, pszBuffer, nChars);
      if (nChars)
      {
#pragma warning(suppress: 26472)
        sTemp.resize(static_cast<size_t>(nChars) - 1);
        sInfo = ATL::CW2T(sTemp.c_str());
      }
    }
    return sInfo;
  }

  inline static String SafeGetDateFormat(LPCWSTR lpLocaleName, DWORD dwFlags, const SYSTEMTIME* pSystemTime) //NOLINT(clang-diagnostic-unused-function)
  {
    //What will be the return value from this function
    String sInfo;

    //First ask the API to give us the buffer size
    int nChars{GetDateFormatEx(lpLocaleName, dwFlags, pSystemTime, nullptr, nullptr, 0, nullptr)};
    if (nChars)
    {
      std::wstring sTemp;
      sTemp.resize(nChars);
#pragma warning(suppress: 26446)
      LPWSTR pszBuffer{sTemp.data()};
      nChars = GetDateFormatEx(lpLocaleName, dwFlags, pSystemTime, nullptr, pszBuffer, nChars, nullptr);
      if (nChars)
      {
#pragma warning(suppress: 26472)
        sTemp.resize(static_cast<size_t>(nChars) - 1);
        sInfo = ATL::CW2T(sTemp.c_str());
      }
    }
    return sInfo;
  }

  inline static String SafeGetTimeFormat(LPCWSTR lpLocaleName, DWORD dwFlags, const SYSTEMTIME* pSystemTime) //NOLINT(clang-diagnostic-unused-function)
  {
    //What will be the return value from this function
    String sInfo;

    //First ask the API to give us the buffer size
    int nChars{GetTimeFormatEx(lpLocaleName, dwFlags, pSystemTime, nullptr, nullptr, 0)};
    if (nChars)
    {
      std::wstring sTemp;
      sTemp.resize(nChars);
#pragma warning(suppress: 26446)
      LPWSTR pszBuffer{sTemp.data()};
      nChars = GetTimeFormatEx(lpLocaleName, dwFlags, pSystemTime, nullptr, pszBuffer, nChars);
      if (nChars)
      {
#pragma warning(suppress: 26472)
        sTemp.resize(static_cast<size_t>(nChars) - 1);
        sInfo = ATL::CW2T(sTemp.c_str());
      }
    }
    return sInfo;
  }
#endif //#ifdef _WIN32

  inline static int32_t lmod(int32_t x, int32_t y) noexcept
  {
    if (x >= 0)
    {
      if (y >= 0)
        return x % y;
      else
        return y - ((-x)%-y);
    }
    else
    {
      if (y >= 0)
        return (y - ((-x)%y)) % y;
      else
        return -(-x%-y);
    }
  }

  inline static int32_t dquotient(double x, double y) noexcept
  {
    return static_cast<int32_t>(floor(x / y));
  }

#ifdef _MSC_VER
#pragma warning(suppress: 26497)
#endif //#ifdef _MSC_VER
  inline static int32_t lquotient(int32_t x, int32_t y) noexcept
  {
    DTIME_ASSERT(y > 0);
    return (x >= 0 ? x/y : (x%y == 0) - 1 - (-x)/y);
  }

#ifdef _MSC_VER
#pragma warning(suppress: 26497)
#endif //#ifdef _MSC_VER
  inline static int32_t lquotient(int64_t x, int32_t y) noexcept
  {
    DTIME_ASSERT(y > 0);

#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    return static_cast<int32_t>(x >= 0 ? x / y : (x%y == 0) - static_cast<int64_t>(1) - (-x) / y);
  }

} //namespace DTimePlus

#endif //#ifndef __DTIMEPLUSHELPERS_H__
