/*
Module : DTime+JulianCalendar.h
Purpose: Defines the interface for the DTime+ CJulianCalendar class
Created: PJN / 01-05-2006

Copyright (c) 1995 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise)
when your product is released in binary form. You are allowed to modify the source code in any way you want
except you cannot modify the copyright details at the top of each module. If you want to distribute source
code with your application, then you are only allowed to distribute versions released by the author. This is
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#if _MSC_VER > 1000
#pragma once
#endif //#if _MSC_VER > 1000

#ifndef __DTIMEPLUSJULIANCALENDAR_H__
#define __DTIMEPLUSJULIANCALENDAR_H__


//////////////////// Includes /////////////////////////////////////////////////

#include <cstdint>
#include "DTime+Defines.h"
#include "DTime+Helpers.h"
#include "DTime+JulianGregorianCalendar.h"


//////////////////// Classes //////////////////////////////////////////////////

namespace DTimePlus
{

class CJulianLeapFunctor
{
public:
  constexpr bool operator()(int32_t Year) const noexcept
  {
    return (Year % 4 == 0);
  }
};

//Class which implements the Julian calendar
class CJulianCalendar : public CJulianGregorianCalendar<CJulianLeapFunctor>
{
public:
  static void NumToDMY(int32_t Days, int32_t& Year, uint16_t& Month, uint16_t& Day) noexcept
  {
    Days += 578043;
    const int32_t era{((Days >= 0) ? Days : (Days - 1460)) / 1461};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    const auto doe{static_cast<unsigned>(Days - (era * 1461))};
    const auto yoe{(doe - (doe / 1460)) / 365};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    int32_t y{static_cast<int32_t>(yoe) + (era * 4)};
    const auto doy{doe - (365 * yoe)};
    const auto mp{((5 * doy) + 2) / 153};
    auto d{doy - (((153 * mp) + 2) / 5) + 1};
    auto m{mp + ((mp < 10) ? 3 : -9)};
    Year = y + (m <= 2);
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Month = static_cast<uint16_t>(m);
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Day = static_cast<uint16_t>(d);
  }

#ifdef _DEBUG
  static bool DMYToNum(int32_t Year, uint16_t month, uint16_t Day, int32_t& lDays, bool bDoAssert) noexcept
#else
  static bool DMYToNum(int32_t Year, uint16_t month, uint16_t Day, int32_t& lDays, bool /*bDoAssert*/) noexcept
#endif //#ifdef _DEBUG
  {
    //Validate our parameters
    if ((month < static_cast<int>(Month::January)) || (month > static_cast<int>(Month::December)) || (Day > DaysInMonth(Year, month)))
    {
    #ifdef _DEBUG
      if (bDoAssert)
        DTIME_ASSERT(false);
    #endif //#ifdef _DEBUG
      return false;
    }

    if (month <= 2)
      --Year;
    int32_t era{((Year >= 0) ? Year : (Year - 3)) / 4};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    const auto yoe{Year - (era * 4)};
    const auto doy{((153 * (month + ((month > 2) ? -3 : 9)) + 2) / 5) + Day - 1};
    auto doe{(yoe * 365) + doy};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    lDays = (era * 1461) + static_cast<int32_t>(doe) - 578043;
    return true;
  }

  constexpr inline static bool IsLeap(int32_t Year) noexcept
  {
    return CJulianLeapFunctor()(Year);
  }

  inline static void EasterSunday(int32_t Year, uint16_t& month, uint16_t& Day) noexcept
  {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    const int a{static_cast<int>(Year % 4)};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    const int b{static_cast<int>(Year % 7)};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    const int c{static_cast<int>(Year % 19)};
    const int d{(19 * c + 15) % 30};
    const int e{(2*a + 4*b - d + 34) % 7};
    int f{(d + e + 114) / 31};
    int g{(d + e + 114) % 31};

    //Set the output parameters
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    month = static_cast<uint16_t>(f);
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Day = static_cast<uint16_t>(g + 1);
  }

};

} //namespace DTimePlus

#endif //#ifndef __DTIMEPLUSJULIANCALENDAR_H__
