/*
Module : DTime+JulianGregorianCalendar.h
Purpose: Defines the interface for the DTime+ CJulianGregorianCalendar class which is the base
         class for the CJulianCalendar and CGregorianCalendar classes
Created: PJN / 01-05-2006

Copyright (c) 2006 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#if _MSC_VER > 1000
#pragma once
#endif //#if _MSC_VER > 1000

#ifndef __DTIMEPLUSJULIANGREGORIANCALENDAR_H__
#define __DTIMEPLUSJULIANGREGORIANCALENDAR_H__


//////////////////// Includes /////////////////////////////////////////////////

#include <cstdint>
#include <array>
#include "DTime+Defines.h"
#include "DTime+Helpers.h"


//////////////////// Classes //////////////////////////////////////////////////

namespace DTimePlus
{

//Base class for the Julian and Gregorian calendars
template<class LEAP_YEAR_FUNCTOR>
class CJulianGregorianCalendar
{
public:
  inline static uint16_t DaysSinceEndPreviousYear(int32_t Year, uint16_t month, uint16_t Day) noexcept
  {
    //Validate our parameters
    DTIME_ASSERT((month >= static_cast<int>(Month::January)) && (month <= static_cast<int>(Month::December)));
    DTIME_ASSERT((Day >= 1) && (Day <= DaysInMonth(Year, month)));

    const int k{LEAP_YEAR_FUNCTOR()(Year) ? 1 : 2};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    const auto rVal{static_cast<uint16_t>((static_cast<int>(275*month/9)) - (k*static_cast<int>((month+9)/12)) + Day - 30)};
    return rVal;
  }

  inline static uint16_t DaysSinceFirstOfYear(int32_t Year, uint16_t month, uint16_t Day) noexcept
  {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    return static_cast<uint16_t>(DaysSinceEndPreviousYear(Year, month, Day) - 1);
  }

  inline static uint8_t DaysInMonth(int32_t Year, uint16_t month) noexcept
  {
    //Validate our parameters
    DTIME_ASSUME((month >= static_cast<uint16_t>(Month::January)) && (month <= static_cast<uint16_t>(Month::December)));

    static std::array<uint8_t, 12> s_MonthLength{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
#ifdef _MSC_VER
#pragma warning(suppress: 26446 26482)
#endif //#ifdef _MSC_VER
    return (month != static_cast<int>(Month::February)) || !LEAP_YEAR_FUNCTOR()(Year) ? s_MonthLength[month - 1] : 29;
  }

  constexpr inline static uint16_t DaysInYear(int32_t Year)
  {
    return LEAP_YEAR_FUNCTOR()(Year) ? static_cast<uint16_t>(366) : static_cast<uint16_t>(365);
  }

  inline static DayOfWeek GetBeginningDayOfWeek()
  {
    //What will be the return value from this function, assume Monday
    DayOfWeek dow{DayOfWeek::Monday};

#ifdef _WIN32
#pragma warning(suppress: 26477)
    String sBuf{SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_IFIRSTDAYOFWEEK)};
    if (sBuf == _T("0"))
      dow = DayOfWeek::Monday;
    else if (sBuf == _T("1"))
      dow = DayOfWeek::Tuesday;
    else if (sBuf == _T("2"))
      dow = DayOfWeek::Wednesday;
    else if (sBuf == _T("3"))
      dow = DayOfWeek::Thursday;
    else if (sBuf == _T("4"))
      dow = DayOfWeek::Friday;
    else if (sBuf == _T("5"))
      dow = DayOfWeek::Saturday;
    else if (sBuf == _T("6"))
      dow = DayOfWeek::Sunday;
    else
      DTIME_ASSERT(false);
#endif //#ifdef _WIN32
    return dow;
  }

#ifdef _WIN32
  inline static String GetFullStringDayOfWeek(DayOfWeek dow)
  {
    //What will be the return value from this function
    String rVal;

    switch (dow)
    {
      case DayOfWeek::Monday:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SDAYNAME1);
        break;
      }
      case DayOfWeek::Tuesday:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SDAYNAME2);
        break;
      }
      case DayOfWeek::Wednesday:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SDAYNAME3);
        break;
      }
      case DayOfWeek::Thursday:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SDAYNAME4);
        break;
      }
      case DayOfWeek::Friday:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SDAYNAME5);
        break;
      }
      case DayOfWeek::Saturday:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SDAYNAME6);
        break;
      }
      case DayOfWeek::Sunday:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SDAYNAME7);
        break;
      }
      default:
      {
        DTIME_ASSERT(false);
        break;
      }
    }
    return rVal;
  }
#endif //#ifdef _WIN32

#ifdef _WIN32
  inline static String GetAbrStringDayOfWeek(DayOfWeek dow)
  {
    //What will be the return value from this function
    String rVal;

    switch (dow)
    {
      case DayOfWeek::Monday:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVDAYNAME1);
        break;
      }
      case DayOfWeek::Tuesday:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVDAYNAME2);
        break;
      }
      case DayOfWeek::Wednesday:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVDAYNAME3);
        break;
      }
      case DayOfWeek::Thursday:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVDAYNAME4);
        break;
      }
      case DayOfWeek::Friday:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVDAYNAME5);
        break;
      }
      case DayOfWeek::Saturday:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVDAYNAME6);
        break;
      }
      case DayOfWeek::Sunday:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVDAYNAME7);
        break;
      }
      default:
      {
        DTIME_ASSERT(false);
        break;
      }
    }
    return rVal;
  }
#endif //#ifdef _WIN32

#ifdef _WIN32
  inline static String GetFullStringMonth(uint16_t month)
  {
    //What will be the return value from this function
    String rVal;

    switch (static_cast<Month>(month))
    {
      case Month::January:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SMONTHNAME1);
        break;
      }
      case Month::February:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SMONTHNAME2);
        break;
      }
      case Month::March:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SMONTHNAME3);
        break;
      }
      case Month::April:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SMONTHNAME4);
        break;
      }
      case Month::May:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SMONTHNAME5);
        break;
      }
      case Month::June:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SMONTHNAME6);
        break;
      }
      case Month::July:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SMONTHNAME7);
        break;
      }
      case Month::August:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SMONTHNAME8);
        break;
      }
      case Month::September:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SMONTHNAME9);
        break;
      }
      case Month::October:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SMONTHNAME10);
        break;
      }
      case Month::November:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SMONTHNAME11);
        break;
      }
      case Month::December:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SMONTHNAME12);
        break;
      }
      default:
      {
        DTIME_ASSERT(false);
        break;
      }
    }
    return rVal;
  }
#endif //#ifdef _WIN32

#ifdef _WIN32
  inline static String GetAbrStringMonth(uint16_t month)
  {
    //What will be the return value from this function
    String rVal;

    switch (static_cast<Month>(month))
    {
      case Month::January:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVMONTHNAME1); 
        break;
      }
      case Month::February:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVMONTHNAME2);
        break;
      }
      case Month::March:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVMONTHNAME3);
        break;
      }
      case Month::April:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVMONTHNAME4);
        break;
      }
      case Month::May:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVMONTHNAME5);
        break;
      }
      case Month::June:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVMONTHNAME6);
        break;
      }
      case Month::July:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVMONTHNAME7);
        break;
      }
      case Month::August:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVMONTHNAME8);
        break;
      }
      case Month::September:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVMONTHNAME9);
        break;
      }
      case Month::October:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVMONTHNAME10);
        break;
      }
      case Month::November:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVMONTHNAME11);
        break;
      }
      case Month::December:
      {
#pragma warning(suppress: 26477)
        rVal = SafeGetLocaleInfo(LOCALE_NAME_USER_DEFAULT, LOCALE_SABBREVMONTHNAME12);
        break;
      }
      default:
      {
        DTIME_ASSERT(false);
        break;
      }
    }
    return rVal;
  }
#endif //#ifdef _WIN32
};

} //namespace DTimePlus

#endif //#ifndef __DTIMEPLUSJULIANGREGORIANCALENDAR_H__
