/*
Module : DTime+JulianGregorianHolidays.H
Purpose: Defines the interface for the DTime+ CJulianGregorianHolidays class which defines various methods
         to create holidays, feasts, holy days, Bank Holidays etc in the Julian or Gregorian calendar
Created: PJN / 29-05-2006

Copyright (c) 2006 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#if _MSC_VER > 1000
#pragma once
#endif //#if _MSC_VER > 1000

#ifndef __DTIMEPLUSJULIANGREGORIANHOLIDAYS_H__
#define __DTIMEPLUSJULIANGREGORIANHOLIDAYS_H__


//////////////////// Includes /////////////////////////////////////////////////

#include <cstdint>
#include "DTime+Defines.h"


//////////////////// Classes //////////////////////////////////////////////////

namespace DTimePlus
{

template<class TCALENDAR = CGregorianCalendar>
class CJulianGregorianHolidays
{
public:
//Well known worldwide holidays
  inline static CDate<TCALENDAR> NewYearsDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::January), 1};
  }

  inline static CDate<TCALENDAR> AprilFoolsDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::April), 1};
  }

  inline static CDate<TCALENDAR> Halloween(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::October), 31};
  }


//Saints Feast Days
  inline static CDate<TCALENDAR> StValentinesDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::February), 14};
  }

  inline static CDate<TCALENDAR> StPatricksDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::March), 17};
  }

  inline static CDate<TCALENDAR> StAndrewsDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::November), 30};
  }

  inline static CDate<TCALENDAR> StPaulsDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::January), 25};
  }

  inline static CDate<TCALENDAR> StGeorgesDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::April), 23};
  }

  inline static CDate<TCALENDAR> StDavidsDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::March), 1};
  }

  inline static CDate<TCALENDAR> StStephensDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::December), 26};
  }


//Other National holidays
  inline static CDate<TCALENDAR> CanadaDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::July), 1}; //Note that we always use July 1st and not the rule: July 2nd if July 1st falls on a Sunday
  }

  inline static CDate<TCALENDAR> BastilleDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::July), 14};
  }

  inline static CDate<TCALENDAR> UKBoxingDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::December), 26}; //Aka St. Stephens day in Ireland
  }

  inline static CDate<TCALENDAR> CanadaThanksgivingDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::October), 2, DayOfWeek::Monday};
  }


//US Specific Holidays
  inline static CDate<TCALENDAR> MartinLutherKingDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::January), 3, DayOfWeek::Monday};
  }

  inline static CDate<TCALENDAR> AbrahamLincolnsBirthday(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::February), 12};
  }

  inline static CDate<TCALENDAR> GeorgeWashingtonsBirthday(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::February), 22};
  }

  inline static CDate<TCALENDAR> USPresidentsDay(int32_t Year) noexcept
  {
    if (Year >= 1971)
      return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::February), 3, DayOfWeek::Monday};
    else
      return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::February), 22};
  }

  inline static CDate<TCALENDAR> USArmedForcesDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::May), 3, DayOfWeek::Sunday};
  }

  inline static CDate<TCALENDAR> USMemorialDay(int32_t Year) noexcept
  {
    if (Year >= 1971)
      return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::May), 5, DayOfWeek::Monday};
    else
      return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::May), 30};
  }

  inline static CDate<TCALENDAR> USFlagDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::June), 14};
  }

  inline static CDate<TCALENDAR> USIndependenceDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::July), 4};
  }

  inline static CDate<TCALENDAR> USLaborDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::September), 1, DayOfWeek::Monday};
  }

  inline static CDate<TCALENDAR> USColumbusDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::October), 2, DayOfWeek::Monday};
  }

  inline static CDate<TCALENDAR> USElectionDay(int32_t Year) noexcept
  {
    CDate<TCALENDAR> ElectionDay{Year, static_cast<uint16_t>(Month::November), 2};
    ElectionDay += ((static_cast<int>(DayOfWeek::Tuesday) + 7 - static_cast<int>(ElectionDay.GetDayOfWeek())) % 7);
    return ElectionDay;
  }

  inline static CDate<TCALENDAR> USVeteransDay(int32_t Year) noexcept
  {
    if ((Year >= 1971) && (Year <= 1977))
      return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::October), 4, DayOfWeek::Monday};
    else
      return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::November), 11};
  }

  inline static CDate<TCALENDAR> USThanksgivingDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::November), 4, DayOfWeek::Thursday};
  }

  inline static CDate<TCALENDAR> GroundhogDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::February), 2};
  }


//Christian Feast Days
  inline static CDate<TCALENDAR> Epiphany(int32_t Year) noexcept             //Note DTime+ uses the traditional Roman catholic date for Epiphany, instead of the 
  {                                                                          //newer definition which is the First Sunday after January 1st. For more info
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::January), 6}; //please see http://en.wikipedia.org/wiki/Epiphany
  }

  inline static CDate<TCALENDAR> Candlemas(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::February), 2};
  }

  inline static CDate<TCALENDAR> ShroveTuesday(int32_t Year) noexcept
  {
    return EasterSunday(Year) - 47; //Shrove Tuesday occurs 47 days before Easter Sunday
  }

  inline static CDate<TCALENDAR> AshWednesday(int32_t Year) noexcept
  {
    return EasterSunday(Year) - 46; //Ash Wednesday occurs 46 days before Easter Sunday
  }

  inline static CDate<TCALENDAR> FeastOfTheAnnunciation(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::March), 25};
  }

  inline static CDate<TCALENDAR> PalmSunday(int32_t Year) noexcept
  {
    return EasterSunday(Year) - 7; //Palm Sunday occurs 7 days before Easter Sunday
  }

  inline static CDate<TCALENDAR> MaundyThursday(int32_t Year) noexcept
  {
    return EasterSunday(Year) - 3; //Maundy (or Holy) Thursday occurs 3 days before Easter Sunday
  }

  inline static CDate<TCALENDAR> GoodFriday(int32_t Year) noexcept
  {
    return EasterSunday(Year) - 2; //Good Friday occurs 2 days before Easter Sunday
  }

  inline static CDate<TCALENDAR> EasterSunday(int32_t Year) noexcept
  {
    //Call the helper function to calculate Easter Sunday for us
    uint16_t wMonth{0};
    uint16_t wDay{0};
    TCALENDAR::EasterSunday(Year, wMonth, wDay);
    return CDate<TCALENDAR>(Year, wMonth, wDay);
  }

  inline static CDate<TCALENDAR> EasterMonday(int32_t Year) noexcept
  {
    return EasterSunday(Year) + 1; //Easter Monday is the day after Easter Sunday
  }

  inline static CDate<TCALENDAR> FeastOfTheAscension(int32_t Year) noexcept
  {
    return EasterSunday(Year) + 39; //Ascension Day occurs 39 days after Easter Sunday
  }

  inline static CDate<TCALENDAR> Pentecost(int32_t Year) noexcept
  {
    return EasterSunday(Year) + 49; //Pentecost occurs 49 days after Easter Sunday
  }

  inline static CDate<TCALENDAR> FeastOfTheAssumption(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::August), 15};
  }

  inline static CDate<TCALENDAR> AllSaintsDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::November), 1};
  }

  inline static CDate<TCALENDAR> AllSoulsDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::November), 2};
  }

  inline static CDate<TCALENDAR> FeastOfTheImmaculateConception(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::December), 8};
  }

  inline static CDate<TCALENDAR> ChristmasDay(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(Month::December), 25};
  }
};

} //namespace DTimePlus

#endif //#ifndef __DTIMEPLUSJULIANGREGORIANHOLIDAYS_H__
