/*
Module : DTime+MoslemCalendar.h
Purpose: Defines the interface for the DTime+ CMoslemCalendar class
Created: PJN / 01-05-2006

Copyright (c) 2006 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#if _MSC_VER > 1000
#pragma once
#endif //#if _MSC_VER > 1000

#ifndef __DTIMEPLUSMOSLEMCALENDAR_H__
#define __DTIMEPLUSMOSLEMCALENDAR_H__


//////////////////// Includes /////////////////////////////////////////////////

#include <cstdint>
#include <array>
#include "DTime+Defines.h"
#include "DTime+Helpers.h"


//////////////////// Classes //////////////////////////////////////////////////

namespace DTimePlus
{

//Enums for the Moslem Month names
enum class MoslemMonth
{
  Muharram    = 1,
  Safar       = 2,
  RabiI       = 3,
  RabiII      = 4,
  JumadaI     = 5,
  JumadaII    = 6,
  Rajab       = 7,
  Sha_ban     = 8,
  Ramadan     = 9,
  Shawwal     = 10,
  Dhu_l_Qa_da = 11,
  Dhu_l_Hijja = 12
};

//Base class for the Moslem calendar variants
template<class LEAP_YEAR_FUNCTOR>
class CBaseMoslemCalendar
{
public:
  inline static uint16_t DaysSinceEndPreviousYear(int32_t Year, uint16_t month, uint16_t Day)
  { 
    //Validate our parameters
    DTIME_ASSERT((month >= static_cast<uint16_t>(MoslemMonth::Muharram)) && (month <= static_cast<uint16_t>(MoslemMonth::Dhu_l_Hijja)));
    DTIME_ASSERT((Day >= 1) && (Day <= DaysInMonth(Year, month)));

    //Count up the days
    uint16_t wDays{0};
    for (uint16_t i{1}; i<month; i++)
      wDays = static_cast<uint16_t>(wDays + DaysInMonth(Year, i));
    wDays = static_cast<uint16_t>(wDays + Day);

    return wDays;
  }

  inline static uint16_t DaysSinceFirstOfYear(int32_t Year, uint16_t month, uint16_t Day)
  {
    return static_cast<uint16_t>(DaysSinceEndPreviousYear(Year, month, Day) - 1);
  }

  inline static uint8_t DaysInMonth(int32_t Year, uint16_t month) noexcept
  {
    //Validate our parameters
    DTIME_ASSUME((month >= static_cast<uint16_t>(MoslemMonth::Muharram)) && (month <= static_cast<uint16_t>(MoslemMonth::Dhu_l_Hijja)));

    static std::array<uint8_t, 12> s_MonthLength{30, 29, 30, 29, 30, 29, 30, 29, 30, 29, 30, 29};

#ifdef _MSC_VER
#pragma warning(suppress: 26446 26482)
#endif //#ifdef _MSC_VER
    return (month != static_cast<uint16_t>(MoslemMonth::Dhu_l_Hijja)) || !LEAP_YEAR_FUNCTOR()(Year) ? s_MonthLength[month - 1] : 30;
  }

  inline static uint16_t DaysInYear(int32_t Year) noexcept
  {
    return LEAP_YEAR_FUNCTOR()(Year) ? static_cast<uint16_t>(355) : static_cast<uint16_t>(354);
  }

  inline static DayOfWeek GetBeginningDayOfWeek() noexcept
  {
    //We do not use the Win32 NLS API's here, instead we just return the hardcoded value of Sunday
    return DayOfWeek::Sunday;
  }

  inline static String GetFullStringDayOfWeek(DayOfWeek dow)
  {
    //Validate our parameters
    DTIME_ASSUME((dow >= DayOfWeek::FirstDayOfWeek) && (dow <= DayOfWeek::LastDayOfWeek));

    static std::array<const TCHAR*, 7> s_MoslemDOW
    {
      _T("yaum al-ahad"),
      _T("yaum al-ithnayna"),
      _T("yaum ath-thalatha'"),
      _T("yaum al-arba'a'"),
      _T("yaum al-hamis"),
      _T("yaum al-jum'a"),
      _T("yaum as-sabt")
    };

#ifdef _MSC_VER
#pragma warning(suppress: 26446 26482)
#endif //#ifdef _MSC_VER
    return String{s_MoslemDOW[static_cast<size_t>(dow) - static_cast<size_t>(DayOfWeek::FirstDayOfWeek)]};
  }

  inline static String GetAbrStringDayOfWeek(DayOfWeek dow)
  {
    //There is no concept of an abbreviated day of week name in the Moslem calendar (That I 
    //know of), hence just delegate the work to GetFullStringDayOfWeek
    return GetFullStringDayOfWeek(dow);
  }

  inline static String GetFullStringMonth(uint16_t month)
  {
    //Validate our parameters
    DTIME_ASSUME(month >= static_cast<uint16_t>(MoslemMonth::Muharram) && month <= static_cast<uint16_t>(MoslemMonth::Dhu_l_Hijja));

    static std::array<const TCHAR*, 12> s_MoslemMonths
    {
      _T("Muharram"),
      _T("Safar"),
      _T("Rabi I"),
      _T("Rabi II"),
      _T("Jumada I"),
      _T("Jumada II"),
      _T("Rajab"),
      _T("Sha'ban"),
      _T("Ramadan"),
      _T("Shawwal"),
      _T("Dhu al-Qa'da"),
      _T("Dhu al-Hijja")
    };

#ifdef _MSC_VER
#pragma warning(suppress: 26446 26482)
#endif //#ifdef _MSC_VER
    return String{s_MoslemMonths[month - static_cast<int>(MoslemMonth::Muharram)]};
  }

  inline static String GetAbrStringMonth(uint16_t month)
  {
    //There is no concept of an abbreviated Month name in the Moslem calendar (That I 
    //know of), hence just delegate the work to GetFullStringMonth
    return GetFullStringMonth(month);
  }
};

class CMoslem16LeapFunctor
{
public:
  bool operator()(int32_t Year) const noexcept
  {
    return lmod(14 + 11*Year, 30) < 11;
  }
};

//Class which implements the Moslem calendar with 16 July 622 CE O.S. as the epoch (so called Civil Epoch) and year 16 in the 30 year cycle as a leap year
class CMoslemCalendarCivil16 : public CBaseMoslemCalendar<CMoslem16LeapFunctor>
{
public:
  static void NumToDMY(int32_t Days, int32_t& Year, uint16_t& Month, uint16_t& Day) noexcept
  {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Year = lquotient(30 * (static_cast<int64_t>(Days) - ISLAMIC_EPOCH) + 10646, 10631);
    int32_t lFirstOfYear{0};
    DMYToNum(Year, 1, 1, lFirstOfYear, true);
    const int32_t lPriorDays{Days - lFirstOfYear};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Month = static_cast<uint16_t>(lquotient((11*lPriorDays + 330), 325));
    int32_t lFirstOfMonth{0};
    DMYToNum(Year, Month, 1, lFirstOfMonth, true);
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Day = static_cast<uint16_t>(Days - lFirstOfMonth + 1);
  }

  inline static bool IsLeap(int32_t Year) noexcept
  {
    return CMoslem16LeapFunctor()(Year);
  }

#ifdef _DEBUG
  static bool DMYToNum(int32_t Year, uint16_t month, uint16_t Day, int32_t& lDays, bool bDoAssert) noexcept
#else
  static bool DMYToNum(int32_t Year, uint16_t month, uint16_t Day, int32_t& lDays, bool /*bDoAssert*/) noexcept
#endif //#ifdef _DEBUG
  {
    //Validate our parameters
    if ((month < static_cast<uint16_t>(MoslemMonth::Muharram)) || (month > static_cast<uint16_t>(MoslemMonth::Dhu_l_Hijja)) || (Day > DaysInMonth(Year, month)))
    {
    #ifdef _DEBUG
      if (bDoAssert)
        DTIME_ASSERT(false);
    #endif //#ifdef _DEBUG
      return false;
    }

    lDays = Day + 29*(month-1) + lquotient((6*month-1), 11) + (Year-1)*354 + lquotient(3+11*Year, 30) + ISLAMIC_EPOCH - 1;
    return true;
  }

protected:
//Enums
  enum
  {
    ISLAMIC_EPOCH = -350721
  };
};

class CMoslem15LeapFunctor
{
public:
  bool operator()(int32_t Year) const noexcept
  {
    return lmod(15 + 11*Year, 30) < 11;
  }
};

//Class which implements the Moslem calendar with 16 July 622 CE O.S. as the epoch (so called Civil Epoch) and year 15 in the 30 year cycle as a leap year
class CMoslemCalendarCivil15 : public CBaseMoslemCalendar<CMoslem15LeapFunctor>
{
public:
  static void NumToDMY(int32_t Days, int32_t& Year, uint16_t& Month, uint16_t& Day) noexcept
  {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Year = lquotient(30 * (static_cast<int64_t>(Days) - ISLAMIC_EPOCH) + 10645, 10631);
    int32_t lFirstOfYear{0};
    DMYToNum(Year, 1, 1, lFirstOfYear, true);
    const int32_t lPriorDays{Days - lFirstOfYear};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Month = static_cast<uint16_t>(lquotient((11*lPriorDays + 330), 325));
    int32_t lFirstOfMonth{0};
    DMYToNum(Year, Month, 1, lFirstOfMonth, true);
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Day = static_cast<uint16_t>(Days - lFirstOfMonth + 1);
  }

  inline static bool IsLeap(int32_t Year) noexcept
  {
    return CMoslem15LeapFunctor()(Year);
  }

#ifdef _DEBUG
  static bool DMYToNum(int32_t Year, uint16_t month, uint16_t Day, int32_t& lDays, bool bDoAssert) noexcept
#else
  static bool DMYToNum(int32_t Year, uint16_t month, uint16_t Day, int32_t& lDays, bool /*bDoAssert*/) noexcept
#endif //#ifdef _DEBUG
  {
    //Validate our parameters
    if ((month < static_cast<uint16_t>(MoslemMonth::Muharram)) || (month > static_cast<uint16_t>(MoslemMonth::Dhu_l_Hijja)) || (Day > DaysInMonth(Year, month)))
    {
    #ifdef _DEBUG
      if (bDoAssert)
        DTIME_ASSERT(false);
    #endif //#ifdef _DEBUG
      return false;
    }

    lDays = Day + 29*(month-1) + lquotient(6*month-1, 11) + (Year-1)*354 + lquotient(4+11*Year, 30) + ISLAMIC_EPOCH - 1;
    return true;
  }

protected:
//Enums
  enum
  {
    ISLAMIC_EPOCH = -350721
  };
};

//Class which implements the Moslem calendar with 15 July 622 CE O.S. as the epoch (so called Astronomical Epoch) and year 16 in the 30 year cycle as a leap year
class CMoslemCalendarAstronomical16 : public CBaseMoslemCalendar<CMoslem16LeapFunctor>
{
public:
  static void NumToDMY(int32_t Days, int32_t& Year, uint16_t& Month, uint16_t& Day) noexcept
  {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Year = lquotient(30 * (static_cast<int64_t>(Days) - ISLAMIC_EPOCH) + 10646, 10631);
    int32_t lFirstOfYear{0};
    DMYToNum(Year, 1, 1, lFirstOfYear, true);
    const int32_t lPriorDays{Days - lFirstOfYear};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Month = static_cast<uint16_t>(lquotient((11*lPriorDays + 330), 325));
    int32_t lFirstOfMonth{0};
    DMYToNum(Year, Month, 1, lFirstOfMonth, true);
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Day = static_cast<uint16_t>(Days - lFirstOfMonth + 1);
  }

  inline static bool IsLeap(int32_t Year) noexcept
  {
    return CMoslem16LeapFunctor()(Year);
  }

#ifdef _DEBUG
  static bool DMYToNum(int32_t Year, uint16_t month, uint16_t Day, int32_t& lDays, bool bDoAssert) noexcept
#else
  static bool DMYToNum(int32_t Year, uint16_t month, uint16_t Day, int32_t& lDays, bool /*bDoAssert*/) noexcept
#endif //#ifdef _DEBUG
  {
    //Validate our parameters
    if ((month < static_cast<uint16_t>(MoslemMonth::Muharram)) || (month > static_cast<uint16_t>(MoslemMonth::Dhu_l_Hijja)) || (Day > DaysInMonth(Year, month)))
    {
    #ifdef _DEBUG
      if (bDoAssert)
        DTIME_ASSERT(false);
    #endif //#ifdef _DEBUG
      return false;
    }

    lDays = Day + 29*(month-1) + lquotient((6*month-1), 11) + (Year-1)*354 + lquotient(3+11*Year, 30) + ISLAMIC_EPOCH - 1;
    return true;
  }

protected:
//Enums
  enum
  {
    ISLAMIC_EPOCH = -350722
  };
};

//Class which implements the Moslem calendar with 15 July 622 CE O.S. as the epoch (so called Astronomical Epoch) and year 15 in the 30 year cycle as a leap year
class CMoslemCalendarAstronomical15 : public CBaseMoslemCalendar<CMoslem15LeapFunctor>
{
public:
  static void NumToDMY(int32_t Days, int32_t& Year, uint16_t& Month, uint16_t& Day) noexcept
  {
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Year = lquotient(30 * (static_cast<int64_t>(Days) - ISLAMIC_EPOCH) + 10645, 10631);
    int32_t lFirstOfYear{0};
    DMYToNum(Year, 1, 1, lFirstOfYear, true);
    const int32_t lPriorDays{Days - lFirstOfYear};
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Month = static_cast<uint16_t>(lquotient((11*lPriorDays + 330), 325));
    int32_t lFirstOfMonth{0};
    DMYToNum(Year, Month, 1, lFirstOfMonth, true);
#ifdef _MSC_VER
#pragma warning(suppress: 26472)
#endif //#ifdef _MSC_VER
    Day = static_cast<uint16_t>(Days - lFirstOfMonth + 1);
  }

  inline static bool IsLeap(int32_t Year) noexcept
  {
    return CMoslem15LeapFunctor()(Year);
  }

#ifdef _DEBUG
  static bool DMYToNum(int32_t Year, uint16_t month, uint16_t Day, int32_t& lDays, bool bDoAssert) noexcept
#else
  static bool DMYToNum(int32_t Year, uint16_t month, uint16_t Day, int32_t& lDays, bool /*bDoAssert*/) noexcept
#endif //#ifdef _DEBUG
  {
    //Validate our parameters
    if ((month < static_cast<uint16_t>(MoslemMonth::Muharram)) || (month > static_cast<uint16_t>(MoslemMonth::Dhu_l_Hijja)) || (Day > DaysInMonth(Year, month)))
    {
    #ifdef _DEBUG
      if (bDoAssert)
        DTIME_ASSERT(false);
    #endif //#ifdef _DEBUG
      return false;
    }

    lDays = Day + 29*(month-1) + lquotient(6*month-1, 11) + (Year-1)*354 + lquotient(4+11*Year, 30) + ISLAMIC_EPOCH - 1;
    return true;
  }

protected:
//Enums
  enum
  {
    ISLAMIC_EPOCH = -350722
  };
};

} //namespace DTimePlus

#endif //#ifndef __DTIMEPLUSMOSLEMCALENDAR_H__
