/*
Module : DTime+MoslemHolidays.H
Purpose: Defines the interface for the DTime+ CMoslemHolidays class which defines various methods
         to create holidays in the Moslem calendar
Created: PJN / 29-05-2006

Copyright (c) 2006 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#if _MSC_VER > 1000
#pragma once
#endif //#if _MSC_VER > 1000

#ifndef __DTIMEPLUSMOSLEMHOLIDAYS_H__
#define __DTIMEPLUSMOSLEMHOLIDAYS_H__


//////////////////// Includes /////////////////////////////////////////////////

#include <vector>
#include "DTime+MoslemCalendar.h"
#include "DTime+GregorianCalendar.h"
#include "DTime+CDate.h"


//////////////////// Classes //////////////////////////////////////////////////

namespace DTimePlus
{

template<class TCALENDAR>
class CMoslemNewYearFunctor
{
public:
  CDate<TCALENDAR> operator()(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(MoslemMonth::Muharram), 1};
  }
};

template<class TCALENDAR>
class CAshuraFunctor
{
public:
  CDate<TCALENDAR> operator()(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(MoslemMonth::Muharram), 10};
  }
};

template<class TCALENDAR>
class CNawlidAnNabiFunctor
{
public:
  CDate<TCALENDAR> operator()(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(MoslemMonth::RabiI), 12};
  }
};

template<class TCALENDAR>
class CLailatAlMiRajFunctor
{
public:
  CDate<TCALENDAR> operator()(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(MoslemMonth::Rajab), 27};
  }
};

template<class TCALENDAR>
class CLailatAlBaraaFunctor
{
public:
  CDate<TCALENDAR> operator()(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(MoslemMonth::Sha_ban), 15};
  }
};

template<class TCALENDAR>
class CRamdadanFunctor
{
public:
  CDate<TCALENDAR> operator()(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(MoslemMonth::Ramadan), 1};
  }
};

template<class TCALENDAR>
class CLailatAlKadrFunctor
{
public:
  CDate<TCALENDAR> operator()(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(MoslemMonth::Ramadan), 27};
  }
};

template<class TCALENDAR>
class CIdAlFitrFunctor
{
public:
  CDate<TCALENDAR> operator()(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(MoslemMonth::Shawwal), 1};
  }
};

template<class TCALENDAR>
class CIdAlAdhaFunctor
{
public:
  CDate<TCALENDAR> operator()(int32_t Year) noexcept
  {
    return CDate<TCALENDAR>{Year, static_cast<uint16_t>(MoslemMonth::Dhu_l_Hijja), 10};
  }
};

//Function which returns all occurrences of a specific date in the Moslem calendar (generated by a functor)
//in a specified Gregorian year. Note that because we are using the flexibility of functors, this function
//is reusable if we ever had the requirement to calculate Moslem holidays which do not occur on simple
//fixed days in the Moslem calendar
template<class TDATEFUNCTOR, class TCALENDAR>
inline static void MoslemOccurrencesInGregorianYear(int32_t Year, std::vector<CDate<CGregorianCalendar> >& Occurrences)
{
  //Empty out the array before we add any values to it
  Occurrences.clear();

  const CDate<CGregorianCalendar> jan1{Year, static_cast<uint16_t>(Month::January), 1};
  const int32_t jan1value{jan1.GetValue()};
  const CDate<CGregorianCalendar> dec31{Year, static_cast<uint16_t>(Month::December), 31};
  const int32_t dec31value{dec31.GetValue()};
  const int32_t y{CDate<TCALENDAR>(jan1value, DateEpoch::EPOCH_GREG).GetYear()};

  TDATEFUNCTOR tdate;
  CDate<TCALENDAR> date{tdate(y - 2)};
  int32_t datevalue{date.GetValue()};
  if ((jan1value <= datevalue) && (datevalue <= dec31value))
    Occurrences.emplace_back(datevalue, DateEpoch::EPOCH_GREG);

  date = tdate(y - 1);
  datevalue = date.GetValue();
  if ((jan1value <= datevalue) && (datevalue <= dec31value))
  {
#ifdef _MSC_VER
#pragma warning(suppress: 26489)
#endif //#ifdef _MSC_VER
    Occurrences.emplace_back(datevalue, DateEpoch::EPOCH_GREG);
  }

  date = tdate(y);
  datevalue = date.GetValue();
  if ((jan1value <= datevalue) && (datevalue <= dec31value))
  {
#ifdef _MSC_VER
#pragma warning(suppress: 26489)
#endif //#ifdef _MSC_VER
    Occurrences.emplace_back(datevalue, DateEpoch::EPOCH_GREG);
  }

  date = tdate(y + 1);
  datevalue = date.GetValue();
  if ((jan1value <= datevalue) && (datevalue <= dec31value))
  {
#ifdef _MSC_VER
#pragma warning(suppress: 26489)
#endif //#ifdef _MSC_VER
    Occurrences.emplace_back(datevalue, DateEpoch::EPOCH_GREG);
  }

  date = tdate(y + 2);
  datevalue = date.GetValue();
  if ((jan1value <= datevalue) && (datevalue <= dec31value))
  {
#ifdef _MSC_VER
#pragma warning(suppress: 26489)
#endif //#ifdef _MSC_VER
    Occurrences.emplace_back(datevalue, DateEpoch::EPOCH_GREG);
  }
}

} //namespace DTimePlus

#endif //#ifndef __DTIMEPLUSMOSLEMHOLIDAYS_H__
